{------------------------------------------------------------------------------}
{                                                                              }
{                                 This code is                                 }
{              Copyright (C) 2005-2010 by Michael in der Wiesche               }
{                                                                              }
{------------------------------------------------------------------------------}
{$J+,Z4}
unit CertStore;

interface

uses
  Windows, Messages, SysUtils, UTF8, PGPStrTools, PSMimeTools,
  pgpBase, pgpErrors, pgpMemoryMgr, X509Types, X509Funcs;

// CertData may either be a BASE64 or PKCS#7 encoded certificate (chain) or a BASE64-decoded S/MIME signature
function SystemHasPrivateCerts: Longbool;
function ShowCertificateDialog(const CertData, Title: RawByteString; FileName: Longbool; WinHandle: THandle): Integer;
function ImportPrivateSystemCerts(const CertKeyPrompt, CertPassPrompt: RawByteString; TempPassphrase: PAnsiChar;
				  var CertPropsList: TX509CertPropsList; WinHandle: THandle): Integer;


implementation

uses
  PGPDialogs;


// declarations translated from WinCrypt.h

type
  CRYPTOAPI_BLOB = Record
    cbData: DWORD;
    pbData: PBYTE;
  end;
  CRYPT_DATA_BLOB = CRYPTOAPI_BLOB;
  PCRYPT_DATA_BLOB = ^CRYPT_DATA_BLOB;
  CRYPT_HASH_BLOB =  CRYPTOAPI_BLOB;
  PCRYPT_HASH_BLOB = ^CRYPT_HASH_BLOB;
  CRYPT_INTEGER_BLOB = CRYPTOAPI_BLOB;
  PCRYPT_INTEGER_BLOB = ^CRYPT_INTEGER_BLOB;
  CERT_NAME_BLOB = CRYPTOAPI_BLOB;
  PCERT_NAME_BLOB = ^CERT_NAME_BLOB;
  CRYPT_OBJID_BLOB = CRYPTOAPI_BLOB;
  PCRYPT_OBJID_BLOB = ^CRYPT_OBJID_BLOB;

type
  HCRYPTPROV = Longint;
  HCERTSTORE = Pointer;

const
  MY	= 'MY';         // Cert Store holds certs with associated Private Keys
  CA	= 'CA';		// Certifying Authority certs
  ROOT	= 'ROOT';	// Root Certs
  SPC	= 'SPC';	// Software Publisher Certs

var
  CertOpenSystemStore: function(hProv: HCRYPTPROV; szSubsystemProtocol: PAnsiChar): HCERTSTORE; stdcall;

const
  CERT_STORE_PROV_MSG		= PAnsiChar(1);
  CERT_STORE_PROV_MEMORY	= PAnsiChar(2);
  CERT_STORE_PROV_FILE		= PAnsiChar(3);
  CERT_STORE_PROV_REG		= PAnsiChar(4);
  CERT_STORE_PROV_PKCS7		= PAnsiChar(5);
  CERT_STORE_PROV_SERIALIZED	= PAnsiChar(6);
  CERT_STORE_PROV_FILENAME_A	= PAnsiChar(7);
  CERT_STORE_PROV_FILENAME_W	= PAnsiChar(8);

const
  X509_ASN_ENCODING	= $00000001;
  PKCS_7_ASN_ENCODING	= $00010000;

const
  CERT_STORE_NO_CRYPT_RELEASE_FLAG		= $00000001;
  CERT_STORE_SET_LOCALIZED_NAME_FLAG		= $00000002;
  CERT_STORE_DEFER_CLOSE_UNTIL_LAST_FREE_FLAG	= $00000004;
  CERT_STORE_DELETE_FLAG			= $00000010;
  CERT_STORE_UNSAFE_PHYSICAL_FLAG		= $00000020;
  CERT_STORE_SHARE_STORE_FLAG			= $00000040;
  CERT_STORE_SHARE_CONTEXT_FLAG			= $00000080;
  CERT_STORE_MANIFOLD_FLAG			= $00000100;
  CERT_STORE_ENUM_ARCHIVED_FLAG			= $00000200;
  CERT_STORE_UPDATE_KEYID_FLAG			= $00000400;
  CERT_STORE_BACKUP_RESTORE_FLAG		= $00000800;
  CERT_STORE_READONLY_FLAG			= $00008000;
  CERT_STORE_OPEN_EXISTING_FLAG			= $00004000;
  CERT_STORE_CREATE_NEW_FLAG			= $00002000;
  CERT_STORE_MAXIMUM_ALLOWED_FLAG		= $00001000;

var
  CertOpenStore: function(lpszStoreProvider: PAnsiChar; dwMsgAndCertEncodingType: DWORD;
			  hCryptProv: HCRYPTPROV; dwFlags: DWORD; pvPara: Pointer): HCERTSTORE; stdcall;

const
  CERT_CLOSE_STORE_FORCE_FLAG	= $00000001;
  CERT_CLOSE_STORE_CHECK_FLAG	= $00000002;

var
  CertCloseStore: function(hCertStore: HCERTSTORE; dwFlags: DWORD): BOOL; stdcall;

type
  CRYPT_ALGORITHM_IDENTIFIER = Record
    pszObjId: PAnsiChar;
    Parameters: CRYPT_OBJID_BLOB;
  end;
  PCRYPT_ALGORITHM_IDENTIFIER = ^CRYPT_ALGORITHM_IDENTIFIER;

type
  CRYPT_BIT_BLOB = Record
    cbData: DWORD;
    pbData: PBYTE;
    cUnusedBits: DWORD;
  end;
  PCRYPT_BIT_BLOB = ^CRYPT_BIT_BLOB;

type
  CERT_PUBLIC_KEY_INFO = Record
    Algorithm: CRYPT_ALGORITHM_IDENTIFIER;
    PublicKey: CRYPT_BIT_BLOB;
  end;
  PCERT_PUBLIC_KEY_INFO = ^CERT_PUBLIC_KEY_INFO;

type
  CERT_EXTENSION = Record
    pszObjId: PAnsiChar;
    fCritical: BOOL;
    Value: CRYPT_OBJID_BLOB;
  end;
  PCERT_EXTENSION = ^CERT_EXTENSION;

type
  CERT_INFO = Record
    dwVersion: DWORD;
    SerialNumber: CRYPT_INTEGER_BLOB;
    SignatureAlgorithm: CRYPT_ALGORITHM_IDENTIFIER;
    Issuer: CERT_NAME_BLOB;
    NotBefore: TFileTime;
    NotAfter: TFileTime;
    Subject: CERT_NAME_BLOB;
    SubjectPublicKeyInfo: CERT_PUBLIC_KEY_INFO;
    IssuerUniqueId: CRYPT_BIT_BLOB;
    SubjectUniqueId: CRYPT_BIT_BLOB;
    cExtension: DWORD;
    rgExtension: PCERT_EXTENSION;
  end;
  PCERT_INFO = ^CERT_INFO;

type
  CERT_CONTEXT = Record
    dwCertEncodingType: DWORD;
    pbCertEncoded: PBYTE;
    cbCertEncoded: DWORD;
    pCertInfo: PCERT_INFO;
    hCertStore: HCERTSTORE;
  end;
  PCCERT_CONTEXT = ^CERT_CONTEXT;
  PPCCERT_CONTEXT = ^PCCERT_CONTEXT;

const
  CERT_STORE_ADD_NEW					= 1;
  CERT_STORE_ADD_USE_EXISTING				= 2;
  CERT_STORE_ADD_REPLACE_EXISTING			= 3;
  CERT_STORE_ADD_ALWAYS					= 4;
  CERT_STORE_ADD_REPLACE_EXISTING_INHERIT_PROPERTIES	= 5;
  CERT_STORE_ADD_NEWER					= 6;
  CERT_STORE_ADD_NEWER_INHERIT_PROPERTIES		= 7;

var
  CertAddCertificateLinkToStore: function(hCertStore: HCERTSTORE; pCertContext: PCCERT_CONTEXT;
					  dwAddDisposition: DWORD; ppStoreContext: PPCCERT_CONTEXT): BOOL; stdcall;
  CertEnumCertificatesInStore: function(hCertStore: HCERTSTORE; pPrevCertContext: PCCERT_CONTEXT): PCCERT_CONTEXT; stdcall;
  CertDeleteCertificateFromStore: function(pCertContext: PCCERT_CONTEXT): BOOL; stdcall;
  CertDuplicateCertificateContext: function(pCertContext: PCCERT_CONTEXT): PCCERT_CONTEXT; stdcall;
  CertFreeCertificateContext: function(pCertContext: PCCERT_CONTEXT): BOOL; stdcall;

const
  CERT_KEY_PROV_HANDLE_PROP_ID      = 1;
  CERT_KEY_PROV_INFO_PROP_ID        = 2;
  CERT_SHA1_HASH_PROP_ID            = 3;
  CERT_MD5_HASH_PROP_ID             = 4;
  CERT_KEY_CONTEXT_PROP_ID          = 5;
  CERT_KEY_SPEC_PROP_ID             = 6;

var
  CertGetCertificateContextProperty: function(pCertContext: PCCERT_CONTEXT; dwPropId: DWORD;
					      pvData: Pointer; var pcbData: DWORD): BOOL; stdcall;

const
  CERT_NAME_EMAIL_TYPE			= 1;
  CERT_NAME_RDN_TYPE			= 2;
  CERT_NAME_ATTR_TYPE			= 3;
  CERT_NAME_SIMPLE_DISPLAY_TYPE		= 4;
  CERT_NAME_FRIENDLY_DISPLAY_TYPE	= 5;
  CERT_NAME_DNS_TYPE			= 6;
  CERT_NAME_URL_TYPE			= 7;
  CERT_NAME_UPN_TYPE			= 8;

const
  CERT_NAME_ISSUER_FLAG			= $00000001;
  CERT_NAME_DISABLE_IE4_UTF8_FLAG	= $00010000;

var
  CertGetNameString: function(pCertContext: PCCERT_CONTEXT; dwType: DWORD; dwFlags: DWORD; pvTypePara: Pointer;
			      pszNameString: PAnsiChar; cchNameString: DWORD): DWORD; stdcall;

const
  CRYPT_FIND_USER_KEYSET_FLAG		= $00000001;
  CRYPT_FIND_MACHINE_KEYSET_FLAG	= $00000002;
  CRYPT_FIND_SILENT_KEYSET_FLAG		= $00000040;

var
  CryptFindCertificateKeyProvInfo: function(pCert: PCCERT_CONTEXT; dwFlags: DWORD; pvReserved: Pointer): BOOL; stdcall;

const
  REPORT_NO_PRIVATE_KEY			= $0001;
  REPORT_NOT_ABLE_TO_EXPORT_PRIVATE_KEY	= $0002;
  EXPORT_PRIVATE_KEYS			= $0004;

var
  PFXExportCertStoreEx: function(hStore: HCERTSTORE; var pPFX: CRYPT_DATA_BLOB;
				 szPassword: PWideChar; pvReserved: Pointer;
				 dwFlags: DWORD): BOOL; stdcall;
  PFXExportCertStore: function(hStore: HCERTSTORE; var pPFX: CRYPT_DATA_BLOB;
			       szPassword: PWideChar; dwFlags: DWORD): BOOL; stdcall;

const
  CRYPTUI_HIDE_HIERARCHYPAGE			= $00000001;
  CRYPTUI_HIDE_DETAILPAGE			= $00000002;
  CRYPTUI_DISABLE_EDITPROPERTIES		= $00000004;
  CRYPTUI_ENABLE_EDITPROPERTIES			= $00000008;
  CRYPTUI_DISABLE_ADDTOSTORE			= $00000010;
  CRYPTUI_ENABLE_ADDTOSTORE			= $00000020;
  CRYPTUI_ACCEPT_DECLINE_STYLE			= $00000040;
  CRYPTUI_IGNORE_UNTRUSTED_ROOT			= $00000080;
  CRYPTUI_DONT_OPEN_STORES			= $00000100;
  CRYPTUI_ONLY_OPEN_ROOT_STORE			= $00000200;
  CRYPTUI_WARN_UNTRUSTED_ROOT			= $00000400; // For use with viewing of certificates on remote
							     // machines only.  If this flag is used rghStores[0]
							     // must be the handle of the root store on the remote machine.
  CRYPTUI_ENABLE_REVOCATION_CHECKING		= $00000800; // This flag is only valid if pCryptProviderData/hWVTStateData
							     // is not passed in.
  CRYPTUI_WARN_REMOTE_TRUST			= $00001000;
  CRYPTUI_DISABLE_EXPORT			= $00002000; // If this flag is set, then the "Copy to file" button will be
							     // disabled on the Details page.

  // Revocation flags is only valid if pCryptProviderData/hWVTStateData is not passed in.
  CRYPTUI_ENABLE_REVOCATION_CHECK_END_CERT	= $00004000;
  CRYPTUI_ENABLE_REVOCATION_CHECK_CHAIN		= $00008000;


// declarations translated from CryptUIApi.h

type
  // simplified declaration as we don't use most values
  CRYPTUI_VIEWCERTIFICATE_STRUCT = Record
    dwSize: DWORD;
    hwndParent: HWND;				// OPTIONAL
    dwFlags: DWORD;				// OPTIONAL
    szTitle: PAnsiChar;				// OPTIONAL
    pCertContext: PCCERT_CONTEXT;
    rgszPurposes: PAnsiChar;			// OPTIONAL
    cPurposes: DWORD;				// OPTIONAL
    WinTrust: Record case Boolean of
      TRUE: (pCryptProviderData: Pointer);	// OPTIONAL
      FALSE: (hWVTStateData: THandle);		// OPTIONAL
    end;
    fpCryptProviderDataTrustedUsage: BOOL;	// OPTIONAL
    idxSigner: DWORD;				// OPTIONAL
    idxCert: DWORD;				// OPTIONAL
    fCounterSigner: BOOL;			// OPTIONAL
    idxCounterSigner: DWORD;			// OPTIONAL
    cStores: DWORD;				// OPTIONAL
    rghStores: HCERTSTORE;			// OPTIONAL
    cPropSheetPages: DWORD;			// OPTIONAL
    rgPropSheetPages: Pointer;			// OPTIONAL
    nStartPage: DWORD;
  end;
  PCCRYPTUI_VIEWCERTIFICATE_STRUCT = ^CRYPTUI_VIEWCERTIFICATE_STRUCT;

var
  CryptUIDlgViewCertificate: function(const pCertViewInfo: CRYPTUI_VIEWCERTIFICATE_STRUCT;
				      var pfPropertiesChanged: BOOL): BOOL; stdcall;

const
  CRYPTUI_SELECT_ISSUEDTO_COLUMN	= $000000001;
  CRYPTUI_SELECT_ISSUEDBY_COLUMN	= $000000002;
  CRYPTUI_SELECT_INTENDEDUSE_COLUMN	= $000000004;
  CRYPTUI_SELECT_FRIENDLYNAME_COLUMN	= $000000008;
  CRYPTUI_SELECT_LOCATION_COLUMN	= $000000010;
  CRYPTUI_SELECT_EXPIRATION_COLUMN	= $000000020;

var
  CryptUIDlgSelectCertificateFromStore: function(hCertStore: HCERTSTORE; hwnd: HWND; pwszTitle: PWideChar;
						 pwszDisplayString: PWideChar; dwDontUseColumn: DWORD;
						 dwFlags: DWORD; pvReserved: Pointer): PCCERT_CONTEXT; stdcall;

function SystemHasPrivateCerts: Longbool;
var
  hcsSystem: HCERTSTORE;
  pccEnum: PCCERT_CONTEXT;
  dwSize: DWORD;
begin
  Result := FALSE;
  if Assigned(PFXExportCertStoreEx) or Assigned(PFXExportCertStore) then begin
    hcsSystem := CertOpenSystemStore(0, MY);
    if hcsSystem = NIL then Exit;
    try
      pccEnum := NIL;
      repeat
	pccEnum := CertEnumCertificatesInStore(hcsSystem, pccEnum);
	if (pccEnum <> NIL) and CertGetCertificateContextProperty(pccEnum, CERT_KEY_SPEC_PROP_ID, NIL, dwSize) then begin
	  CertFreeCertificateContext(pccEnum);
	  Result := TRUE;
	  Break;
	end;
      until pccEnum = NIL;
    finally
      CertCloseStore(hcsSystem, CERT_CLOSE_STORE_FORCE_FLAG);
    end;
  end;
end;

function ShowCertificateDialog(const CertData, Title: RawByteString; FileName: Longbool; WinHandle: THandle): Integer;
const
  OwnerWnd: HWND = 0;
  CallWndHook: HHOOK = 0;

  function CallWndProc(nCode: Integer; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
  begin
    if nCode = HC_ACTION then with PCWPStruct(lParam)^ do begin
      Result := 0;
      if GetWindowLongA(hWnd, GWL_STYLE) and WS_CHILD = 0 then begin
	case Message of
	  WM_CREATE: if IsStandardDialog(hWnd) then PCreateStructA(lParam)^.hwndParent := OwnerWnd;
	  WM_DESTROY: if IsStandardDialog(hWnd) then OwnerWnd := GetWindow(hWnd, GW_OWNER);
	  WM_SHOWWINDOW: if wParam <> 0 then begin
	    if IsStandardDialog(hWnd) then begin
	      SetWindowPos(hWnd, HWND_TOP or HWND_NOTOPMOST, 0, 0, 0, 0, SWP_DRAWFRAME or SWP_NOMOVE or SWP_NOSIZE);
	      if OwnerWnd = 0 then OwnerWnd := GetDesktopWindow;
	      CenterDialog(hWnd, OwnerWnd);
	      OwnerWnd := hWnd;
	    end;
	  end;
	end;
      end;
    end
    else Result := CallNextHookEx(CallWndHook, nCode, wParam, lParam);
  end;

var
  cdbSource: CRYPT_DATA_BLOB;
  hcsSource: HCERTSTORE;
  pccEnum, pccView: PCCERT_CONTEXT;
  ftNotBefore: TFileTime;
  cvsData: CRYPTUI_VIEWCERTIFICATE_STRUCT;
  bChanged: BOOL;
begin
  if FileName then
    Result := kPGPError_FileNotFound
  else Result := kPGPError_ItemNotFound;
  if CertData = E then Exit;
  Result := kPGPError_FeatureNotAvailable;
  if not Assigned(CryptUIDlgViewCertificate) then Exit;
  Result := kPGPError_X509InvalidCertificateFormat;
  if FileName then begin
    hcsSource := CertOpenStore(CERT_STORE_PROV_FILENAME_A, X509_ASN_ENCODING or PKCS_7_ASN_ENCODING, 0,
			       CERT_STORE_OPEN_EXISTING_FLAG or CERT_STORE_READONLY_FLAG, Pointer(CertData));
  end
  else begin
    with cdbSource do begin
      cbData := Length(CertData);
      pbData := Pointer(CertData);
    end;
    hcsSource := CertOpenStore(CERT_STORE_PROV_PKCS7, X509_ASN_ENCODING or PKCS_7_ASN_ENCODING, 0,
			       CERT_STORE_CREATE_NEW_FLAG, @cdbSource);
  end;
  if hcsSource = NIL then Exit;
  pccEnum := NIL;
  pccView := NIL;
  try
    Result := kPGPError_X509NeededCertNotAvailable;
    FillChar(ftNotBefore, SizeOf(TFileTime), 0);
    while GetLastError = 0 do begin
      pccEnum := CertEnumCertificatesInStore(hcsSource, pccEnum);
      if (pccEnum <> NIL) and (CompareFileTime(pccEnum.pCertInfo^.NotBefore, ftNotBefore) > 0) then begin
	if pccView <> NIL then CertFreeCertificateContext(pccView);
	pccView := CertDuplicateCertificateContext(pccEnum);
	ftNotBefore := pccEnum.pCertInfo^.NotBefore;
      end;
    end;
    if pccView = NIL then Exit;
    Result := kPGPError_UnknownError;
    FillChar(cvsData, SizeOf(CRYPTUI_VIEWCERTIFICATE_STRUCT), 0);
    with cvsData do begin
      dwSize := SizeOf(CRYPTUI_VIEWCERTIFICATE_STRUCT);
      hwndParent := WinHandle;
      dwFlags := CRYPTUI_DISABLE_EDITPROPERTIES or CRYPTUI_DISABLE_ADDTOSTORE or
		 CRYPTUI_IGNORE_UNTRUSTED_ROOT or CRYPTUI_ENABLE_REVOCATION_CHECKING;
      if Title <> E then szTitle := PAnsiChar(Title);
      pCertContext := pccView;
      cStores := 1;
      rghStores := @hcsSource;
    end;
    OwnerWnd := WinHandle;
    CallWndHook := SetWindowsHookExA(WH_CALLWNDPROC, @CallWndProc, 0, GetCurrentThreadID);
    try
      if CryptUIDlgViewCertificate(cvsData, bChanged) or (GetLastError = ERROR_CANCELLED) then Result := 0;
    finally
      UnhookWindowsHookEx(CallWndHook);
    end;
    ProcessMessages;
  finally
    CertFreeCertificateContext(pccView);
    CertFreeCertificateContext(pccEnum);
    CertCloseStore(hcsSource, CERT_CLOSE_STORE_FORCE_FLAG);
  end;
end;

function GetCertNameString(pCertContext: PCCERT_CONTEXT; Issuer: Longbool): RawByteString;
var
  dwFlags, dwLen: DWORD;
begin
  dwFlags := 0;
  if Issuer then dwFlags := CERT_NAME_ISSUER_FLAG;
  dwLen := CertGetNameString(pCertContext, CERT_NAME_SIMPLE_DISPLAY_TYPE, dwFlags, NIL, NIL, 0);
  if dwLen > 1 then begin
    SetString(Result, NIL, pred(dwLen));
    CertGetNameString(pCertContext, CERT_NAME_SIMPLE_DISPLAY_TYPE, dwFlags, NIL, Pointer(Result), dwLen);
  end
  else Result := E;
end;

function ImportPrivateSystemCerts(const CertKeyPrompt, CertPassPrompt: RawByteString; TempPassphrase: PAnsiChar;
				  var CertPropsList: TX509CertPropsList; WinHandle: THandle): Integer;
const
  OwnerWnd: HWND = 0;
  CallWndHook: HHOOK = 0;
  DialogResult: Integer = 0;
  CertTitle: RawByteString = E;
  PassPrompt: RawByteString = E;
  CertPassphrase: PAnsiChar = NIL;

  function PostButtonCommand(Dialog: THandle; ButtonID: Integer): Longbool;
  var
    hButton: THandle;
  begin
    Result := FALSE;
    DialogResult := ButtonID;
    hButton := GetDlgItem(Dialog, ButtonID);
    if not IsWindowEnabled(hButton) then Exit;
    Result := PostMessageA(Dialog, WM_COMMAND, BN_CLICKED shl 16 + ButtonID, GetDlgItem(Dialog, ButtonID));
  end;

  procedure PassphraseCallback(const Passphrase: PAnsiChar; const UserValue);
  begin
    CertPassphrase := PGPNewSecureData(PGPGetDefaultMemoryMgr, succ(StrLen(Passphrase)), kPGPMemoryMgrFlags_Clear);
    if CertPassphrase <> NIL then StrCopy(CertPassphrase, Passphrase);
    SetWindowTextA(THandle(UserValue), Passphrase);
  end;

  function CallWndProc(nCode: Integer; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
  var
    hEdit: THandle;
    sWarning: RawByteString;
  begin
    if nCode = HC_ACTION then with PCWPStruct(lParam)^ do begin
      Result := 0;
      if GetWindowLongA(hWnd, GWL_STYLE) and WS_CHILD = 0 then begin
	case Message of
	  WM_CREATE: if IsStandardDialog(hWnd) then PCreateStructA(lParam)^.hwndParent := OwnerWnd;
	  WM_SHOWWINDOW: if wParam <> 0 then begin
	    if IsStandardDialog(hWnd) then begin
	      if GetWindowText(hWnd) <> CertTitle then begin
		if OwnerWnd = 0 then OwnerWnd := GetDesktopWindow;
		SetWindowPos(hWnd, 0, 0, 0, 0, 0, SWP_HIDEWINDOW);
		hEdit := GetDlgItem(hWnd, $3FC);
		if hEdit <> 0 then begin
		  if IsWindowEnabled(hEdit) then begin
		    if SimplePassphraseDialog(CertTitle, PassPrompt, OwnerWnd, @PassphraseCallback, hEdit) = 0 then
		      PostButtonCommand(hWnd, ID_OK)
		    else PostButtonCommand(hWnd, ID_CANCEL);
		  end
		  else begin
		    sWarning := GetWindowText(GetDlgItem(hWnd, $400));
		    PostButtonCommand(hWnd, MessageBoxA(OwnerWnd, Pointer(sWarning), Pointer(CertTitle), MB_ICONWARNING or MB_OKCANCEL));
		  end;
		end
		else PostButtonCommand(hWnd, ID_CANCEL);
	      end
	      else CenterDialog(hWnd, OwnerWnd);
	    end;
	  end;
	end;
      end;
    end
    else Result := CallNextHookEx(CallWndHook, nCode, wParam, lParam);
  end;

var
  hcsSystem, hcsPrivate: HCERTSTORE;
  wcTempPassphrase: PWideChar;
  pccEnum: PCCERT_CONTEXT;
  dwStrLen, dwExportFlags: DWORD;
  cdbPFX: CRYPT_DATA_BLOB;
  sPFXData: RawByteString;
begin
  if Assigned(PFXExportCertStoreEx) or Assigned(PFXExportCertStore) then begin
    Result := kPGPError_X509NeededCertNotAvailable;
    hcsSystem := CertOpenSystemStore(0, MY);
    if hcsSystem = NIL then Exit;
    wcTempPassphrase := NIL;
    pccEnum := NIL;
    try
      Result := kPGPError_OutOfMemory;
      dwStrLen := succ(StrLen(TempPassphrase));
      wcTempPassphrase := PGPNewSecureData(PGPGetDefaultMemoryMgr, dwStrLen shl 1, kPGPMemoryMgrFlags_Clear);
      if wcTempPassphrase = NIL then Exit;
      Result := kPGPError_NoUnicodeEquivalent;
      if MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, TempPassphrase, -1, wcTempPassphrase, dwStrLen) = 0 then Exit;
      Result := kPGPError_X509NeededCertNotAvailable;
      pccEnum := CertEnumCertificatesInStore(hcsSystem, NIL);
      if pccEnum <> NIL then begin
	OwnerWnd := WinHandle;
	dwExportFlags := REPORT_NO_PRIVATE_KEY or REPORT_NOT_ABLE_TO_EXPORT_PRIVATE_KEY or EXPORT_PRIVATE_KEYS;
	repeat
	  hcsPrivate := CertOpenStore(CERT_STORE_PROV_MEMORY, 0, 0, CERT_STORE_CREATE_NEW_FLAG, NIL);
	  if hcsPrivate <> NIL then begin
	    CertPassphrase := NIL;
	    try
	      if CryptFindCertificateKeyProvInfo(pccEnum, CRYPT_FIND_USER_KEYSET_FLAG, NIL)
	      and CertAddCertificateLinkToStore(hcsPrivate, pccEnum, CERT_STORE_ADD_ALWAYS, NIL) then begin
		with cdbPFX do begin
		  cbData := 0;
		  pbData := NIL;
		end;
		if Assigned(PFXExportCertStoreEx) and PFXExportCertStoreEx(hcsPrivate, cdbPFX, NIL, NIL, dwExportFlags)
		or PFXExportCertStore(hcsPrivate, cdbPFX, NIL, dwExportFlags) then begin
		  SetString(sPFXData, NIL, cdbPFX.cbData);
		  cdbPFX.pbData := Pointer(sPFXData);
		  CallWndHook := SetWindowsHookExA(WH_CALLWNDPROC, @CallWndProc, 0, GetCurrentThreadID);
		  try
		    DialogResult := 0;
		    CertTitle := GetCertNameString(pccEnum, FALSE) + ' / ' + GetCertNameString(pccEnum, TRUE);
		    PassPrompt := CertPassPrompt;
		    if Assigned(PFXExportCertStoreEx) then
		      PFXExportCertStoreEx(hcsPrivate, cdbPFX, wcTempPassphrase, NIL, dwExportFlags)
		    else PFXExportCertStore(hcsPrivate, cdbPFX, wcTempPassphrase, dwExportFlags);
		  finally
		    Result := GetLastError;
		    UnhookWindowsHookEx(CallWndHook);
		  end;
		  ProcessMessages;
		  case DialogResult of
		    ID_OK: begin
		      Result := kPGPError_NoErr;
		      SetCursor(LoadCursor(0, IDC_WAIT));
		      Delete(sPFXData, succ(cdbPFX.cbData), MAXINT);
		      Result := ImportX509Certificate(CertKeyPrompt, E, sPFXData,
						      TempPassphrase, CertPassphrase,
						      FALSE, CertPropsList, WinHandle);
		    end;
		    ID_CANCEL: Result := kPGPError_UserAbort;
		  else
		    Result := kPGPError_X509NeededCertNotAvailable;
		  end;
		end;
	      end;
	    finally
	      CertCloseStore(hcsPrivate, 0);
	      PGPFreeData(CertPassphrase);
	    end;
	  end;
	  pccEnum := CertEnumCertificatesInStore(hcsSystem, pccEnum);
	until pccEnum = NIL;
      end;
    finally
      PGPFreeData(wcTempPassphrase);
      CertFreeCertificateContext(pccEnum);
      CertCloseStore(hcsSystem, CERT_CLOSE_STORE_FORCE_FLAG);
    end;
  end
  else Result := kPGPError_FeatureNotAvailable;
end;

procedure LoadCryptLibs;
const
  CRYPT32 = 'Crypt32.dll';
  CRYPTUI = 'CryptUI.dll';
var
  hCrypt32: HMODULE;
  hCryptUI: HMODULE;
begin
  hCrypt32 := LoadLibrary(CRYPT32);
  hCryptUI := LoadLibrary(CRYPTUI);
  // available since Win95 + IE 3.02
  CertOpenSystemStore := GetProcAddress(hCrypt32, 'CertOpenSystemStoreA');
  CertOpenStore := GetProcAddress(hCrypt32, 'CertOpenStore');
  CertCloseStore := GetProcAddress(hCrypt32, 'CertCloseStore');
  CertAddCertificateLinkToStore := GetProcAddress(hCrypt32, 'CertAddCertificateLinkToStore');
  CertEnumCertificatesInStore := GetProcAddress(hCrypt32, 'CertEnumCertificatesInStore');
  CertDeleteCertificateFromStore := GetProcAddress(hCrypt32, 'CertDeleteCertificateFromStore');
  CertDuplicateCertificateContext := GetProcAddress(hCrypt32, 'CertDuplicateCertificateContext');
  CertFreeCertificateContext := GetProcAddress(hCrypt32, 'CertFreeCertificateContext');
  CertGetCertificateContextProperty := GetProcAddress(hCrypt32, 'CertGetCertificateContextProperty');
  CertGetNameString := GetProcAddress(hCrypt32, 'CertGetNameStringA');
  // available since W2K Pro and WXP
  CryptFindCertificateKeyProvInfo := GetProcAddress(hCrypt32, 'CryptFindCertificateKeyProvInfo');
  PFXExportCertStore := GetProcAddress(hCrypt32, 'PFXExportCertStore');
  PFXExportCertStoreEx := GetProcAddress(hCrypt32, 'PFXExportCertStoreEx');
  CryptUIDlgViewCertificate := GetProcAddress(hCryptUI, 'CryptUIDlgViewCertificateA');
  CryptUIDlgSelectCertificateFromStore := GetProcAddress(hCryptUI, 'CryptUIDlgSelectCertificateFromStore');
end;


begin
  LoadCryptLibs;
end.

