{------------------------------------------------------------------------------}
{                                                                              }
{       This unit is based on Steve Heller's spgpPreferences.pas from his      }
{         SPGP sources available from http://www.oz.net/~srheller/spgp/        }
{                                                                              }
{                Portions created by Michael in der Wiesche are                }
{              Copyright (C) 2001-2010 by Michael in der Wiesche               }
{                                                                              }
{------------------------------------------------------------------------------}
{$J+,Z4}
unit KeyFuncs;

interface

uses
  Windows, ShellApi, Classes, SysUtils, TimeUtils, KeyPropTypes, X509Types, ASN1Tools, PGPStrTools, PGPStringList, UTF8,
  pgpBase, pgpErrors, pgpPubTypes, pgpUtilities, pgpOptionList, pgpMemoryMgr, pgpGroups, pgpKeys, pgpCl;

type
  TKeyRings = Class
  private
    FPubringFile: RawByteString;
    FSecringFile: RawByteString;
    InitCount: Longint;
    Recursion: Longbool;
    RingKeyDB: pPGPKeyDB;
    RingKeySet: pPGPKeySet;
    RingContext: pPGPContext;
    function  GetPubringFile: RawByteString;
    function  GetSecringFile: RawByteString;
    procedure SetPubringFile(const Value: RawByteString);
    procedure SetSecringFile(const Value: RawByteString);
  public
    GroupsFile: RawByteString;
    destructor Destroy; override;
    function  UpdateKeyRings: PGPError;
    function  SetKeyRings(const Pubring, Secring: RawByteString): PGPError;
    function  InitKeyRings(var Context: pPGPContext; var KeySetMain: pPGPKeySet): PGPError;
    procedure FreeKeyRings;
    property  PubringFile: RawByteString read GetPubringFile write SetPubringFile;
    property  SecringFile: RawByteString read GetSecringFile write SetSecringFile;
  end;

{ "public" functions }
// returns hex ID length
function IsHexID(const ToCheck: RawByteString): Longint;
function ShortHexID(const HexID: RawByteString): RawByteString;
function KeyOnRing(const HexKeyID: RawByteString): Integer;
function KeyRemove(KeyIDCommaText: PAnsiChar): Integer;
function KeyEnable(HexKeyID: PAnsiChar): Integer;
function KeyDisable(HexKeyID: PAnsiChar): Integer;
function KeyRevoke(HexKeyID, Passphrase: PAnsiChar): Integer;
function ChangePassphrase(HexKeyID, OldPassphrase, NewPassphrase: PAnsiChar): Integer;
function GetUserIDCertOwner(const HexKeyID, UserID: RawByteString): RawByteString;
function GetSimpleKeyProp(const HexKeyID: RawByteString; PropertyFlag: DWord): Variant;
// properties, key & user IDs for all keys matching the specified KeyIDCommaText
// if KeyIDCommaText = '' then all keys will be checked
// - returns number of keys found or error
function FindKeyProps(const KeyIDCommaText: RawByteString;
		      var KeyPropsList: TKeyPropsList;
		      PropertyFlags, FilterFlags: DWord;
		      KeyOrder: TPGPKeyOrdering): Integer;
{ "private" functions }
function SystemTimeToUnixTimeNum: Integer;
function UnixTimeToLocalTimeStr(UnixSeconds: Integer): RawByteString;
function GetCertPropIASN(Cert: pPGPKeyDBObj): RawByteString;
function GetCertPropIssuerName(Cert: pPGPKeyDBObj): RawByteString;
function GetCertPropIssuerSerialNumber(Cert: pPGPKeyDBObj): RawByteString;
function GetCertPropOwnerKeyID(Cert: pPGPKeyDBObj): RawByteString;
function GetCertPropOwnerUserID(Cert: pPGPKeyDBObj): RawByteString;
function GetCertPropOwnerEmail(Cert: pPGPKeyDBObj): RawByteString;
function GetCertPropOwnerName(Cert: pPGPKeyDBObj): RawByteString;
function GetCertPropOwnerSerialNumber(Cert: pPGPKeyDBObj): RawByteString;
function GetCertPropCertCreationTimeStr(Cert: pPGPKeyDBObj): RawByteString;
function GetCertPropCertExpirationTimeStr(Cert: pPGPKeyDBObj): RawByteString;
function GetCertPropCertCreationTime(Cert: pPGPKeyDBObj): Integer;
function GetCertPropCertExpirationTime(Cert: pPGPKeyDBObj): Integer;
function GetCertKeyUsage(Cert: pPGPKeyDBObj): Word;
function GetCertPropCanEncrypt(Cert: pPGPKeyDBObj; KeyUsage: Word): Longbool;
function GetCertPropCanSign(Cert: pPGPKeyDBObj; KeyUsage: Word): Longbool;
function GetCertPropIsVerified(Cert: pPGPKeyDBObj): Longbool;
function GetCertPropIsCorrupt(Cert: pPGPKeyDBObj): Longbool;
function GetCertPropIsRevoked(Cert: pPGPKeyDBObj): Longbool;
function GetCertPropIsExpired(Cert: pPGPKeyDBObj): Longbool;
function GetCertPropIsRoot(Cert: pPGPKeyDBObj): Longbool;
function GetCertPropIsValid(Cert: pPGPKeyDBObj): Longbool;
function GetSubKeyPropKeyID(SubKey: pPGPKey): RawByteString;
function GetSubKeyPropIsRevoked(SubKey: pPGPKey): Longbool;
function GetSubKeyPropIsExpired(SubKey: pPGPKey): Longbool;
function GetKeyPropKeyID(Key: pPGPKey): RawByteString;
function GetKeyPropUserID(Key: pPGPKey): RawByteString;
function GetKeyPropFingerprint(Key: pPGPKey): RawByteString;
function GetKeyPropKeyCreationTimeStr(Key: pPGPKey): RawByteString;
function GetKeyPropKeyExpirationTimeStr(Key: pPGPKey): RawByteString;
function GetKeyPropKeyBits(Key: pPGPKey): RawByteString;
function GetKeyPropX509Cert(Key: pPGPKey): TX509Type;
function GetKeyPropKeyAlg(Key: pPGPKey): TKeyAlgorithm;
function GetKeyPropHashAlg(Key: pPGPKey): THashAlgorithm;
function GetKeyPropLegacy(Key: pPGPKey): Longbool;
function GetKeyPropTrust(Key: pPGPKey): TTrustLevel;
function GetKeyPropValidity(Key: pPGPKey): TValidityLevel;
function GetKeyPropKeyCreationTime(Key: pPGPKey): Integer;
function GetKeyPropKeyExpirationTime(Key: pPGPKey): Integer;
function GetKeyPropIsCorrupt(Key: pPGPKey): Longbool;
function GetKeyPropIsSecret(Key: pPGPKey): Longbool;
function GetKeyPropIsAxiomatic(Key: pPGPKey): Longbool;
function GetKeyPropIsSigningKey(Key: pPGPKey): Longbool;
function GetKeyPropIsRevoked(Key: pPGPKey): Longbool;
function GetKeyPropIsDisabled(Key: pPGPKey): Longbool;
function GetKeyPropIsExpired(Key: pPGPKey): Longbool;
function GetKeyPropIsSecretShared(Key: pPGPKey): Longbool;
function GetKeyPropCanEncrypt(Key: pPGPKey): Longbool;
function GetKeyPropCanDecrypt(Key: pPGPKey): Longbool;
function GetKeyPropCanSign(Key: pPGPKey): Longbool;
function GetKeyPropCanVerify(Key: pPGPKey): Longbool;
function GetKeyPropHasCertificate(Key: pPGPKey): Longbool;
function GetKeyPropHasRevoker(KeySet: pPGPKeySet; Key: pPGPKey; const RevKeyIDList: TRevKeyIDs): Longbool;
function GetKeyPropHasARR(KeySet: pPGPKeySet; Key: pPGPKey; const ADKeyIDList: TADKeyIDs): TADKType;
function GetKeyPropHasSubKey(KeySet: pPGPKeySet; Key: pPGPKey): Longbool;
function GetKeyFromKeySet(KeySet: pPGPKeySet; var Key: pPGPKey): PGPError;
function HasKeyUserID(Key: pPGPKey; const UserID: RawByteString): Longbool;
function GetKeyByHexID(KeySet: pPGPKeySet; const HexID: RawByteString; var Key: pPGPKey): PGPError;
function GetKeyFromNewSet(Key: pPGPKey; NewSet: pPGPKeySet; var NewKey: pPGPKey): PGPError;
function RemoveKeyFromKeySet(Key: pPGPKey; TargetKeySet: pPGPKeySet): PGPError;
function AddKeyToKeySet(Key: pPGPKey; TargetKeySet: pPGPKeySet): PGPError;
function GetKeyIterAnyOrdering(KeySet: pPGPKeySet; var KeyList: pPGPKeyList; var KeyIter: pPGPKeyIter): PGPError;
function GetSingleKeyDBObjIter(KeyDBObj: pPGPKeyDBObj; var ObjSet: pPGPKeySet;
			       var ObjList: pPGPKeyList; var ObjIter: pPGPKeyIter): PGPError;
function GetKeyFilter(Context: pPGPContext; FilterFlags: DWord; var KeyFilter: pPGPFilter): PGPError;
function GetHexIDFilter(Context: pPGPContext; const HexID: RawByteString;
			IncludeSubKeys: Longbool; var HexIDFilter: pPGPFilter): PGPError;
function GetUserIDFilter(Context: pPGPContext; const UserID: RawByteString;
			 IncludeUTF8: Longbool; var UserIDFilter: pPGPFilter): PGPError;
function GetKeyExcludeFilter(Context: pPGPContext; Key: pPGPKeyDBObj): pPGPFilter;
function GetKeySetWithoutSigner(Context: pPGPContext; const KeySetToFilter: pPGPKeySet; SignKey: pPGPKeyDBObj;
				var KeySetFiltered: pPGPKeySet): PGPError;
function GetKeyFilterByAnyID(Context: pPGPContext; const AnyID: RawByteString;
			     IncludeSubKeys, IncludeUTF8: Longbool;
			     var KeyFilter: pPGPFilter): PGPError;
function GetKeySetByAnyIDs(Context: pPGPContext;
			   KeySetMain: pPGPKeySet;
			   const KeyIDCommaText: RawByteString;
			   var KeySetFound: pPGPKeySet): PGPError;
function GetExclusiveKeySet(var KeySetToCheck: pPGPKeySet; KeySetMain: pPGPKeySet;
			    Context: pPGPContext; IgnoreKnownFlag: Integer): PGPError;
function GetCertKeySetFromKeySet(const KeySet: pPGPKeySet; EncryptOnly: Longbool; var CertKeySet: pPGPKeySet): PGPError;
function GetCertProps(Cert: pPGPKeyDBObj; var CertPropsRec: TX509CertPropsRec): PGPError;
function GetKeyProps(KeySet: pPGPKeySet; Key: pPGPKey; Props, Filters: DWord; var KeyPropsRec: TKeyPropsRec): PGPError;
// - returns number of keys found or error
function GetKeySetProps(Context: pPGPContext; KeySet: pPGPKeySet;
			var KeyPropsList: TKeyPropsList;
			PropertyFlags, FilterFlags: DWord;
			KeyOrder: TPGPKeyOrdering): Integer;
// - returns number of keys found or error
function AddKeysToKeyRing(Context: pPGPContext; KeySetMain, KeysToImport: pPGPKeySet;
			  var KeyPropsList: TKeyPropsList; PropertyFlags: DWord): Integer;
// - returns number of keys found or error
function GetHexIDByAnyID(Context: pPGPContext; KeySetMain: pPGPKeySet; const AnyID: RawByteString; var HexID: RawByteString): Integer;
// - returns number of user IDs found or error
function CountUserIDs(KeySet: pPGPKeySet): Integer;
function GetMatchingKey(Context: pPGPContext; KeySet: pPGPKeySet; Passphrase: PAnsiChar): pPGPKey;
function GetUtf8Passphrase(const Passphrase: PAnsiChar; var Utf8Passphrase: PAnsiChar): PGPError;
function PassphraseIsValid(Context: pPGPContext; KeySetMain: pPGPKeySet; AnyID, Passphrase: PAnsiChar): Longbool;
function GetCachedPassphrase(Context: pPGPContext; var Key: pPGPKey; KeySet: pPGPKeySet; var Passphrase: PAnsiChar): Longbool;
function CachePassphrase(Context: pPGPContext; Key: pPGPKey; TimeInterval: PGPTime;
			 ShareCache: Longbool; var Passphrase: PAnsiChar): Longbool;
function IsPassphraseCacheEmpty(Context: pPGPContext): Longbool;
function PurgePassphraseCache(Context: pPGPContext): Longbool;

const
  KeyRings: TKeyRings = NIL;


implementation

uses
  {$IFDEF UNICODE}AnsiStrings, {$ENDIF} PrefFuncs;

function IsHexID(const ToCheck: RawByteString): Longint; register; assembler;
asm	// EAX = @ToCheck
  OR	EAX,EAX
  JE	@EXIT
  MOV	ECX,[EAX - 4]
  PUSH	ECX
  CMP	ECX,LongHexIDLen
  JE	@START
  CMP	ECX,ShortHexIDLen
  JNE	@FALSE
  @START:
  MOV	DX,[EAX]
  AND	DH,0DFh
  CMP	DX,'X0'
  JNE	@FALSE
  SUB	ECX,2
  ADD	EAX,2
  @LOOP:
  DEC	ECX
  JS	@TRUE
  MOV	DL,[EAX + ECX]
  CMP	DL,'0'
  JB	@FALSE
  CMP	DL,'9'
  JBE	@LOOP
  AND	DL,0DFh
  CMP	DL,'A'
  JB	@FALSE
  CMP	DL,'F'
  JBE	@LOOP
  @FALSE:
  XOR	EAX,EAX
  POP	ECX
  RET
  @TRUE:
  POP	EAX
  @EXIT:
end;

function ShortHexID(const HexID: RawByteString): RawByteString;
begin
  Result := E;
  case IsHexID(HexID) of
    ShortHexIDLen: Result := HexID;
    LongHexIDLen: Result := '0x' + Copy(HexID, ShortHexIDLen + 1, ShortHexIDLen - 2);
  end;
end;

function KeyOnRing(const HexKeyID: RawByteString): Integer;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  Key		: pPGPKey;
begin
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  try
    Result := GetKeyByHexID(KeySetMain, HexKeyID, Key);
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function KeyRemove(KeyIDCommaText: PAnsiChar): Integer;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  KeySetFound	: pPGPKeySet;
  KeyCount	: PGPUInt32;
begin
  KeySetFound := NIL;
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  try
    Result := GetKeySetByAnyIDs(Context, KeySetMain, KeyIDCommaText, KeySetFound);
    if Result <> 0 then Exit;
    try
      Result := PGPCountKeys(KeySetFound, KeyCount);
      if Result <> 0 then Exit;
      if KeyCount <> 0 then begin
	Result := PGPRemoveKeys(KeySetFound, KeySetMain);
	if Result <> 0 then Exit;
      end;
    finally
      PGPFreeKeySet(KeySetFound);
    end;
    Result := KeyRings.UpdateKeyRings;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function KeyEnable(HexKeyID: PAnsiChar): Integer;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  Key		: pPGPKey;
begin
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  try
    Result := GetKeyByHexID(KeySetMain, HexKeyID, Key);
    if Result <> 0 then Exit;
    Result := PGPEnableKey(Key);
    if Result <> 0 then Exit;
    Result := KeyRings.UpdateKeyRings;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function KeyDisable(HexKeyID: PAnsiChar): Integer;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  Key		: pPGPKey;
begin
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  try
    Result := GetKeyByHexID(KeySetMain, HexKeyID, Key);
    if Result <> 0 then Exit;
    Result := PGPDisableKey(Key);
    if Result <> 0 then Exit;
    Result := KeyRings.UpdateKeyRings;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function KeyRevoke(HexKeyID, Passphrase: PAnsiChar): Integer;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  Key		: pPGPKey;
begin
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  try
    Result := GetKeyByHexID(KeySetMain, HexKeyID, Key);
    if Result <> 0 then Exit;
    Result := PGPRevokeKey(Key, PGPOPassphrase(Context, Passphrase), PGPOLastOption(Context));
    if Result <> 0 then Exit;
    Result := KeyRings.UpdateKeyRings;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function ChangePassphrase(HexKeyID, OldPassphrase, NewPassphrase: PAnsiChar): Integer;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  OptionList	: pPGPOptionList;
  Key		: pPGPKey;
  KeyList	: pPGPKeyList;
  KeyIter	: pPGPKeyIter;
  SubKey	: pPGPSubKey;
begin
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  OptionList := NIL;
  try
    Result := PGPBuildOptionList(Context, OptionList,
      [
	PGPOPassphrase(Context, OldPassphrase),
	PGPOPassphrase(Context, NewPassphrase)
      ]);
    if Result <> 0 then Exit;
    try
      Result := GetKeyByHexID(KeySetMain, HexKeyID, Key);
      if Result <> 0 then Exit;
      Result := PGPChangePassphrase(Key, OptionList, PGPOLastOption(Context));
      if Result <> 0 then Exit;
      if not GetKeyPropLegacy(Key) then begin
	Result := GetKeyIterAnyOrdering(KeySetMain, KeyList, KeyIter);
	if Result <> 0 then Exit;
	try
	  PGPKeyIterSeek(KeyIter, Key);
	  while PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_SubKey, SubKey) = 0 do begin
	    Result := PGPChangeSubKeyPassphrase(SubKey, OptionList, PGPOLastOption(Context));
	    if Result <> 0 then Exit;
	  end;
	finally
	  PGPFreeKeyIter(KeyIter);
	  PGPFreeKeyList(KeyList);
	end;
      end;
      Result := KeyRings.UpdateKeyRings;
    finally
      PGPFreeOptionList(OptionList);
    end;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function GetUserIDCertOwner(const HexKeyID, UserID: RawByteString): RawByteString;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  Key		: pPGPKey;
  KeySet	: pPGPKeySet;
  KeyIter	: pPGPKeyIter;
  KeyDBObj	: pPGPKeyDBObj;
  UserIDToCheck	: TUserID;
  IDSize	: PGPSize;
  X509Prop	: PGPBoolean;
  ONData	: TX509CertData;
  ONSize	: PGPSize;
begin
  Result := E;
  if KeyRings.InitKeyRings(Context, KeySetMain) = 0 then begin
    KeySet := NIL;
    KeyIter := NIL;
    try
      if (GetKeyByHexID(KeySetMain, HexKeyID, Key) = 0)
      and (PGPNewSingletonKeySet(Key, KeySet) = 0) and (PGPNewKeyIterFromKeySet(KeySet, KeyIter) = 0) then begin
	while PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_UserID, KeyDBObj) = 0 do begin
	  if (PGPGetKeyDBObjDataProperty(KeyDBObj, kPGPUserIDProperty_Name, @UserIDToCheck, SizeOf(TUserID), IDSize) = 0)
	  and (UserID = Utf8OrAnsi(UserIDToCheck)) then begin
	    while PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_Signature, KeyDBObj) = 0 do begin
	      if (PGPGetKeyDBObjBooleanProperty(KeyDBObj, kPGPSigProperty_IsX509, X509Prop) = 0)
	      and (X509Prop = PGPTrue) and GetCertPropIsValid(KeyDBObj)
	      and (PGPGetKeyDBObjDataProperty(KeyDBObj, kPGPSigProperty_X509LongName,
					      @ONData, MAX_509_CERT_SIZE, ONSize) = 0) then begin
		Result := Utf8ToAnsi(ONData);
		Exit;
	      end;
	    end;
	  end;
	end;
      end;
    finally
      PGPFreeKeyIter(KeyIter);
      PGPFreeKeySet(KeySet);
      KeyRings.FreeKeyRings;
    end;
  end
end;


function GetSimpleKeyProp(const HexKeyID: RawByteString; PropertyFlag: DWord): Variant;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  Key		: pPGPKey;
begin
  Result := varNull;
  if KeyRings.InitKeyRings(Context, KeySetMain) <> 0 then Exit;
  try
    if GetKeyByHexID(KeySetMain, HexKeyID, Key) <> 0 then Exit;
    case PropertyFlag of
      spgpKeyPropFlag_KeyID: Result := GetKeyPropKeyID(Key);
      spgpKeyPropFlag_UserID: Result := GetKeyPropUserID(Key);
      spgpKeyPropFlag_Fingerprint: Result := GetKeyPropFingerprint(Key);
      spgpKeyPropFlag_CreationTimeStr: Result := GetKeyPropKeyCreationTimeStr(Key);
      spgpKeyPropFlag_ExpirationTimeStr: Result := GetKeyPropKeyExpirationTimeStr(Key);
      // "number" properties
      spgpKeyPropFlag_KeyBits: Result := GetKeyPropKeyBits(Key);
      spgpKeyPropFlag_X509Cert: Result:= GetKeyPropX509Cert(Key);
      spgpKeyPropFlag_KeyAlg: Result := GetKeyPropKeyAlg(Key);
      spgpKeyPropFlag_HashAlg: Result := GetKeyPropHashAlg(Key);
      spgpKeyPropFlag_Trust: Result := GetKeyPropTrust(Key);
      spgpKeyPropFlag_Validity: Result := GetKeyPropValidity(Key);
      spgpKeyPropFlag_CreationTime: Result := GetKeyPropKeyCreationTime(Key);
      spgpKeyPropFlag_ExpirationTime: Result := GetKeyPropKeyExpirationTime(Key);
      // "boolean" properties
      spgpKeyPropFlag_IsCorrupt: Result := GetKeyPropIsCorrupt(Key);
      spgpKeyPropFlag_IsSecret: Result := GetKeyPropIsSecret(Key);
      spgpKeyPropFlag_IsAxiomatic: Result := GetKeyPropIsAxiomatic(Key);
      spgpKeyPropFlag_IsRevoked: Result := GetKeyPropIsRevoked(Key);
      spgpKeyPropFlag_IsDisabled: Result := GetKeyPropIsDisabled(Key);
      spgpKeyPropFlag_IsExpired: Result := GetKeyPropIsExpired(Key);
      spgpKeyPropFlag_IsSecretShared: Result := GetKeyPropIsSecretShared(Key);
      spgpKeyPropFlag_CanEncrypt: Result := GetKeyPropCanEncrypt(Key);
      spgpKeyPropFlag_CanDecrypt: Result := GetKeyPropCanDecrypt(Key);
      spgpKeyPropFlag_CanSign: Result := GetKeyPropCanSign(Key);
      spgpKeyPropFlag_CanVerify: Result := GetKeyPropCanVerify(Key);
      spgpKeyPropFlag_HasRevoker: Result := GetKeyPropHasRevoker(KeySetMain, Key, NIL);
      spgpKeyPropFlag_HasADK: Result := GetKeyPropHasARR(KeySetMain, Key, NIL);
      spgpKeyPropFlag_HasSubKey: Result := GetKeyPropHasSubKey(KeySetMain, Key);
      spgpKeyPropFlag_LegacyKey: Result := GetKeyPropLegacy(Key);
    end;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function FindKeyProps(const KeyIDCommaText: RawByteString;
		      var KeyPropsList: TKeyPropsList;
		      PropertyFlags, FilterFlags: DWord;
		      KeyOrder: TPGPKeyOrdering): Integer;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  KeySetFound	: pPGPKeySet;
  KeyFilter	: pPGPFilter;
  KeySetFiltered: pPGPKeySet;
  KeyCount	: PGPUInt32;
begin
  KeySetFound := NIL;
  KeyFilter := NIL;
  KeySetFiltered := NIL;
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  try
    if KeyIDCommaText = E then
      KeySetFound := KeySetMain // all keys on key rings
    else Result := GetKeySetByAnyIDs(Context, KeySetMain, KeyIDCommaText, KeySetFound); // key(s) matching KeyData
    if Result <> 0 then Exit;
    try
      if PropertyFlags <> 0 then begin
	Result := GetKeySetProps(Context, KeySetFound,
				 KeyPropsList, PropertyFlags,
				 FilterFlags, KeyOrder);
      end
      else begin
	if FilterFlags <> KeyFilterFlag_AllKeys then begin
	  if FilterFlags and KeyFilterFlag_CountUserIDs <> 0 then
	    Result := CountUserIDs(KeySetFound)
	  else begin
	    Result := GetKeyFilter(Context, FilterFlags, KeyFilter);
	    try
	      if Result <> 0 then Exit;
	      Result := PGPFilterKeySet(KeySetFound, KeyFilter, KeySetFiltered);
	    finally
	      PGPFreeFilter(KeyFilter);
	    end;
	  end;
	end
	else KeySetFiltered := KeySetFound;
	if Result <> 0 then Exit;
	try
	  Result := PGPCountKeys(KeySetFiltered, KeyCount);
	  if Result = 0 then Result := KeyCount;
	finally
	  if KeySetFiltered <> KeySetFound then PGPFreeKeySet(KeySetFiltered);
	end;
      end;
    finally
      if KeySetFound <> KeySetMain then PGPFreeKeySet(KeySetFound);
    end;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function GetKeyFromKeySet(KeySet: pPGPKeySet; var Key: pPGPKey): PGPError;
var
  KeyList	: pPGPKeyList;
  KeyIter	: pPGPKeyIter;
begin
  Key := NIL;
  Result := GetKeyIterAnyOrdering(KeySet, KeyList, KeyIter);
  if Result <> 0 then Exit;
  try
    Result := PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_Key, Key);
  finally
    PGPFreeKeyIter(KeyIter);
    PGPFreeKeyList(KeyList);
  end;
end;

function HasKeyUserID(Key: pPGPKey; const UserID: RawByteString): Longbool;
var
  KeySet	: pPGPKeySet;
  KeyFilter	: pPGPFilter;
  Context	: pPGPContext;
  KeySetFiltered: pPGPKeySet;
  KeyCount	: PGPUInt32;
begin
  Result := FALSE;
  KeySet := NIL;
  KeyFilter := NIL;
  Context := PGPPeekKeyDBObjContext(Key);
  if PGPNewSingletonKeySet(Key, KeySet) <> 0 then Exit;
  try
    if GetUserIDFilter(Context, UserID, TRUE, KeyFilter) <> 0 then Exit;
    try
      if PGPFilterKeySet(KeySet, KeyFilter, KeySetFiltered) <> 0 then Exit;
      try
	Result := (PGPCountKeys(KeySetFiltered, KeyCount) = 0) and (KeyCount <> 0);
      finally
	PGPFreeKeySet(KeySetFiltered);
      end;
    finally
      PGPFreeFilter(KeyFilter);
    end;
  finally
    PGPFreeKeySet(KeySet);
  end;
end;

function GetKeyByHexID(KeySet: pPGPKeySet; const HexID: RawByteString; var Key: pPGPKey): PGPError;
var
  PGPKeyID	: TPGPKeyID7;
begin
  Key := NIL;
  Result := PGPGetKeyIDFromString(PAnsiChar(LowerCase(HexID)), kPGPPublicKeyAlgorithm_Invalid, PGPKeyID);
  if Result = 0 then Result := PGPGetKeyByKeyID(KeySet, PGPKeyID, kPGPPublicKeyAlgorithm_Invalid, Key);
end;

function GetKeyFromNewSet(Key: pPGPKey; NewSet: pPGPKeySet; var NewKey: pPGPKey): PGPError;
var
  PGPKeyID	: TPGPKeyID7;
begin
  NewKey := NIL;
  Result := PGPGetKeyIDFromKey(Key, PGPKeyID);
  if Result = 0 then Result := PGPGetKeyByKeyID(NewSet, PGPKeyID, kPGPPublicKeyAlgorithm_Invalid, NewKey);
end;

function RemoveKeyFromKeySet(Key: pPGPKey; TargetKeySet: pPGPKeySet): PGPError;
var
  SingleKeySet	: pPGPKeySet;
begin
  Result := PGPNewSingletonKeySet(Key, SingleKeySet);
  try
    if Result = 0 then Result := PGPRemoveKeys(SingleKeySet, TargetKeySet);
  finally
    PGPFreeKeySet(SingleKeySet);
  end;
end;

function AddKeyToKeySet(Key: pPGPKey; TargetKeySet: pPGPKeySet): PGPError;
var
  SingleKeySet	: pPGPKeySet;
begin
  Result := PGPNewSingletonKeySet(Key, SingleKeySet);
  try
    if Result = 0 then begin
      if SDKVersion >= PGP7XX then
	Result := PGPCopyKeys(SingleKeySet, PGPPeekKeySetKeyDB(TargetKeySet), TargetKeySet)
      else Result := PGPAddKeys(SingleKeySet, TargetKeySet);
    end;
  finally
    PGPFreeKeySet(SingleKeySet);
  end;
end;

function GetKeyIterAnyOrdering(KeySet: pPGPKeySet; var KeyList: pPGPKeyList; var KeyIter: pPGPKeyIter): PGPError;
begin
  KeyList := NIL;
  KeyIter := NIL;
  if SDKVersion >= PGP7XX then
    Result := PGPNewKeyIterFromKeySet(KeySet, KeyIter)
  else begin
    Result := PGPOrderKeySet(KeySet, kPGPKeyOrdering_Any, PGPFalse, KeyList);
    if Result = 0 then begin
      Result := PGPNewKeyIter(KeyList, KeyIter);
      if Result <> 0 then begin
	PGPFreeKeyList(KeyList);
	KeyList := NIL;
      end;
    end;
  end;
end;

function GetSingleKeyDBObjIter(KeyDBObj: pPGPKeyDBObj; var ObjSet: pPGPKeySet;
			       var ObjList: pPGPKeyList; var ObjIter: pPGPKeyIter): PGPError;
begin
  ObjSet := NIL;
  ObjList := NIL;
  ObjIter := NIL;
  Result := PGPNewSingletonKeySet(KeyDBObj, ObjSet);
  if Result = 0 then Result := GetKeyIterAnyOrdering(ObjSet, ObjList, ObjIter);
  if Result <> 0 then begin
    PGPFreeKeySet(ObjSet);
    ObjSet := NIL;
  end;
end;

function GetKeyFilter(Context: pPGPContext; FilterFlags: DWord; var KeyFilter: pPGPFilter): PGPError;
var
  BoolFilter	: pPGPFilter;
  NumFilter	: pPGPFilter;
begin
  Result := 0;
  KeyFilter := NIL;
  BoolFilter := NIL;
  NumFilter := NIL;
  try
    FilterFlags := FilterFlags or (DWord(FilterFlags and KeyFilterFlag_X509CanEncrypt <> 0) * KeyFilterFlag_CanEncrypt)
			       or (DWord(FilterFlags and KeyFilterFlag_X509CanSign <> 0) * KeyFilterFlag_CanSign);
    if FilterFlags and KeyFilterMask_Boolean <> 0 then begin
      if FilterFlags and KeyFilterFlag_CanEncrypt <> 0 then begin
	Result := PGPNewKeyDBObjBooleanFilter(Context, kPGPKeyProperty_CanEncrypt, PGPTrue, BoolFilter);
      end
      else if FilterFlags and KeyFilterFlag_CanDecrypt <> 0 then begin
	Result := PGPNewKeyDBObjBooleanFilter(Context, kPGPKeyProperty_CanDecrypt, PGPTrue, BoolFilter);
      end
      else if FilterFlags and KeyFilterFlag_CanSign <> 0 then begin
	Result := PGPNewKeyDBObjBooleanFilter(Context, kPGPKeyProperty_CanSign, PGPTrue, BoolFilter);
      end
      else if FilterFlags and KeyFilterFlag_CanVerify <> 0 then begin
	Result := PGPNewKeyDBObjBooleanFilter(Context, kPGPKeyProperty_CanVerify, PGPTrue, BoolFilter);
      end
      else if FilterFlags and KeyFilterFlag_Enabled <> 0 then begin
	Result := PGPNewKeyDBObjBooleanFilter(Context, kPGPKeyProperty_IsDisabled, PGPFalse, BoolFilter);
      end
      else if FilterFlags and KeyFilterFlag_Disabled <> 0 then begin
	Result := PGPNewKeyDBObjBooleanFilter(Context, kPGPKeyProperty_IsDisabled, PGPTrue, BoolFilter);
      end;
      if Result <> 0 then Exit;
    end;
    if FilterFlags and KeyFilterMask_Algorithm <> 0 then begin
      if FilterFlags and KeyFilterFlag_DHDSS <> 0 then begin
	Result := PGPNewKeyDBObjNumericFilter(Context, kPGPKeyProperty_AlgorithmID,
					      kPGPPublicKeyAlgorithm_DSA, kPGPMatchCriterion_Equal, NumFilter);
      end
      else if FilterFlags and KeyFilterFlag_RSA <> 0 then begin
	Result := PGPNewKeyDBObjNumericFilter(Context, kPGPKeyProperty_AlgorithmID,
					      kPGPPublicKeyAlgorithm_RSA, kPGPMatchCriterion_Equal, NumFilter);
      end
      else if FilterFlags and KeyFilterFlag_V4 <> 0 then begin
	if SDKVersion >= PGP7XX then begin
	  Result := PGPNewKeyDBObjNumericFilter(Context, kPGPKeyProperty_Version,
						kPGPKeyVersion_V4, kPGPMatchCriterion_Equal, NumFilter);
	end
	else begin
	  Result := PGPNewKeyDBObjNumericFilter(Context, kPGPKeyProperty_AlgorithmID,
						kPGPPublicKeyAlgorithm_DSA, kPGPMatchCriterion_Equal, NumFilter);
	end;
      end
      else if FilterFlags and KeyFilterFlag_V3 <> 0 then begin
	if SDKVersion >= PGP7XX then begin
	  Result := PGPNewKeyDBObjNumericFilter(Context, kPGPKeyProperty_Version,
						kPGPKeyVersion_V3, kPGPMatchCriterion_LessOrEqual, NumFilter);
	end
	else begin
	  Result := PGPNewKeyDBObjNumericFilter(Context, kPGPKeyProperty_AlgorithmID,
						kPGPPublicKeyAlgorithm_RSA, kPGPMatchCriterion_Equal, NumFilter);
	end;
      end;
      if Result <> 0 then Exit;
    end;
    if (BoolFilter <> NIL) and (NumFilter <> NIL) then begin
      Result := PGPIntersectFilters(BoolFilter, NumFilter, KeyFilter);
      if Result <> 0 then Exit;
      BoolFilter := NIL;
      NumFilter := NIL;
    end
    else if BoolFilter <> NIL then begin
      KeyFilter := BoolFilter;
      BoolFilter := NIL;
    end
    else if NumFilter <> NIL then begin
      KeyFilter := NumFilter;
      NumFilter := NIL;
    end;
    if FilterFlags and KeyFilterMask_X509Props <> 0 then begin
      Result := PGPNewKeyDBObjBooleanFilter(Context, kPGPSigProperty_IsX509, PGPTrue, BoolFilter);
      if Result <> 0 then Exit;
      if KeyFilter <> NIL then
	Result := PGPIntersectFilters(BoolFilter, KeyFilter, KeyFilter)
      else KeyFilter := BoolFilter;
      if Result <> 0 then Exit;
      BoolFilter := NIL;
    end;
  finally
    if (KeyFilter <> BoolFilter) and (BoolFilter <> NIL) then PGPFreeFilter(BoolFilter);
    if (KeyFilter <> NumFilter) and (NumFilter <> NIL) then PGPFreeFilter(NumFilter);
  end;
end;

function GetHexIDFilter(Context: pPGPContext; const HexID: RawByteString;
			IncludeSubKeys: Longbool; var HexIDFilter: pPGPFilter): PGPError;
var
  PGPKeyID	: TPGPKeyID7;
  KeyFilter	: pPGPFilter;
  SubKeyFilter	: pPGPFilter;
begin
  HexIDFilter := NIL;
  KeyFilter := NIL;
  SubKeyFilter := NIL;
  Result := PGPGetKeyIDFromString(PAnsiChar(HexID), kPGPPublicKeyAlgorithm_Invalid, PGPKeyID);
  if Result <> 0 then Exit;
  Result := PGPNewKeyDBObjDataFilter(Context, kPGPKeyProperty_KeyID, @PGPKeyID,
				     SizeOf(TPGPKeyID7), kPGPMatchCriterion_Equal, KeyFilter);
  if Result <> 0 then Exit;
  try
    if IncludeSubKeys then begin
      Result := PGPNewKeyDBObjDataFilter(Context, kPGPSubKeyProperty_KeyID, @PGPKeyID,
					 SizeOf(TPGPKeyID7), kPGPMatchCriterion_Equal, SubKeyFilter);
      if Result <> 0 then Exit;
      try
	Result := PGPUnionFilters(KeyFilter, SubKeyFilter, HexIDFilter);
	KeyFilter := NIL;
	SubKeyFilter := NIL;
      finally
	if SubKeyFilter <> NIL then PGPFreeFilter(SubKeyFilter);
      end;
    end
    else HexIDFilter := KeyFilter;
  finally
    if IncludeSubKeys and (KeyFilter <> NIL) then PGPFreeFilter(KeyFilter);
  end;
end;

function GetUserIDFilter(Context: pPGPContext; const UserID: RawByteString;
			 IncludeUTF8: Longbool; var UserIDFilter: pPGPFilter): PGPError;
var
  AnsiFilter	: pPGPFilter;
  UTF8Filter	: pPGPFilter;
  UTF8ID	: RawByteString;
begin
  UserIDFilter := NIL;
  AnsiFilter := NIL;
  UTF8Filter := NIL;
  Result := PGPNewKeyDBObjDataFilter(Context, kPGPUserIDProperty_Name, PAnsiChar(UserID),
				     Length(UserID), kPGPMatchCriterion_SubString, AnsiFilter);
  if Result <> 0 then Exit;
  try
    if (SDKVersion >= PGP8XX) and IncludeUTF8 then begin
      UTF8ID := AnsiToUtf8(UserID);
      if Length(UTF8ID) > Length(UserID) then begin
	Result := PGPNewKeyDBObjDataFilter(Context, kPGPUserIDProperty_Name, PAnsiChar(UTF8ID),
					   Length(UTF8ID), kPGPMatchCriterion_SubString, UTF8Filter);
	if Result <> 0 then Exit;
	try
	  Result := PGPUnionFilters(AnsiFilter, UTF8Filter, UserIDFilter);
	  AnsiFilter := NIL;
	  UTF8Filter := NIL;
	finally
	  if UTF8Filter <> NIL then PGPFreeFilter(UTF8Filter);
	end;
      end
      else begin
	UserIDFilter := AnsiFilter;
	UTF8ID := E;
      end;
    end
    else UserIDFilter := AnsiFilter;
  finally
    if (UTF8ID <> E) and (AnsiFilter <> NIL) then PGPFreeFilter(AnsiFilter);
  end;
end;

function GetKeyExcludeFilter(Context: pPGPContext; Key: pPGPKeyDBObj): pPGPFilter;
var
  KeyID		: TPGPKeyID7;
  KeyFilter	: pPGPFilter;
begin
  Result := NIL;
  if PGPGetKeyIDFromKey(Key, KeyID) = 0 then begin
    if PGPNewKeyDBObjDataFilter(Context, kPGPKeyProperty_KeyID, @KeyID,
				SizeOf(TPGPKeyID7), kPGPMatchCriterion_Equal, KeyFilter) = 0 then begin
      try
	PGPNegateFilter(KeyFilter, Result);
	KeyFilter := NIL;
      finally
	if KeyFilter <> NIL then PGPFreeFilter(KeyFilter);
      end;
    end;
  end;
end;

function GetKeySetWithoutSigner(Context: pPGPContext; const KeySetToFilter: pPGPKeySet; SignKey: pPGPKeyDBObj;
				var KeySetFiltered: pPGPKeySet): PGPError;
var
  ExcludeFilter	: pPGPFilter;
begin
  KeySetFiltered := NIL;
  ExcludeFilter := GetKeyExcludeFilter(Context, SignKey);
  try
    Result := PGPFilterKeySet(KeySetToFilter, ExcludeFilter, KeySetFiltered);
  finally
    PGPFreeFilter(ExcludeFilter);
  end;
  if Result <> 0 then KeySetFiltered := KeySetToFilter;
end;

function GetKeyFilterByAnyID(Context: pPGPContext; const AnyID: RawByteString;
			     IncludeSubKeys, IncludeUTF8: Longbool;
			     var KeyFilter: pPGPFilter): PGPError;
begin
  if IsHexID(AnyID) <> 0 then
    Result := GetHexIDFilter(Context, AnyID, IncludeSubKeys, KeyFilter)
  else Result := GetUserIDFilter(Context, AnyID, IncludeUTF8, KeyFilter);
end;

function GetKeySetByAnyIDs(Context: pPGPContext;
			   KeySetMain: pPGPKeySet;
			   const KeyIDCommaText: RawByteString;
			   var KeySetFound: pPGPKeySet): PGPError;
var
  KeyDataList	: TRawByteStrings;
  KeyIndex	: PGPUInt32;
  KeyString	: RawByteString;
  KeyFilter	: pPGPFilter;
  KeySetFiltered: pPGPKeySet;
  KeyCount	: PGPUInt32;
begin
  KeySetFound := NIL;
  if Trim(KeyIDCommaText) <> E then begin
    KeyDataList := TRawByteStrings.Create;
    KeyFilter := NIL;
    KeySetFiltered := NIL;
    try
      KeyDataList.CommaText := KeyIDCommaText;
      if SDKVersion >= PGP7XX then
	Result := PGPNewEmptyKeySet(PGPPeekKeySetKeyDB(KeySetMain), KeySetFound)
      else Result := PGPNewKeySet(Context, KeySetFound);
      if Result <> 0 then Exit;
      try
	for KeyIndex := 0 to pred(KeyDataList.Count) do begin
	  KeyString := Trim(KeyDataList[KeyIndex]);
	  if KeyString <> E then begin
	    Result := GetKeyFilterByAnyID(Context, KeyString, FALSE, TRUE, KeyFilter);
	    if Result <> 0 then Exit;
	    try
	      Result := PGPFilterKeySet(KeySetMain, KeyFilter, KeySetFiltered);
	      if Result <> 0 then Exit;
	      try
		Result := PGPCountKeys(KeySetFiltered, KeyCount);
		if Result <> 0 then Exit;
		if KeyCount <> 0 then begin
		  Result := PGPAddKeys(KeySetFiltered, KeySetFound);
		  if Result <> 0 then Exit;
		end
		else begin
		  Result := kPGPError_PublicKeyNotFound;
		  Exit;
		end;
	      finally
		PGPFreeKeySet(KeySetFiltered);
	      end;
	    finally
	      PGPFreeFilter(KeyFilter);
	    end;
	  end;
	end;
      finally
	if Result <> 0 then begin
	  PGPFreeKeySet(KeySetFound);
	  KeySetFound := NIL;
	end;
      end;
    finally
      KeyDataList.Free;
    end;
  end
  else Result := kPGPError_PublicKeyNotFound;
end;

function GetExclusiveKeySet(var KeySetToCheck: pPGPKeySet; KeySetMain: pPGPKeySet;
			    Context: pPGPContext; IgnoreKnownFlag: Integer): PGPError;
var
  KeyListToCheck: pPGPKeyList;
  KeyIterToCheck: pPGPKeyIter;
  KeyToCheck	: pPGPKey;
  KeyIDToCheck	: TKeyID;
  UserIDToCheck	: TUserID;
  KeyCount	: PGPUInt32;
  IDSize	: PGPSize;
  UserIDFound	: pPGPUserID;
  UserIDProp	: PGPBoolean;
  KeyFilter	: pPGPFilter;
  KeySetFound	: pPGPKeySet;
  KeySetToRemove: pPGPKeySet;
begin
  KeyToCheck := NIL;
  Result := GetKeyIterAnyOrdering(KeySetToCheck, KeyListToCheck, KeyIterToCheck);
  if Result <> 0 then Exit;
  try
    if (IgnoreKnownFlag and IgnoreFlag_ByHexID) <> 0 then begin
      while PGPKeyIterNextKeyDBObj(KeyIterToCheck, kPGPKeyDBObjType_Key, KeyToCheck) = 0 do begin
	KeyFilter := NIL;
	KeySetFound := NIL;
	KeySetToRemove := NIL;
	try
	  if (PGPGetKeyDBObjDataProperty(KeyToCheck, kPGPKeyProperty_KeyID,
					 @KeyIDToCheck, kPGPMaxKeyIDStringSize, IDSize) = 0)
	  and (PGPNewKeyDBObjDataFilter(Context, kPGPKeyProperty_KeyID, @KeyIDToCheck,
					IDSize, kPGPMatchCriterion_Equal, KeyFilter) = 0)
	  and (PGPFilterKeySet(KeySetMain, KeyFilter, KeySetFound) = 0)
	  and (PGPCountKeys(KeySetFound, KeyCount) = 0) and (KeyCount <> 0)
	  and (PGPNewSingletonKeySet(KeyToCheck, KeySetToRemove) = 0) then begin
	    PGPRemoveKeys(KeySetToRemove, KeySetToCheck);
	  end;
	finally
	  PGPFreeFilter(KeyFilter);
	  PGPFreeKeySet(KeySetFound);
	  PGPFreeKeySet(KeySetToRemove);
	end;
      end;
      PGPKeyIterRewindKeyDBObj(KeyIterToCheck, kPGPKeyDBObjType_Key);
    end;
    if (PGPCountKeys(KeySetToCheck, KeyCount) = 0) and (KeyCount = 0) then Exit;
    if (IgnoreKnownFlag and IgnoreFlag_ByUserID) <> 0 then begin
      while PGPKeyIterNextKeyDBObj(KeyIterToCheck, kPGPKeyDBObjType_Key, KeyToCheck) = 0 do begin
	while (PGPKeyIterNextKeyDBObj(KeyIterToCheck, kPGPKeyDBObjType_UserID, UserIDFound) = 0)
	and (PGPGetKeyDBObjBooleanProperty(UserIDFound, kPGPUserIDProperty_IsAttribute, UserIDProp) = 0)
	and (UserIDProp = PGPFalse) do begin
	  KeyFilter := NIL;
	  KeySetFound := NIL;
	  KeySetToRemove := NIL;
	  try
	    if (PGPGetKeyDBObjDataProperty(UserIDFound, kPGPUserIDProperty_Name,
					   @UserIDToCheck, SizeOf(TUserID), IDSize) = 0)
	    and (PGPNewKeyDBObjDataFilter(Context, kPGPUserIDProperty_Name, @UserIDToCheck,
					  IDSize, kPGPMatchCriterion_Equal, KeyFilter) = 0)
	    and (PGPFilterKeySet(KeySetMain, KeyFilter, KeySetFound) = 0)
	    and (PGPCountKeys(KeySetFound, KeyCount) = 0) and (KeyCount <> 0)
	    and (PGPNewSingletonKeySet(KeyToCheck, KeySetToRemove) = 0) then begin
	      PGPRemoveKeys(KeySetToRemove, KeySetToCheck);
	      Break;
	    end;
	  finally
	    PGPFreeFilter(KeyFilter);
	    PGPFreeKeySet(KeySetFound);
	    PGPFreeKeySet(KeySetToRemove);
	  end;
	end;
	if (PGPCountKeys(KeySetToCheck, KeyCount) = 0) and (KeyCount = 0) then Break;
      end;
    end;
  finally
    PGPFreeKeyIter(KeyIterToCheck);
    PGPFreeKeyList(KeyListToCheck);
  end;
end;

function GetCertKeySetFromKeySet(const KeySet: pPGPKeySet; EncryptOnly: Longbool; var CertKeySet: pPGPKeySet): PGPError;
var
  X509Filter	: pPGPFilter;
begin
  Result := GetKeyFilter(PGPGetKeySetContext(KeySet),
			 KeyFilterMask_X509Props or (Ord(EncryptOnly = TRUE) * KeyFilterFlag_CanEncrypt), X509Filter);
  if Result <> 0 then Exit;
  try
    Result := PGPFilterKeySet(KeySet, X509Filter, CertKeySet);
  finally
    PGPFreeFilter(X509Filter);
  end;
end;

function GetFingerprintString(const Fingerprint: TFingerprint; FPSize: PGPSize): RawByteString; register; assembler;
asm	// EAX = @Fingerprint, EDX = FPSize, ECX = @Result
  PUSH	EBX
  PUSH	EDI
  PUSH	ESI
  PUSH	EBP
  MOV	ESI,EAX
  MOV	EDI,ECX
  MOV	EAX,EDX
  MOV	EBX,EDX
  SHL	EAX,1
  SHR	EBX,1
  MOV	EBP,EBX
  ADD	EBX,EAX
  INC	EBP
  PUSH	EDX
  MOV	EAX,EBX
  CALL	SYSTEM.@NEWANSISTRING
  MOV	[EDI],EAX
  MOV	EDI,EAX
  POP	ECX
  SUB	EBX,2
  SHR	ECX,1
  SHR	EBP,1
  DEC	ECX
  @LOOP:
  XOR	EAX,EAX
  MOV	AL,[ESI + ECX * 2 + 1]
  MOV	EDX,EAX
  AND	EAX,0Fh
  SHR	EDX,4
  MOV	AH,BYTE PTR[@HEXCHARS + EAX]
  MOV	AL,BYTE PTR[@HEXCHARS + EDX]
  MOV	[EDI + EBX],AX
  XOR	EAX,EAX
  MOV	AL,[ESI + ECX * 2]
  MOV	EDX,EAX
  AND	EAX,0Fh
  SHR	EDX,4
  MOV	AH,BYTE PTR[@HEXCHARS + EAX]
  MOV	AL,BYTE PTR[@HEXCHARS + EDX]
  MOV	[EDI + EBX - 2],AX
  CMP	EBP,ECX
  JE	@MIDDLE
  SUB	EBX,5
  DEC	ECX
  JS	@END
  MOV	BYTE PTR[EDI + EBX + 2],' '
  JMP	@LOOP
  @MIDDLE:
  SUB	EBX,6
  DEC	ECX
  MOV	WORD PTR[EDI + EBX + 2],'  '
  JMP	@LOOP
  @END:
  POP	EBP
  POP	ESI
  POP	EDI
  POP	EBX
  RET
  @HEXCHARS:
  DB	'0123456789ABCDEF';
end;

function SystemTimeToUnixTimeNum: Integer;
var SystemTime: TSystemTime;
begin
  GetSystemTime(SystemTime);
  Result := SystemTimeToUnixTime(SystemTime);
end;

function UnixTimeToLocalTimeStr(UnixSeconds: Integer): RawByteString;
begin
  Result := RawByteString(FormatDateTime('ddd, dd mmm yyyy, tt', UnixTimeToLocalTime(UnixSeconds)));
end;

function GetCertPropIASN(Cert: pPGPKeyDBObj): RawByteString;
var
  IDData	: TX509CertData;
  IDSize	: PGPSize;
begin
  Result := E;
  if PGPGetKeyDBObjDataProperty(Cert, kPGPSigProperty_X509IASN, @IDData, MAX_509_CERT_SIZE, IDSize) = 0 then begin
    SetString(Result, NIL, IDSize);
    Move(IDData, Pointer(Result)^, IDSize);
  end;
end;

function GetCertPropIssuerName(Cert: pPGPKeyDBObj): RawByteString;
var
  INData	: TX509CertData;
  INSize	: PGPSize;
begin
  Result := E;
  if PGPGetKeyDBObjDataProperty(Cert, kPGPSigProperty_X509IssuerLongName, @INData, MAX_509_CERT_SIZE, INSize) = 0 then begin
    Result := Utf8ToAnsi(INData);
  end;
end;

function GetCertPropIssuerSerialNumber(Cert: pPGPKeyDBObj): RawByteString;
var
  ISNData	: TX509CertData;
  ISNSize	: PGPSize;
begin
  Result := E;
  if PGPGetKeyDBObjDataProperty(Cert, kPGPSigProperty_X509IASN, @ISNData, MAX_509_CERT_SIZE, ISNSize) = 0 then begin
    Result := GetSerialNumber(ISNData, ISNSize);
  end;
end;

function GetCertPropOwnerKeyID(Cert: pPGPKeyDBObj): RawByteString;
var
  PGPKeyID	: TPGPKeyID7;
  KeyID		: TKeyID;
begin
  Result := E;
  if (PGPGetKeyIDFromKey(PGPPeekKeyDBObjKey(Cert), PGPKeyID) = 0)
  and (PGPGetKeyIDString(PGPKeyID, kPGPKeyIDString_Full, KeyID) = 0) then begin
    Result := KeyID;
  end;
end;

function GetCertPropOwnerUserID(Cert: pPGPKeyDBObj): RawByteString;
var
  UserIDBuf	: TUserID;
  IDSize	: PGPSize;
begin
  Result := E;
  if PGPGetKeyDBObjDataProperty(PGPPeekKeyDBObjUserID(Cert), kPGPUserIDProperty_Name,
				@UserIDBuf, SizeOf(TUserID), IDSize) = 0 then begin
    if SDKVersion >= PGP8XX then
      Result := Utf8OrAnsi(UserIDBuf)
    else Result := UserIDBuf;
  end;
end;

function GetCertPropOwnerEmail(Cert: pPGPKeyDBObj): RawByteString;
var
  EmailBuf	: TUserID;
  AddrSize	: PGPSize;
begin
  Result := E;
  if PGPGetKeyDBObjDataProperty(PGPPeekKeyDBObjUserID(Cert), kPGPUserIDProperty_EmailAddress,
				@EmailBuf, SizeOf(TUserID), AddrSize) = 0 then begin
    Result := EmailBuf;
  end;
end;

function GetCertPropOwnerName(Cert: pPGPKeyDBObj): RawByteString;
var
  ONData	: TX509CertData;
  ONSize	: PGPSize;
begin
  Result := E;
  if PGPGetKeyDBObjDataProperty(Cert, kPGPSigProperty_X509LongName, @ONData, MAX_509_CERT_SIZE, ONSize) = 0 then begin
    Result := Utf8ToAnsi(ONData);
  end;
end;

function GetCertPropOwnerSerialNumber(Cert: pPGPKeyDBObj): RawByteString;
var
  OSNData	: TX509CertData;
  OSNSize	: PGPSize;
begin
  Result := E;
  if PGPGetKeyDBObjDataProperty(Cert, kPGPSigProperty_X509DERDName, @OSNData, MAX_509_CERT_SIZE, OSNSize) = 0 then begin
    Result := GetSerialNumber(OSNData, OSNSize);
  end;
end;

// local time
function GetCertPropCertCreationTimeStr(Cert: pPGPKeyDBObj): RawByteString;
var
  CreaTime	: PGPTime;
begin
  Result := E;
  if PGPGetKeyDBObjTimeProperty(Cert, kPGPSigProperty_Creation, CreaTime) = 0 then begin
    Result := UnixTimeToLocalTimeStr(PGPGetStdTimeFromPGPTime(CreaTime));
  end;
end;

// local time
function GetCertPropCertExpirationTimeStr(Cert: pPGPKeyDBObj): RawByteString;
var
  ExpTime	: PGPTime;
  CreaTime	: PGPTime;
begin
  Result := E;
  if (PGPGetKeyDBObjTimeProperty(Cert, kPGPSigProperty_Expiration, ExpTime) = 0)
  and (PGPGetKeyDBObjTimeProperty(Cert, kPGPSigProperty_Creation, CreaTime) = 0) then begin
    if ExpTime > CreaTime then Result := UnixTimeToLocalTimeStr(PGPGetStdTimeFromPGPTime(ExpTime));
  end;
end;

// UTC/GMT Unix format seconds
function GetCertPropCertCreationTime(Cert: pPGPKeyDBObj): Integer;
var
  CreaTime	: PGPTime;
begin
  Result := 0;
  if PGPGetKeyDBObjTimeProperty(Cert, kPGPSigProperty_Creation, CreaTime) = 0 then begin
    Result := PGPGetStdTimeFromPGPTime(CreaTime);
  end;
end;

// UTC/GMT Unix format seconds
function GetCertPropCertExpirationTime(Cert: pPGPKeyDBObj): Integer;
var
  ExpTime	: PGPTime;
  CreaTime	: PGPTime;
begin
  Result := 0;
  if (PGPGetKeyDBObjTimeProperty(Cert, kPGPSigProperty_Expiration, ExpTime) = 0)
  and (PGPGetKeyDBObjTimeProperty(Cert, kPGPSigProperty_Creation, CreaTime) = 0) then begin
    if ExpTime > CreaTime then Result := PGPGetStdTimeFromPGPTime(ExpTime);
  end;
end;

function GetCertKeyUsage(Cert: pPGPKeyDBObj): Word;
var
  CertData	: TX509CertData;
  CertSize	: PGPSize;
begin
  Result := 0;
  if (PGPGetKeyDBObjDataProperty(Cert, kPGPSigProperty_X509Certificate,
				 @CertData, MAX_509_CERT_SIZE, CertSize) = 0) then begin
     Result := GetKeyUsage(CertData, CertSize, 0);
  end;
end;

function GetCertPropCanEncrypt(Cert: pPGPKeyDBObj; KeyUsage: Word): Longbool;
begin
  Result := GetCertPropIsValid(Cert) and (KeyUsage and x509KeyUsage_dataEncipherment <> 0);
end;

function GetCertPropCanSign(Cert: pPGPKeyDBObj; KeyUsage: Word): Longbool;
begin
  Result := GetCertPropIsValid(Cert) and (KeyUsage and x509KeyUsage_digitalSignature <> 0);
end;

function GetCertPropIsVerified(Cert: pPGPKeyDBObj): Longbool;
var
  Prop		: PGPBoolean;
begin
  Result := FALSE;
  if PGPGetKeyDBObjBooleanProperty(Cert, kPGPSigProperty_IsVerified, Prop) = 0 then begin
    Result := Boolean(Prop);
  end;
end;

function GetCertPropIsCorrupt(Cert: pPGPKeyDBObj): Longbool;
var
  Prop		: PGPBoolean;
begin
  Result := FALSE;
  if PGPGetKeyDBObjBooleanProperty(Cert, kPGPSigProperty_IsNotCorrupt, Prop) = 0 then begin
    Result := not Boolean(Prop);
  end;
end;

function GetCertPropIsRevoked(Cert: pPGPKeyDBObj): Longbool;
var
  Prop		: PGPBoolean;
begin
  Result := FALSE;
  if PGPGetKeyDBObjBooleanProperty(Cert, kPGPSigProperty_IsRevoked, Prop) = 0 then begin
    Result := Boolean(Prop);
  end;
end;

function GetCertPropIsExpired(Cert: pPGPKeyDBObj): Longbool;
var
  Prop		: PGPBoolean;
begin
  Result := FALSE;
  if PGPGetKeyDBObjBooleanProperty(Cert, kPGPSigProperty_IsExpired, Prop) = 0 then begin
    Result := Boolean(Prop);
  end;
end;

function GetCertPropIsRoot(Cert: pPGPKeyDBObj): Longbool;
var
  INData	: TX509CertData;
  INSize	: PGPSize;
  ONData	: TX509CertData;
  ONSize	: PGPSize;
  SignKey	: pPGPKeyDBObj;
begin
  Result := FALSE;
  if (PGPGetKeyDBObjDataProperty(Cert, kPGPSigProperty_X509IssuerLongName, @INData, MAX_509_CERT_SIZE, INSize) = 0)
  and (PGPGetKeyDBObjDataProperty(Cert, kPGPSigProperty_X509LongName, @ONData, MAX_509_CERT_SIZE, ONSize) = 0)
  and (PGPGetSigCertifierKey(Cert, PGPPeekKeyDBRootKeySet(PGPPeekKeyDBObjKeyDB(Cert)), SignKey) = 0)
  and (PGPPeekKeyDBObjKey(Cert) = SignKey) and (INData = ONData) then begin
    Result := TRUE;
  end;
end;

function GetCertPropIsValid(Cert: pPGPKeyDBObj): Longbool;
var
  NotCorrupt	: PGPBoolean;
  Expired	: PGPBoolean;
  Revoked	: PGPBoolean;
begin
  Result := FALSE;
  if (PGPGetKeyDBObjBooleanProperty(Cert, kPGPSigProperty_IsNotCorrupt, NotCorrupt) = 0)
  and (PGPGetKeyDBObjBooleanProperty(Cert, kPGPSigProperty_IsExpired, Expired) = 0)
  and (PGPGetKeyDBObjBooleanProperty(Cert, kPGPSigProperty_IsRevoked, Revoked) = 0) then begin
    Result := (NotCorrupt = PGPTrue) and (Expired = PGPFalse) and (Revoked = PGPFalse);
  end;
end;

function GetSubKeyPropKeyID(SubKey: pPGPKey): RawByteString;
var
  PGPKeyID	: TPGPKeyID7;
  KeyID		: TKeyID;
begin
  Result := E;
  if (PGPGetKeyIDFromSubKey(SubKey, PGPKeyID) = 0)
  and (PGPGetKeyIDString(PGPKeyID, kPGPKeyIDString_Full, KeyID) = 0) then begin
    Result := KeyID;
  end;
end;

function GetSubKeyPropIsRevoked(SubKey: pPGPKey): Longbool;
var
  Prop		: PGPBoolean;
begin
  Result := FALSE;
  if PGPGetKeyDBObjBooleanProperty(SubKey, kPGPSubKeyProperty_IsRevoked, Prop) = 0 then begin
    Result := Boolean(Prop);
  end;
end;

function GetSubKeyPropIsExpired(SubKey: pPGPKey): Longbool;
var
  Prop		: PGPBoolean;
begin
  Result := FALSE;
  if PGPGetKeyDBObjBooleanProperty(SubKey, kPGPSubKeyProperty_IsExpired, Prop) = 0 then begin
    Result := Boolean(Prop);
  end;
end;

function GetKeyPropKeyID(Key: pPGPKey): RawByteString;
var
  PGPKeyID	: TPGPKeyID7;
  KeyID		: TKeyID;
begin
  Result := E;
  if (PGPGetKeyIDFromKey(Key, PGPKeyID) = 0)
  and (PGPGetKeyIDString(PGPKeyID, kPGPKeyIDString_Full, KeyID) = 0) then begin
    Result := KeyID;
  end;
end;

function GetKeyPropUserID(Key: pPGPKey): RawByteString;
var
  UserIDBuf	: TUserID;
  IDSize	: PGPSize;
begin
  Result := E;
  if PGPGetPrimaryUserIDNameBuffer(Key, UserIDBuf, SizeOf(TUserID), IDSize) = 0 then begin
    if SDKVersion >= PGP8XX then
      Result := Utf8OrAnsi(UserIDBuf)
    else Result := UserIDBuf;
  end;
end;

function GetKeyPropFingerprint(Key: pPGPKey): RawByteString;
var
  Fingerprint	: TFingerprint;
  FPSize	: PGPSize;
begin
  Result := E;
  if (PGPGetKeyDBObjDataProperty(Key, kPGPKeyProperty_Fingerprint, @Fingerprint, kPGPmaxFingerprintSize, FPSize) = 0)
  and (FPSize <> 0) then begin
    Result := GetFingerprintString(Fingerprint, FPSize);
  end;
end;

// local time
function GetKeyPropKeyCreationTimeStr(Key: pPGPKey): RawByteString;
var
  CreaTime	: PGPTime;
begin
  Result := E;
  if PGPGetKeyDBObjTimeProperty(Key, kPGPKeyProperty_Creation, CreaTime) = 0 then begin
    Result := UnixTimeToLocalTimeStr(PGPGetStdTimeFromPGPTime(CreaTime));
  end;
end;

// local time
function GetKeyPropKeyExpirationTimeStr(Key: pPGPKey): RawByteString;
var
  ExpTime	: PGPTime;
  CreaTime	: PGPTime;
begin
  Result := E;
  if (PGPGetKeyDBObjTimeProperty(Key, kPGPKeyProperty_Expiration, ExpTime) = 0)
  and (PGPGetKeyDBObjTimeProperty(Key, kPGPKeyProperty_Creation, CreaTime) = 0) then begin
    if ExpTime > CreaTime then Result := UnixTimeToLocalTimeStr(PGPGetStdTimeFromPGPTime(ExpTime));
  end;
end;

function GetKeyPropKeyBits(Key: pPGPKey): RawByteString;
var
  Error		: PGPError;
  KeySet	: pPGPKeySet;
  KeyList	: pPGPKeyList;
  KeyIter	: pPGPKeyIter;
  SubKey	: pPGPSubKey;
  KeyAlg	: PGPInt32;
  KeyBits	: PGPInt32;
  SubKeyBits	: PGPInt32;
begin
  Result := E;
  Error := 0;
  KeySet := NIL;
  if (PGPGetKeyDBObjNumericProperty(Key, kPGPKeyProperty_AlgorithmID, KeyAlg) = 0)
  and (PGPGetKeyDBObjNumericProperty(Key, kPGPKeyProperty_Bits, KeyBits) = 0)
  and (GetSingleKeyDBObjIter(Key, KeySet, KeyList, KeyIter) = 0) then begin
    try
      SubKeyBits := 0;
      PGPKeyIterSeek(KeyIter, Key);
      while (PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_SubKey, SubKey) = 0) do begin
	if not (GetSubKeyPropIsRevoked(SubKey) or GetSubKeyPropIsExpired(SubKey)) then begin
	  Error := PGPGetKeyDBObjNumericProperty(SubKey, kPGPSubKeyProperty_Bits, SubKeyBits);
	  Break;
	end;
      end;
      if Error = 0 then begin
	if SubKeyBits > 0 then
	  Result := IntToStr(SubKeyBits) + SL + IntToStr(KeyBits) + ' bits'
	else Result := IntToStr(KeyBits) + ' bits';
      end;
    finally
      PGPFreeKeyIter(KeyIter);
      PGPFreeKeyList(KeyList);
      PGPFreeKeySet(KeySet);
    end;
  end;
end;

function GetKeyPropX509Cert(Key: pPGPKey): TX509Type;
var
  SingleKeySet	: pPGPKeySet;
  SingleKeyList : pPGPKeyList;
  SingleKeyIter	: pPGPKeyIter;
  NextResult	: TX509Type;
  ObjFound	: pPGPKeyDBObj;
  X509Prop	: PGPBoolean;
begin
  Result := NoCert;
  if (SDKVersion >= PGP7XX) and (GetSingleKeyDBObjIter(Key, SingleKeySet, SingleKeyList, SingleKeyIter) = 0) then begin
    try
      while PGPKeyIterNextKeyDBObj(SingleKeyIter, kPGPKeyDBObjType_Any, ObjFound) = 0 do begin
	NextResult := NoCert;
	if (PGPGetKeyDBObjBooleanProperty(ObjFound, kPGPSigProperty_IsX509, X509Prop) = 0)
	and (X509Prop = PGPTrue) then begin
	  if GetCertPropIsValid(ObjFound) then begin
	    if GetCertPropIsRoot(ObjFound) then
	      NextResult := RootCert
	    else NextResult := StdCert;
	  end
	  else NextResult := BadCert;
	end;
	if NextResult > Result then Result := NextResult;
      end;
    finally
      PGPFreeKeyIter(SingleKeyIter);
      PGPFreeKeyList(SingleKeyList);
      PGPFreeKEySet(SingleKeySet);
    end;
  end;
end;

function GetKeyPropKeyAlg(Key: pPGPKey): TKeyAlgorithm;
var
  KeyAlg	: PGPInt32;
  KeySet	: pPGPKeySet;
  KeyList	: pPGPKeyList;
  KeyIter	: pPGPKeyIter;
  SubKey	: pPGPSubKey;
begin
  Result := KeyAlgorithm_Invalid;
  if PGPGetKeyDBObjNumericProperty(Key, kPGPKeyProperty_AlgorithmID, KeyAlg) = 0 then begin
    Result := TKeyAlgorithm(KeyAlg);
    if (Result = KeyAlgorithm_DSS) and (GetSingleKeyDBObjIter(Key, KeySet, KeyList, KeyIter) = 0) then begin
      try
	PGPKeyIterSeek(KeyIter, Key);
	if PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_SubKey, SubKey) = 0 then Result := KeyAlgorithm_DHDSS;
      finally
	PGPFreeKeyIter(KeyIter);
	PGPFreeKeyList(KeyList);
	PGPFreeKeySet(KeySet);
      end;
    end;
  end;
end;

function GetKeyPropHashAlg(Key: pPGPKey): THashAlgorithm;
var
  HashAlg	: PGPInt32;
begin
  Result := HashAlgorithm_Invalid;
  if SDKVersion >= PGP7XX then begin
    if PGPGetKeyDBObjNumericProperty(Key, kPGPKeyProperty_HashAlgorithmID, HashAlg) <> 0 then Exit;
  end
  else if PGPGetHashAlgUsed(Key, HashAlg) <> 0 then Exit;
  Result := GetTHashAlgorithm(HashAlg);
end;

function GetKeyPropLegacy(Key: pPGPKey): Longbool;
var
  KeyVersion	: PGPInt32;
begin
  Result := TRUE;
  if SDKVersion >= PGP7XX then begin
    if PGPGetKeyDBObjNumericProperty(Key, kPGPKeyProperty_Version, KeyVersion) = 0 then begin
      Result := (KeyVersion <= kPGPKeyVersion_V3);
    end;
  end
  else Result := (GetKeyPropKeyAlg(Key) = KeyAlgorithm_RSA);
end;

function GetKeyPropTrust(Key: pPGPKey): TTrustLevel;
var
  KeyTrust	: PGPInt32;
begin
  Result := KeyTrust_Undefined;
  if PGPGetKeyDBObjNumericProperty(Key, kPGPKeyProperty_Trust, KeyTrust) = 0 then begin
    Result := TTrustLevel(KeyTrust);
  end;
end;

function GetKeyPropValidity(Key: pPGPKey): TValidityLevel;
var
  KeyValidity	: PGPInt32;
begin
  Result := Validity_Unknown;
  if PGPGetKeyDBObjNumericProperty(Key, kPGPKeyProperty_Validity, KeyValidity) = 0 then begin
    Result := TValidityLevel(KeyValidity);
  end;
end;

// UTC/GMT Unix format seconds
function GetKeyPropKeyCreationTime(Key: pPGPKey): Integer;
var
  CreaTime	: PGPTime;
begin
  Result := 0;
  if PGPGetKeyDBObjTimeProperty(Key, kPGPKeyProperty_Creation, CreaTime) = 0 then begin
    Result := PGPGetStdTimeFromPGPTime(CreaTime);
  end;
end;

// UTC/GMT Unix format seconds
function GetKeyPropKeyExpirationTime(Key: pPGPKey): Integer;
var
  ExpTime	: PGPTime;
  CreaTime	: PGPTime;
begin
  Result := 0;
  if (PGPGetKeyDBObjTimeProperty(Key, kPGPKeyProperty_Expiration, ExpTime) = 0)
  and (PGPGetKeyDBObjTimeProperty(Key, kPGPKeyProperty_Creation, CreaTime) = 0) then begin
    if ExpTime > CreaTime then Result := PGPGetStdTimeFromPGPTime(ExpTime);
  end;
end;

function GetKeyPropIsCorrupt(Key: pPGPKey): Longbool;
var
  Prop		: PGPBoolean;
begin
  Result := FALSE;
  if PGPGetKeyDBObjBooleanProperty(Key, kPGPKeyProperty_IsNotCorrupt, Prop) = 0 then begin
    Result := not Boolean(Prop);
  end;
end;

function GetKeyPropIsSecret(Key: pPGPKey): Longbool;
var
  Prop		: PGPBoolean;
begin
  Result := FALSE;
  if PGPGetKeyDBObjBooleanProperty(Key, kPGPKeyProperty_IsSecret, Prop) = 0 then begin
    Result := Boolean(Prop);
  end;
end;

function GetKeyPropIsAxiomatic(Key: pPGPKey): Longbool;
var
  Prop		: PGPBoolean;
begin
  Result := FALSE;
  if PGPGetKeyDBObjBooleanProperty(Key, kPGPKeyProperty_IsAxiomatic, Prop) = 0 then begin
    Result := Boolean(Prop);
  end;
end;

function GetKeyPropIsSigningKey(Key: pPGPKey): Longbool;
var
  Prop		: PGPBoolean;
begin
  Result := FALSE;
  if PGPGetKeyDBObjBooleanProperty(Key, kPGPKeyProperty_IsSigningKey, Prop) = 0 then begin
    Result := Boolean(Prop);
  end;
end;

function GetKeyPropIsRevoked(Key: pPGPKey): Longbool;
var
  Prop		: PGPBoolean;
begin
  Result := FALSE;
  if PGPGetKeyDBObjBooleanProperty(Key, kPGPKeyProperty_IsRevoked, Prop) = 0 then begin
    Result := Boolean(Prop);
  end;
end;

function GetKeyPropIsDisabled(Key: pPGPKey): Longbool;
var
  Prop		: PGPBoolean;
begin
  Result := FALSE;
  if PGPGetKeyDBObjBooleanProperty(Key, kPGPKeyProperty_IsDisabled, Prop) = 0 then begin
    Result := Boolean(Prop);
  end;
end;

function GetKeyPropIsExpired(Key: pPGPKey): Longbool;
var
  Prop		: PGPBoolean;
begin
  Result := FALSE;
  if PGPGetKeyDBObjBooleanProperty(Key, kPGPKeyProperty_IsExpired, Prop) = 0 then begin
    Result := Boolean(Prop);
  end;
end;

function GetKeyPropIsSecretShared(Key: pPGPKey): Longbool;
var
  Prop		: PGPBoolean;
begin
  Result := FALSE;
  if PGPGetKeyDBObjBooleanProperty(Key, kPGPKeyProperty_IsSecretShared, Prop) = 0 then begin
    Result := Boolean(Prop);
  end;
end;

function GetKeyPropCanEncrypt(Key: pPGPKey): Longbool;
var
  Prop		: PGPBoolean;
begin
  Result := FALSE;
  if PGPGetKeyDBObjBooleanProperty(Key, kPGPKeyProperty_CanEncrypt, Prop) = 0 then begin
    Result := Boolean(Prop);
  end;
end;

function GetKeyPropCanDecrypt(Key: pPGPKey): Longbool;
var
  Prop		: PGPBoolean;
begin
  Result := FALSE;
  if PGPGetKeyDBObjBooleanProperty(Key, kPGPKeyProperty_CanDecrypt, Prop) = 0 then begin
    Result := Boolean(Prop);
  end;
end;

function GetKeyPropCanSign(Key: pPGPKey): Longbool;
var
  Prop		: PGPBoolean;
begin
  Result := FALSE;
  if PGPGetKeyDBObjBooleanProperty(Key, kPGPKeyProperty_CanSign, Prop) = 0 then begin
    Result := Boolean(Prop);
  end;
end;

function GetKeyPropCanVerify(Key: pPGPKey): Longbool;
var
  Prop		: PGPBoolean;
begin
  Result := FALSE;
  if PGPGetKeyDBObjBooleanProperty(Key, kPGPKeyProperty_CanVerify, Prop) = 0 then begin
    Result := Boolean(Prop);
  end;
end;

function GetKeyPropHasCertificate(Key: pPGPKey): Longbool;
begin
  Result := GetKeyPropCanSign(Key) and (GetKeyPropX509Cert(Key) <> NoCert);
end;

function GetKeyPropHasRevoker(KeySet: pPGPKeySet; Key: pPGPKey; const RevKeyIDList: TRevKeyIDs): Longbool;
var
  RevKeyCount	: PGPUInt32;
  RevKeyIndex	: PGPUInt32;
  RevKey	: pPGPKey;
  RevKeyID	: TPGPKeyID7;
  KeyID		: TKeyID;
begin
  Result := FALSE;
  if (PGPCountRevocationKeys(Key, RevKeyCount) = 0) and (RevKeyCount > 0) then begin
    Result := TRUE;
    if RevKeyIDList <> NIL then begin
      for RevKeyIndex := 0 to pred(RevKeyCount) do begin
	if PGPGetIndexedRevocationKey(Key, KeySet, RevKeyIndex, RevKey, RevKeyID) = 0 then begin
	  if PGPGetKeyIDString(RevKeyID, kPGPKeyIDString_Full, KeyID) = 0 then begin
	    RevKeyIDList.Add(KeyID);
	  end;
	end;
      end;
    end;
  end;
end;

function GetKeyPropHasARR(KeySet: pPGPKeySet; Key: pPGPKey; const ADKeyIDList: TADKeyIDs): TADKType;
var
  ARKeyCount	: PGPUInt32;
  ARKeyIndex	: PGPUInt32;
  ARKey		: pPGPKey;
  ARKeyID	: TPGPKeyID7;
  ARClass	: PGPByte;
  KeyID		: TKeyID;
begin
  Result := NoADK;
  ARKey := NIL;
  if (PGPCountAdditionalRecipientRequests(Key, ARKeyCount) = 0) and (ARKeyCount > 0) then begin
    Result := SimpleADK;
    for ARKeyIndex := 0 to pred(ARKeyCount) do begin
      if PGPGetIndexedAdditionalRecipientRequestKey(Key, KeySet, ARKeyIndex, ARKey, ARKeyID, ARClass) = 0 then begin
	if ARClass > 0 then Result := EnforcedADK;
	if ADKeyIDList <> NIL then begin
	  if PGPGetKeyIDString(ARKeyID, kPGPKeyIDString_Full, KeyID) = 0 then begin
	    ADKeyIDList.Add(KeyID);
	  end;
	end;
      end;
    end;
  end;
end;

function GetKeyPropHasSubKey(KeySet: pPGPKeySet; Key: pPGPKey): Longbool;
var
  KeyList	: pPGPKeyList;
  KeyIter	: pPGPKeyIter;
  SubKey	: pPGPSubKey;
begin
  Result := FALSE;
  if not GetKeyPropLegacy(Key) then begin
    if GetKeyIterAnyOrdering(KeySet, KeyList, KeyIter) = 0 then begin
      try
	PGPKeyIterSeek(KeyIter, Key);
	Result := (PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_SubKey, SubKey) = 0);
      finally
	PGPFreeKeyIter(KeyIter);
	PGPFreeKeyList(KeyList);
      end;
    end;
  end;
end;

function GetKeyPropUserIDs(Key: pPGPKey; var KeyPropsRec: TKeyPropsRec; Props, Filters: DWord): Integer;
const
  F_CanEncrypt	= KeyFilterFlag_X509CanEncrypt or KeyFilterFlag_CanEncrypt;
  F_CanSign	= KeyFilterFlag_X509CanSign or KeyFilterFlag_CanSign;
var
  KeySet	: pPGPKeySet;
  KeyList	: pPGPKeyList;
  KeyIter	: pPGPKeyIter;
  AddUserIDs	: Longbool;
  AddSignerIDs	: Longbool;
  AddX509Data	: Longbool;
  AddAny	: Longbool;
  SigCount	: Integer;
  UserID	: pPGPUserID;
  UserIDProp	: PGPBoolean;
  UserIDNum	: PGPInt32;
  UserIDBuf	: TUserID;
  IDSize	: PGPSize;
  IDBuffer	: RawByteString;
  KeySig	: pPGPSig;
  SigProp	: PGPBoolean;
  LastUIDIndex	: Integer;
  SignKeyID	: TPGPKeyID7;
  KeyID		: TKeyID;
begin
  Result := GetSingleKeyDBObjIter(Key, KeySet, KeyList, KeyIter);
  if Result <> 0 then Exit;
  try
    PGPKeyIterSeek(KeyIter, Key);
    with KeyPropsRec do begin
      AddUserIDs := (Props and spgpKeyPropFlag_IncludeUserIDs <> 0) and (kUserIDList <> NIL);
      AddSignerIDs := AddUserIDs and (Props and spgpKeyPropFlag_IncludeSignerIDs <> 0) and (kSignerIDList <> NIL);
      AddX509Data := (SDKVersion >= PGP7XX) and (Props and spgpKeyPropFlag_X509Cert <> 0);
      AddAny := AddSignerIDs or AddX509Data;
      kX509Key := AddX509Data;
    end;
    SigCount := 0;
    // reset kX509Cert for filtering
    if AddUserIDs and AddX509Data then KeyPropsRec.kX509Cert := NoCert;
    // add UserIDs one by one to UserIDList
    while PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_UserID, UserID) = 0 do begin
      if AddUserIDs then begin
	Result := PGPGetKeyDBObjBooleanProperty(UserID, kPGPUserIDProperty_IsAttribute, UserIDProp);
	if Result <> 0 then Exit;
	// check for attributes (picture, e.g.)
	if UserIDProp = PGPFalse then begin
	  // UserID validity
	  Result := PGPGetKeyDBObjNumericProperty(UserID, kPGPUserIDProperty_Validity, UserIDNum);
	  if Result <> 0 then Exit;
	  // UserID RawByteString
	  Result := PGPGetKeyDBObjDataProperty(UserID, kPGPUserIDProperty_Name, @UserIDBuf, SizeOf(TUserID), IDSize);
	  if Result <> 0 then Exit;
	  try
	    with KeyPropsRec, kUserIDList do begin
	      if SDKVersion >= PGP8XX then
		KeyValidities[Add(Utf8OrAnsi(UserIDBuf))] := TValidityLevel(UserIDNum)
	      else KeyValidities[Add(UserIDBuf)] := TValidityLevel(UserIDNum);
	      if AddSignerIDs and (kSignerIDList <> NIL) then kSignerIDList.Add(E);
	    end;
	  except
	    Result := kPGPError_OutOfMemory;
	    Exit;
	  end;
	end;
      end;
      if AddAny then begin
	while PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_Signature, KeySig) = 0 do begin
	  inc(SigCount);
	  if (PGPGetKeyDBObjBooleanProperty(KeySig, kPGPSigProperty_IsX509, SigProp) = 0) then begin
	    if SigProp = PGPTrue then begin
	      if AddUserIDs then with KeyPropsRec.kUserIDList do begin
		LastUIDIndex := pred(Count);
		// add X509Type to last UserIDList entry
		if GetCertPropIsRoot(KeySig) then
		  X509Types[LastUIDIndex] := RootCert
		else X509Types[LastUIDIndex] := StdCert;
		// add X509KeyUsage to last UserIDList entry
		X509KeyUsage[LastUIDIndex] := GetCertKeyUsage(KeySig);
		// hide non-encrypting certificates
		if Filters and F_CanEncrypt <> 0 then begin
		  if not GetCertPropCanEncrypt(KeySig, X509KeyUsage[LastUIDIndex]) then begin
		    X509Types[LastUIDIndex] := NoCert;
		  end;
		end
		// hide non-signing certificates
		else if Filters and F_CanSign <> 0 then begin
		  if not GetCertPropCanSign(KeySig, X509KeyUsage[LastUIDIndex]) then begin
		    X509Types[LastUIDIndex] := NoCert;
		  end;
		end;
		if X509Types[LastUIDIndex] > NoCert then begin
		  if not GetCertPropIsValid(KeySig) then
		    // mark invalid certificates
		    X509Types[LastUIDIndex] := BadCert
		  else begin
		    // add ChainValidity to last UserIDList entry
		    if GetCertPropIsVerified(KeySig) then begin
		      ChainValidities[LastUIDIndex] := TX509ChainValidity(Ord(X509Types[LastUIDIndex]) - Ord(BadCert));
		    end;
		    // change last UserIDList entry to certificate owner's long name
		    if AddX509Data then begin
		      Strings[LastUIDIndex] := GetCertPropOwnerUserID(KeySig) + #9 + GetCertPropOwnerName(KeySig);
		    end;
		  end;
		  // update kX509Cert with highest value after certificate filtering
		  with KeyPropsRec do if X509Types[LastUIDIndex] > kX509Cert then kX509Cert := X509Types[LastUIDIndex];
		end;
	      end;
	    end
	    else KeyPropsRec.kX509Key := FALSE;
	  end;
	  // add comma separated HexIDs of signing keys to SignerIDList
	  if AddSignerIDs then begin
	    if SDKVersion >= PGP7XX then
	      IDBuffer := IDBuffer + GetKeyPropKeyID(PGPPeekKeyDBObjKey(KeySig)) + CM
	    else begin
	      Result := PGPGetKeyIDOfCertifier(KeySig, SignKeyID);
	      if Result <> 0 then Exit;
	      Result := PGPGetKeyIDString(SignKeyID, kPGPKeyIDString_Full, KeyID);
	      if Result <> 0 then Exit;
	      IDBuffer := IDBuffer + KeyID + CM;
	    end;
	  end;
	end;
	if AddSignerIDs then begin
	  Delete(IDBuffer, Length(IDBuffer), 1);
	  try
	    with KeyPropsRec.kSignerIDList do Strings[pred(Count)] := IDBuffer;
	  except
	    Result := kPGPError_OutOfMemory;
	    Exit;
	  end;
	end;
      end;
    end;
    if SigCount = 0 then KeyPropsRec.kX509Key := FALSE;
  finally
    PGPFreeKeyIter(KeyIter);
    PGPFreeKeyList(KeyList);
    PGPFreeKeySet(KeySet);
  end;
end;

function GetGroupsList(Context: pPGPContext; KeySetMain: pPGPKeySet; var GroupsList: TGroupsList): Integer;
var
  GroupFileSpec	: pPGPFileSpec;
  GroupSet	: pPGPGroupSet;
  GroupCount	: PGPUInt32;
  GroupIndex	: PGPUInt32;
  GroupID	: PGPGroupID;
  GroupInfo	: TPGPGroupInfo;
  KeyCount	: PGPUInt32;
  ItemCount	: PGPUInt32;
  GroupItemIter	: pPGPGroupItemIter;
  GroupItem	: TPGPGroupItem;
  HexIDList	: RawByteString;
  Key		: pPGPKey;
  HexID		: RawByteString;
begin
  Result := kPGPError_OutOfMemory;
  if GroupsList <> NIL then begin
    try
      GroupsList.Clear;
      GroupFileSpec := NIL;
      GroupSet := NIL;
      Result := PGPNewFileSpecFromFullPath(Context, PAnsiChar(KeyRings.GroupsFile), GroupFileSpec);
      if Result <> 0 then Exit;
      try
	Result := PGPNewGroupSetFromFile(Context, GroupFileSpec, GroupSet);
	if Result <> 0 then Exit;
	try
	  Result := PGPCountGroupsInSet(GroupSet, GroupCount);
	  if (Result <> 0) or (GroupCount = 0) then Exit;
	  for GroupIndex := 0 to pred(GroupCount) do begin
	    Result := PGPGetIndGroupID(GroupSet, GroupIndex, GroupID);
	    if Result <> 0 then Continue;
	    Result := PGPGetGroupInfo(GroupSet, GroupID, GroupInfo);
	    if Result <> 0 then Continue;
	    Result := PGPCountGroupItems(GroupSet, GroupID, PGPTrue, KeyCount, ItemCount);
	    if (Result <> 0) or (KeyCount = 0) then Continue;
	    with GroupInfo do begin
	      if Description <> E then
		GroupsList.Add(RawByteString(Description) + SP + LT + RawByteString(Name) + GT + TAB)
	      else GroupsList.Add(LT + RawByteString(Name) + GT + TAB);
	    end;
	    GroupItemIter := NIL;
	    Result := PGPNewGroupItemIter(GroupSet, GroupID, kPGPGroupIterFlags_AllKeysRecursive, GroupItemIter);
	    if Result <> 0 then Continue;
	    try
	      HexIDList := E;
	      while PGPGroupItemIterNext(GroupItemIter, GroupItem) = 0 do begin
		with GroupItem.Item.Key do begin
		  if SDKVersion >= PGP7XX then
		    Result := PGPGetKeyByKeyID(KeySetMain, KeyStruct7.KeyID, kPGPPublicKeyAlgorithm_Invalid, Key)
		  else Result := PGPGetKeyByKeyID(KeySetMain, KeyStruct6.KeyID, kPGPPublicKeyAlgorithm_Invalid, Key);
		end;
		if Result <> 0 then Continue;
		HexID := GetKeyPropKeyID(Key);
		if HexID <> E then HexIDList := HexIDList + HexID + CM;
	      end;
	      if HexIDList <> E then begin
		Delete(HexIDList, Length(HexIDList), 1);
		with GroupsList do Strings[GroupIndex] := Strings[GroupIndex] + HexIDList;
	      end;
	    finally
	      PGPFreeGroupItemIter(GroupItemIter);
	    end;
	  end;
	  Result := GroupsList.Count;
	finally
	  if GroupSet <> NIL then PGPFreeGroupSet(GroupSet);
	end;
      finally
	PGPFreeFileSpec(GroupFileSpec);
      end;
    except
      Result := 0;
    end;
    if (Result <= 0) and (GroupsList <> NIL) then GroupsList.Clear;
  end;
end;

function GetCertProps(Cert: pPGPKeyDBObj; var CertPropsRec: TX509CertPropsRec): PGPError;
var
  CertASN1Parser: TX509CertASN1Parser;
  CertData	: TX509CertData;
  CertSize	: PGPSize;
  CertKey	: pPGPKeyDBObj;
begin
  Result := 0;
  try
    CertASN1Parser := TX509CertASN1Parser.Create;
    if CertASN1Parser <> NIL then begin
      try
	Result := PGPGetKeyDBObjDataProperty(Cert, kPGPSigProperty_X509Certificate,
					     @CertData, MAX_509_CERT_SIZE, CertSize);
	if Result <> 0 then Exit;
	CertASN1Parser.ParseCertificate(CertData, CertSize);
	with CertPropsRec, CertASN1Parser do begin
	  x509CertificateID := IssuerCertID;
	  x509IssuerKeyAlgorithm := IssuerKeyAlgorithm;
	  x509IssuerHashAlgorithm := IssuerHashAlgorithm;
	  x509IssuerLongName := GetCertPropIssuerName(Cert);
	  x509IssuerSerialNumber := GetCertPropIssuerSerialNumber(Cert);
	  x509OwnerKeyHexID := GetCertPropOwnerKeyID(Cert);
	  x509OwnerCertUserID := GetCertPropOwnerUserID(Cert);
	  x509OwnerLongName := GetCertPropOwnerName(Cert);
	  x509OwnerSerialNumber := GetCertPropOwnerSerialNumber(Cert);
	  x509CreaTimeStr := GetCertPropCertCreationTimeStr(Cert);
	  x509ExpTimeStr := GetCertPropCertExpirationTimeStr(Cert);
	  x509CreaTimeNum := GetCertPropCertCreationTime(Cert);
	  x509ExpTimeNum := GetCertPropCertExpirationTime(Cert);
	  x509KeyUsages := KeyUsages;
	  x509IsVerified := GetCertPropIsVerified(Cert);
	  x509IsCorrupt := GetCertPropIsCorrupt(Cert);
	  x509IsRevoked := GetCertPropIsRevoked(Cert);
	  x509IsExpired := GetCertPropIsExpired(Cert);
	  x509IsRoot := GetCertPropIsRoot(Cert);
	  if not (x509IsCorrupt or x509IsRevoked or x509IsExpired) then begin
	    CertKey := PGPPeekKeyDBObjKey(Cert);
	    if GetKeyPropCanEncrypt(CertKey) then x509CanEncryptData := (KeyUsage_dataEncipherment in x509KeyUsages);
	    if GetKeyPropCanSign(CertKey) then begin
	      x509CanSignData := (KeyUsage_digitalSignature in x509KeyUsages);
	      x509CanSignCerts := (KeyUsage_keyCertSign in x509KeyUsages);
	      x509CanSignCRLs := (KeyUsage_cRLSign in x509KeyUsages);
	    end;
	  end;
	end;
      finally
	CertASN1Parser.Free;
      end;
    end;
  except
    Result := -1;
  end;
end;

function GetKeyProps(KeySet: pPGPKeySet; Key: pPGPKey; Props, Filters: DWord; var KeyPropsRec: TKeyPropsRec): PGPError;
const
  F_X509CanEncryptOrSign = KeyFilterFlag_X509CanEncrypt or KeyFilterFlag_X509CanSign;
begin
  Result := 0;
  try
    with KeyPropsRec do begin
      // "RawByteString" properties
      if Props and spgpKeyPropFlag_KeyID <> 0 then begin
	kHexID := ShortHexID(GetKeyPropKeyID(Key));
      end;
      if Props and spgpKeyPropFlag_UserID <> 0 then kUserID := GetKeyPropUserID(Key);
      if Props and spgpKeyPropFlag_Fingerprint <> 0 then kFingerprint := GetKeyPropFingerprint(Key);
      if Props and spgpKeyPropFlag_CreationTimeStr <> 0 then kCreaTimeStr := GetKeyPropKeyCreationTimeStr(Key);
      if Props and spgpKeyPropFlag_ExpirationTimeStr <> 0 then kExpTimeStr := GetKeyPropKeyExpirationTimeStr(Key);
      // "number" properties
      if Props and spgpKeyPropFlag_KeyBits <> 0 then kSize := GetKeyPropKeyBits(Key);
      if Props and spgpKeyPropFlag_X509Cert <> 0 then kX509Cert := GetKeyPropX509Cert(Key);
      if Props and spgpKeyPropFlag_KeyAlg <> 0 then kKeyAlgorithm := GetKeyPropKeyAlg(Key);
      if Props and spgpKeyPropFlag_HashAlg <> 0 then kHashAlgorithm := GetKeyPropHashAlg(Key);
      if Props and spgpKeyPropFlag_Trust <> 0 then kTrust := GetKeyPropTrust(Key);
      if Props and spgpKeyPropFlag_Validity <> 0 then kValidity := GetKeyPropValidity(Key);
      if Props and spgpKeyPropFlag_CreationTime <> 0 then kCreaTimeNum := GetKeyPropKeyCreationTime(Key);
      if Props and spgpKeyPropFlag_ExpirationTime <> 0 then kExpTimeNum := GetKeyPropKeyExpirationTime(Key);
      // "boolean" properties
      if Props and spgpKeyPropFlag_IsCorrupt <> 0 then kCorrupt := GetKeyPropIsCorrupt(Key);
      if Props and spgpKeyPropFlag_IsSecret <> 0 then kPrivate := GetKeyPropIsSecret(Key);
      if Props and spgpKeyPropFlag_IsAxiomatic <> 0 then kImplicitTrust := GetKeyPropIsAxiomatic(Key);
      if Props and spgpKeyPropFlag_IsRevoked <> 0 then kRevoked := GetKeyPropIsRevoked(Key);
      if Props and spgpKeyPropFlag_IsDisabled <> 0 then kDisabled := GetKeyPropIsDisabled(Key);
      if Props and spgpKeyPropFlag_IsExpired <> 0 then kExpired := GetKeyPropIsExpired(Key);
      if Props and spgpKeyPropFlag_IsSecretShared <> 0 then kSecShared := GetKeyPropIsSecretShared(Key);
      if Props and spgpKeyPropFlag_CanEncrypt <> 0 then kCanEncrypt := GetKeyPropCanEncrypt(Key);
      if Props and spgpKeyPropFlag_CanDecrypt <> 0 then kCanDecrypt := GetKeyPropCanDecrypt(Key);
      if Props and spgpKeyPropFlag_CanSign <> 0 then kCanSign := GetKeyPropCanSign(Key);
      if Props and spgpKeyPropFlag_CanVerify <> 0 then kCanVerify := GetKeyPropCanVerify(Key);
      if Props and spgpKeyPropFlag_HasRevoker <> 0 then kHasRevoker := GetKeyPropHasRevoker(KeySet, Key, kRevKeyIDList);
      if Props and spgpKeyPropFlag_HasADK <> 0 then kHasADK := GetKeyPropHasARR(KeySet, Key, kADKeyIDList);
      if Props and spgpKeyPropFlag_HasSubKey <> 0 then kHasSubKey := GetKeyPropHasSubKey(KeySet, Key);
      if Props and spgpKeyPropFlag_LegacyKey <> 0 then kLegacyKey := GetKeyPropLegacy(Key);
      // "list" properties
      if (Props and spgpKeyPropFlag_IncludeSignerIDs <> 0) or (Filters and F_X509CanEncryptOrSign <> 0) then begin
	Props := Props or spgpKeyPropFlag_IncludeUserIDs;
      end;
      if (Props and (spgpKeyPropFlag_IncludeUserIDs) <> 0) or (kX509Cert > NoCert) then begin
	GetKeyPropUserIDs(Key, KeyPropsRec, Props, Filters);
      end;
    end;
  except
    Result := -1;
  end;
end;

function GetKeySetProps(Context: pPGPContext; KeySet: pPGPKeySet;
			var KeyPropsList: TKeyPropsList;
			PropertyFlags, FilterFlags: DWord;
			KeyOrder: TPGPKeyOrdering): Integer;
var
  KeyFilter	: pPGPFilter;
  KeySetFound	: pPGPKeySet;
  KeyCount	: PGPUInt32;
  KeyList	: pPGPKeyList;
  KeyIter	: pPGPKeyIter;
  KeyFound	: pPGPKey;
  KeyIndex	: Integer;
begin
  KeyFilter := NIL;
  KeySetFound := NIL;
  KeyList := NIL;
  KeyIter := NIL;
  if FilterFlags <> KeyFilterFlag_AllKeys then begin
    Result := GetKeyFilter(Context, FilterFlags, KeyFilter);
    try
      if Result <> 0 then Exit;
      Result := PGPFilterKeySet(KeySet, KeyFilter, KeySetFound);
      if Result <> 0 then Exit;
      KeySet := KeySetFound;
    finally
      PGPFreeFilter(KeyFilter);
    end;
  end
  else Result := 0;
  try
    if Result <> 0 then Exit;
    Result := PGPCountKeys(KeySet, KeyCount);
    if Result <> 0 then Exit;
    if PropertyFlags <> 0 then begin
      if KeyPropsList = NIL then KeyPropsList := TKeyPropsList.Create(PropertyFlags);
      if KeyPropsList = NIL then begin
	Result := kPGPError_OutOfMemory;
	Exit;
      end;
      if (KeyPropsList.Count = 0) and ((PropertyFlags and spgpKeyPropFlag_IncludeGroupsList) <> 0) then begin
	GetGroupsList(Context, KeySet, KeyPropsList.GroupsList);
      end;
      Result := PGPOrderKeySet(KeySet, PGPKeyOrdering(succ(Ord(KeyOrder) shr 1)), PGPBoolean(Odd(Ord(KeyOrder))), KeyList);
      if Result <> 0 then Exit;
      try
	Result := PGPNewKeyIter(KeyList, KeyIter);
	try
	  while PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_Key, KeyFound) = 0 do begin
	    KeyIndex := KeyPropsList.Add(GetKeyPropKeyID(KeyFound));
	    if KeyIndex = -1 then begin
	      Result := kPGPError_OutOfMemory;
	      Exit;
	    end;
	    Result := GetKeyProps(KeySet, KeyFound, PropertyFlags, FilterFlags,
	    			  pKeyPropsRec(KeyPropsList.Objects[KeyIndex])^);
	    if Result <> 0 then Exit;
	    with KeyPropsList, KeyProps[KeyIndex] do if kX509Key and (kX509Cert < StdCert) then Delete(KeyIndex);
	  end;
	  Result := KeyPropsList.Count;
	finally
	  PGPFreeKeyIter(KeyIter);
	end;
      finally
	PGPFreeKeyList(KeyList);
      end;
    end
    else Result := KeyCount;
  finally
    PGPFreeKeySet(KeySetFound);
  end;
end;

function AddKeysToKeyRing(Context: pPGPContext; KeySetMain, KeysToImport: pPGPKeySet;
			  var KeyPropsList: TKeyPropsList; PropertyFlags: DWord): Integer;
var
  KeySetToAdd	: pPGPKeySet;
begin
  if SDKVersion >= PGP7XX then begin
    KeySetToAdd := NIL;
    Result := PGPCopyKeys(KeysToImport, PGPPeekKeySetKeyDB(KeySetMain), KeySetToAdd);
  end
  else begin
    KeySetToAdd := KeysToImport;
    Result := PGPAddKeys(KeySetToAdd, KeySetMain);
  end;
  if Result <> 0 then Exit;
  try
    Result := KeyRings.UpdateKeyRings;
    if Result <> 0 then Exit;
    Result := GetKeySetProps(Context, KeySetToAdd, KeyPropsList,
			     PropertyFlags, KeyFilterFlag_AllKeys, UserID_Ordering);
  finally
    if SDKVersion >= PGP7XX then PGPFreeKeySet(KeySetToAdd);
  end;
end;

function GetHexIDByAnyID(Context: pPGPContext; KeySetMain: pPGPKeySet; const AnyID: RawByteString; var HexID: RawByteString): Integer;
var
  KeySetFound	: pPGPKeySet;
  KeyCount	: PGPUInt32;
  Key		: pPGPKey;
begin
  HexID := E;
  Result := 0;
  KeySetFound := NIL;
  if AnyID <> E then begin
    try
      Result := GetKeySetByAnyIDs(Context, KeySetMain, AnyID, KeySetFound);
      if Result <> 0 then Exit;
      Result := PGPCountKeys(KeySetFound, KeyCount);
      if Result <> 0 then Exit;
      if KeyCount > 0 then begin
	Result := GetKeyFromKeySet(KeySetFound, Key);
	if Result <> 0 then Exit;
	HexID := GetKeyPropKeyID(Key);
      end;
      Result := KeyCount;
    finally
      PGPFreeKeySet(KeySetFound);
    end;
  end;
end;

function CountUserIDs(KeySet: pPGPKeySet): Integer;
var
  KeyList	: pPGPKeyList;
  KeyIter	: pPGPKeyIter;
  Key		: pPGPKey;
  UserID	: pPGPUserID;
  UserIDProp	: PGPBoolean;
  UserIDCount	: PGPUInt32;
begin
  UserIDCount := 0;
  Result := GetKeyIterAnyOrdering(KeySet, KeyList, KeyIter);
  if Result <> 0 then Exit;
  try
    while PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_Key, Key) = 0 do begin
      while PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_UserID, UserID) = 0 do begin
	Result := PGPGetKeyDBObjBooleanProperty(UserID, kPGPUserIDProperty_IsAttribute, UserIDProp);
	if Result <> 0 then Exit;
	if UserIDProp = PGPFalse then inc(UserIDCount);
      end;
    end;
    Result := UserIDCount;
  finally
    PGPFreeKeyIter(KeyIter);
    PGPFreeKeyList(KeyList);
  end;
end;

function GetMatchingKey(Context: pPGPContext; KeySet: pPGPKeySet; Passphrase: PAnsiChar): pPGPKey;
var
  KeyList: pPGPKeyList;
  KeyIter: pPGPKeyIter;
  Key: pPGPKey;
begin
  Result := NIL;
  if GetKeyIterAnyOrdering(KeySet, KeyList, KeyIter) = 0 then begin
    try
      while PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_Key, Key) = 0 do begin
	if PGPPassphraseIsValid(Key, PGPOPassphrase(Context, Passphrase), PGPOLastOption(Context)) = PGPTrue then begin
	  Result := Key;
	  Break;
	end;
      end;
    finally
      PGPFreeKeyIter(KeyIter);
      PGPFreeKeyList(KeyList);
    end;
  end;
end;

function GetUtf8Passphrase(const Passphrase: PAnsiChar; var Utf8Passphrase: PAnsiChar): PGPError;
begin
  Result := kPGPError_OutOfMemory;
  Utf8Passphrase := PGPNewSecureData(PGPGetDefaultMemoryMgr, succ(MaxPassphraseLength) * UTF8Factor,
  				     kPGPMemoryMgrFlags_Clear);
  if Utf8Passphrase <> NIL then begin
    if SDKVersion >= PGP8XX then
      SecureAnsiToUtf8PChar(Passphrase, Utf8Passphrase, succ(MaxPassphraseLength) * UTF8Factor)
    else StrLCopy(Utf8Passphrase, Passphrase, succ(MaxPassphraseLength) * UTF8Factor);
    Result := kPGPError_NoErr;
  end;
end;

function PassphraseIsValid(Context: pPGPContext; KeySetMain: pPGPKeySet; AnyID, Passphrase: PAnsiChar): Longbool;
var
  KeySetFound	: pPGPKeySet;
  KeyCount	: PGPUInt32;
  Key		: pPGPKey;
begin
  Result := FALSE;
  KeySetFound := NIL;
  if GetKeySetByAnyIDs(Context, KeySetMain, AnyID, KeySetFound) = 0 then begin
    try
      Result := (PGPCountKeys(KeySetFound, KeyCount) = 0) and
		(KeyCount = 1) and (GetKeyFromKeySet(KeySetFound, Key) = 0) and
		(PGPPassphraseIsValid(Key, PGPOPassphrase(Context, Passphrase), PGPOLastOption(Context)) <> PGPFalse);
    finally
      PGPFreeKeySet(KeySetFound);
    end;
  end;
end;

function GetCachedPassphrase(Context: pPGPContext; var Key: pPGPKey; KeySet: pPGPKeySet; var Passphrase: PAnsiChar): Longbool;
var
  KeyIter: pPGPKeyIter;
begin
  Result := FALSE;
  if (SDKVersion >= PGP8XX) and ((Key <> NIL) or (KeySet <> NIL)) then begin
    Passphrase := PGPNewSecureData(PGPGetDefaultMemoryMgr, succ(MaxPassphraseLength) * UTF8Factor,
    				   kPGPMemoryMgrFlags_Clear);
    if Passphrase <> NIL then begin
      try
	if Key <> NIL then
	  Result := (PGPPassphraseIsValid(Key, PGPOLastOption(Context), PGPOLastOption(Context)) = PGPTrue)
	else begin
	  KeyIter := NIL;
	  try
	    if PGPNewKeyIterFromKeySet(KeySet, KeyIter) = 0 then begin
	      while PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_Key, Key) = 0 do begin
		Result := (PGPPassphraseIsValid(Key, PGPOLastOption(Context), PGPOLastOption(Context)) = PGPTrue);
		if Result then Break;
	      end;
	    end;
	  finally
	    PGPFreeKeyIter(KeyIter);
	  end;
	end;
      finally
	if not Result and (PGPFreeData(Passphrase) = kPGPError_NoErr) then Passphrase := NIL;
      end;
    end;
  end;
end;

function CachePassphrase(Context: pPGPContext; Key: pPGPKey; TimeInterval: PGPTime;
			 ShareCache: Longbool; var Passphrase: PAnsiChar): Longbool;
var
  OptionList: pPGPOptionList;
begin
  Result := FALSE;
  if (SDKVersion >= PGP8XX) and (Key <> NIL) and (Passphrase <> NIL) then begin
    OptionList := NIL;
    if PGPBuildOptionList(Context, OptionList,
      [
	PGPOPassphrase(Context, Passphrase),
        PGPOCachePassphrase(Context, TimeInterval, PGPBoolean(ShareCache = TRUE))
      ]
    ) = 0 then begin
      try
	Result := (PGPPassphraseIsValid(Key, OptionList, PGPOLastOption(Context)) = PGPTrue) and
		  (PGPCacheKeyDB(PGPPeekKeyDBObjKeyDB(Key), TimeInterval) = 0);
      finally
	if (PGPFreeData(Passphrase) = kPGPError_NoErr) then Passphrase := NIL;
	PGPFreeOptionList(OptionList);
      end;
    end;
  end;
end;

function IsPassphraseCacheEmpty(Context: pPGPContext): Longbool;
var
  Global, Local, Other, Persistent: PGPUInt32;
begin
  Result := (SDKVersion >= PGP8XX) and (PGPCountCachedPassphrases(Context, Global, Local, Other, Persistent) = 0) and
  	    (Global + Local + Other + Persistent = 0);
end;

function PurgePassphraseCache(Context: pPGPContext): Longbool;
begin
  Result := (PGPPurgePassphraseCache(Context) = 0) and (PGPPurgeKeyDBCache(Context) = 0);
end;

const
  SC_MANAGER_CONNECT		= $0001;
  SC_MANAGER_CREATE_SERVICE	= $0002;
  SC_MANAGER_ENUMERATE_SERVICE	= $0004;
  SC_MANAGER_LOCK		= $0008;
  SC_MANAGER_QUERY_LOCK_STATUS	= $0010;
  SC_MANAGER_MODIFY_BOOT_CONFIG	= $0020;

  SERVICE_QUERY_CONFIG		= $0001;
  SERVICE_CHANGE_CONFIG		= $0002;
  SERVICE_QUERY_STATUS		= $0004;
  SERVICE_ENUMERATE_DEPENDENTS	= $0008;
  SERVICE_START			= $0010;
  SERVICE_STOP			= $0020;
  SERVICE_PAUSE_CONTINUE	= $0040;
  SERVICE_INTERROGATE		= $0080;
  SERVICE_USER_DEFINED_CONTROL	= $0100;

  SERVICE_STOPPED		= $00000001;
  SERVICE_START_PENDING		= $00000002;
  SERVICE_STOP_PENDING		= $00000003;
  SERVICE_RUNNING		= $00000004;
  SERVICE_CONTINUE_PENDING	= $00000005;
  SERVICE_PAUSE_PENDING		= $00000006;
  SERVICE_PAUSED		= $00000007;

type
  SC_HANDLE = THandle;
  PServiceStatus = ^TServiceStatus;
  TServiceStatus = Record
    dwServiceType: DWORD;
    dwCurrentState: DWORD;
    dwControlsAccepted: DWORD;
    dwWin32ExitCode: DWORD;
    dwServiceSpecificExitCode: DWORD;
    dwCheckPoint: DWORD;
    dwWaitHint: DWORD;
  end;

var
  OpenSCManager: function(lpMachineName, lpDatabaseName: PAnsiChar; dwDesiredAccess: DWORD): SC_HANDLE; stdcall;
  OpenService: function(hSCManager: SC_HANDLE; lpServiceName: PAnsiChar; dwDesiredAccess: DWORD): SC_HANDLE; stdcall;
  StartService: function(hService: SC_HANDLE; dwNumServiceArgs: DWORD; var lpServiceArgVectors: PAnsiChar): BOOL; stdcall;
  QueryServiceStatus: function(hService: SC_HANDLE; var lpServiceStatus: TServiceStatus): BOOL; stdcall;
  CloseServiceHandle: function(hSCObject: SC_HANDLE): BOOL; stdcall;

function LoadServiceApiFunctions: Longbool;
var
  hAdvapi32: THandle;
begin
  Result := FALSE;
  hAdvapi32 := GetModuleHandleA('ADVAPI32.DLL');
  if hAdvapi32 <> 0 then begin
    OpenSCManager := GetProcAddress(hAdvapi32, 'OpenSCManagerA');
    OpenService := GetProcAddress(hAdvapi32, 'OpenServiceA');
    StartService := GetProcAddress(hAdvapi32, 'StartServiceA');
    QueryServiceStatus := GetProcAddress(hAdvapi32, 'QueryServiceStatus');
    CloseServiceHandle := GetProcAddress(hAdvapi32, 'CloseServiceHandle');
    Result := Assigned(OpenSCManager);
  end;
end;

function StartPGPsdkServ: Longbool;
var
  viOS: TOSVersionInfoA;
  hSCManager: THandle;
  hService: THandle;
  ssBuffer: TServiceStatus;
  savBuffer: PAnsiChar;
  dwWaitTime: DWord;
  dwStartTickCount: DWord;
begin
  Result := FALSE;
  viOS.dwOSVersionInfoSize := SizeOf(TOSVersionInfoA);
  if GetVersionExA(viOS) and (viOS.dwPlatformId = VER_PLATFORM_WIN32_NT) and LoadServiceApiFunctions then begin
    // WinNT, Win2K, WinXP
    hSCManager := OpenSCManager(NIL, NIL, GENERIC_EXECUTE);
    if hSCManager <> 0 then begin
      try
	hService := OpenService(hSCManager, 'PGPSERV', SERVICE_QUERY_STATUS or SERVICE_START); // PGP 8.1x & 9.0x
	if hService = 0 then OpenService(hSCManager, 'PGPSDKSERV', SERVICE_QUERY_STATUS or SERVICE_START); // PGP 8.0x
	if hService <> 0 then begin
	  try
	    if not QueryServiceStatus(hService, ssBuffer) or (ssBuffer.dwCurrentState <> SERVICE_RUNNING) then begin
	      savBuffer := NIL;
	      if StartService(hService, 0, savBuffer) then begin
		dwStartTickCount := 0;
		repeat
		  if QueryServiceStatus(hService, ssBuffer) then with ssBuffer do begin
		    if dwStartTickCount = 0 then dwStartTickCount := GetTickCount;
		    dwWaitTime := dwWaitHint div 10;
		    Sleep(dwWaitTime);
		  end
		  else Exit;
		until (ssBuffer.dwCurrentState = SERVICE_RUNNING)
		or ((GetTickCount - dwStartTickCount) > ssBuffer.dwWaitHint);
		Result := (ssBuffer.dwCurrentState = SERVICE_RUNNING);
	      end;
	    end;
	  finally
	    CloseServiceHandle(hService);
	  end;
	end;
      finally
	CloseServiceHandle(hSCManager);
      end;
    end;
  end
  else begin
    // Win95, Win98, WinME
    Result := (ShellExecuteA(0, NIL, 'PGPSERV.EXE', NIL, NIL, SW_SHOWDEFAULT) > 32); // PGP 8.1x & 9.0x
    if not Result then Result := (ShellExecuteA(0, NIL, 'PGPSDKSERV.EXE', NIL, NIL, SW_SHOWDEFAULT) > 32); // PGP 8.0x
  end;
  ProcessMessages;
  Sleep(0);
end;

destructor TKeyRings.Destroy;
begin
  if InitCount > 0 then begin
    InitCount := 1;
    FreeKeyRings;
  end;
  inherited Destroy;
end;

function TKeyRings.UpdateKeyRings: PGPError;
begin
  Result := kPGPError_NoErr;
  if SDKVersion >= PGP7XX then
    Result := PGPFlushKeyDB(RingKeyDB)
  else if PGPKeySetNeedsCommit(RingKeySet) = PGPTrue then Result := PGPCommitKeyringChanges(RingKeySet);
  if Result = 0 then PGPclNotifyKeyringChanges(GetCurrentProcessID);
end;

function TKeyRings.SetKeyRings(const Pubring, Secring: RawByteString): PGPError;
begin
  if ((Pubring <> E) and (Secring <> E)) then begin
    Result := kPGPError_NoErr;
    SetPubringFile(Pubring);
    SetSecringFile(Secring);
  end
  else Result := kPGPError_ImproperInitialization
end;

function TKeyRings.InitKeyRings(var Context: pPGPContext; var KeySetMain: pPGPKeySet): PGPError;
var
  Prefs: TPreferenceRec;
  PubFileSpec: pPGPFileSpec;
  SecFileSpec: pPGPFileSpec;
  OldInitCount: Longint;
begin
  Result := kPGPError_ImproperInitialization;
  if PGPInitErrorCode = ieNone then begin
    if InitCount <= 0 then begin
      InitCount := 0;
      RingKeyDB := NIL;
      RingKeySet := NIL;
      RingContext := NIL;
      PubFileSpec := NIL;
      SecFileSpec := NIL;
      Result := PGPNewContext(kPGPsdkAPIVersion, RingContext);
      if Result <> 0 then Exit;
      try
	if (FPubringFile = E) or (FSecringFile = E) then begin
	  GetPreferences(Prefs, PrefsFlag_PublicKeyring or PrefsFlag_PrivateKeyring or PrefsFlag_GroupsFile);
	  SetKeyRings(Prefs.PublicKeyring, Prefs.PrivateKeyring);
	  if GroupsFile = E then GroupsFile := Prefs.GroupsFile;
	end
	else if (GroupsFile = E) and (GetPreferences(Prefs, PrefsFlag_GroupsFile) = 0) then begin
	  GroupsFile := Prefs.GroupsFile;
	end;
	try
	  Result := PGPNewFileSpecFromFullPath(RingContext, PAnsiChar(FPubringFile), PubFileSpec);
	  if Result <> 0 then Exit;
	  Result := PGPNewFileSpecFromFullPath(RingContext, PAnsiChar(FSecringFile), SecFileSpec);
	  if Result <> 0 then Exit;
	  if SDKVersion >= PGP7XX then begin
	    Result := PGPOpenKeyDBFile(RingContext, kPGPOpenKeyDBFileOptions_Mutable, PubFileSpec, SecFileSpec, RingKeyDB);
	    if Result = 0 then RingKeySet := PGPPeekKeyDBRootKeySet(RingKeyDB);
	  end
	  else Result := PGPOpenKeyRingPair(RingContext, kPGPKeyRingOpenFlags_Mutable, PubFileSpec, SecFileSpec, RingKeySet);
	finally
	  PGPFreeFileSpec(PubFileSpec);
	  PGPFreeFileSpec(SecFileSpec);
	end;
      finally
	if (Result <> 0) or (RingKeySet = NIL) then begin
	  PGPFreeContext(RingContext);
	  RingContext := NIL;
	  if SDKVersion >= PGP7XX then begin
	    PGPFreeKeyDB(RingKeyDB);
	    RingKeyDB := NIL;
	  end
	  else begin
	    PGPFreeKeySet(RingKeySet);
	    RingKeySet := NIL;
	  end;
	end;
      end;
    end
    else if SDKVersion >= PGP7XX then begin
      Result := PGPsdkReconnect;
      // prevents PGP 8.1 problems
      if Result > 0 then Result := 0;
    end
    else Result := 0;
    if Result = 0 then inc(InitCount);
    KeySetMain := RingKeySet;
    Context := RingContext;
    if (SDKVersion >= PGP7XX) and not Recursion and (Result = kPGPError_RPCFailed) then begin
      if InitCount > 0 then begin
	OldInitCount := InitCount;
	InitCount := 1;
	FreeKeyRings;
      end
      else OldInitCount := 1;
      PGPsdkCleanup;
      Recursion := TRUE;
      if StartPGPsdkServ then Result := PGPsdkInit(InitFlags);
      if Result = 0 then Result := InitKeyRings(Context, KeySetMain);
      if Result = 0 then InitCount := OldInitCount;
    end
    else Recursion := FALSE;
  end;
end;

procedure TKeyRings.FreeKeyRings;
begin
  try
    if InitCount = 1 then begin
      try
	if SDKVersion >= PGP7XX then
	  PGPFreeKeyDB(RingKeyDB)
	else PGPFreeKeySet(RingKeySet);
      finally
	RingKeySet := NIL;
	RingKeyDB := NIL;
	try
	  PGPFreeContext(RingContext);
	finally
	  RingContext := NIL;
	end;
      end;
    end;
  finally
    if InitCount > 0 then dec(InitCount);
  end;
end;

function TKeyRings.GetPubringFile: RawByteString;
begin
  if SDKVersion >= PGP9XX then
    Result := Utf8ToAnsi(FPubringFile)
  else Result := FPubringFile;
end;

function TKeyRings.GetSecringFile: RawByteString;
begin
  if SDKVersion >= PGP9XX then
    Result := Utf8ToAnsi(FSecringFile)
  else Result := FSecringFile;
end;

procedure TKeyRings.SetPubringFile(const Value: RawByteString);
begin
  if Value <> FPubringFile then begin
    if SDKVersion >= PGP9XX then
      FPubringFile := AnsiToUtf8(Value)
    else FPubringFile := Value;
  end;
end;

procedure TKeyRings.SetSecringFile(const Value: RawByteString);
begin
  if Value <> FSecringFile then begin
    if SDKVersion >= PGP9XX then
      FSecringFile := AnsiToUtf8(Value)
    else FSecringFile := Value;
  end;
end;


initialization

  KeyRings := TKeyRings.Create;


finalization

  KeyRings.Free;
  KeyRings := NIL;


end.

