{------------------------------------------------------------------------------}
{                                                                              }
{                     This code is Copyright (c) 1999-2010                     }
{                           by Michael in der Wiesche                          }
{                                                                              }
{------------------------------------------------------------------------------}
{$A+,B-,C-,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z4}
unit PGP2Comp;							// Version 5.0.0

interface

uses
  Windows, Classes, Forms, StdCtrls, SysUtils, PGPStrTools, PGPStringList, UTF8, KeyPropTypes, KeyFuncs, PrefFuncs,
  pgpBase, pgpErrors, pgpPubTypes, pgpUtilities, pgpOptionList, pgpMemoryMgr, pgpEvents, pgpKeys,
  pgpKeyServer, pgpTLS, pgpUI, pgpCL, pgpSC, PGPDialogs, PGPKeyGenerate, PGPEncode, PGPDecode;


// Register --------------------------------------------------------------------

procedure Register;


// PGP Initialization Error ----------------------------------------------------

type
  TPGPInitErrorCode = pgpBase.TPGPInitErrorCode;

const
  PGPInitErrorCode: TPGPInitErrorCode = ieNone;


// Notification Messages -------------------------------------------------------

const
  WM_PGP_ReloadPrefs: UInt = 0;
  WM_PGP_ReloadKeyring: UInt = 0;
  WM_PGP_ReloadKeyserverPrefs: UInt = 0;


// Common Constants ------------------------------------------------------------

const
  MaxPassphraseLength = pgpBase.MaxPassphraseLength;


// Common types ----------------------------------------------------------------

type

  // Preferences ---------------------------------------------------------------

  TPGPPref = (
    Pref_DefaultKeyID,
    Pref_PublicKeyring,
    Pref_PrivateKeyring,
    Pref_RandomSeedFile,
    Pref_GroupsFile
  );

  TPGPPrefs = Set of TPGPPref;

  TPrefsPage = (
    PrefsPage_GeneralPrefs,
    PrefsPage_KeyringPrefs,
    PrefsPage_EmailPrefs,
    PrefsPage_HotkeyPrefs,
    PrefsPage_KeyserverPrefs,
    PrefsPage_CAPrefs,
    PrefsPage_AdvancedPrefs
  );

  {$IFNDEF _PGP2COMP_BCPPB_}
  // BCPPB has ambiguity issues in method declarations if these are included
  TPreferenceRec = PrefFuncs.TPreferenceRec;
  {$ENDIF}

  // Key properties ------------------------------------------------------------

  {$IFNDEF _PGP2COMP_BCPPB_}
  // BCPPB has ambiguity issues in method declarations if these are included
  TKeyPropsList = KeyPropTypes.TKeyPropsList;
  TKeyPropsRec = KeyPropTypes.TKeyPropsRec;
  PKeyPropsRec = KeyPropTypes.PKeyPropsRec;
  TKeyProp = KeyPropTypes.TKeyProp;
  TKeyProps = KeyPropTypes.TKeyProps;
  {$ENDIF}

  TAlgorithmKeyFilter = (
    AlgorithmFilter_AllKeys,
    AlgorithmFilter_DHDSS,
    AlgorithmFilter_RSA,
    AlgorithmFilter_RSALegacy
  );

  TBooleanKeyFilter = (
    BoolFilter_AllKeys,
    BoolFilter_CanEncrypt,
    BoolFilter_CanDecrypt,
    BoolFilter_CanSign,
    BoolFilter_CanVerify
  );

  TCertKeyFilter = (
    CertFilter_AllKeys,
    CertFilter_X509Cert,
    CertFilter_X509Root,
    CertFilter_X509CanEncrypt,
    CertFilter_X509CanSign
  );

  TKeyOrdering = (
    Any_Order,
    UserID_Order,
    ReverseUserID_Order,
    KeyID_Order,
    ReverseKeyID_Order,
    Validity_Order,
    ReverseValidity_Order,
    Trust_Order,
    ReverseTrust_Order,
    EncryptKeySize_Order,
    ReverseEncryptKeySize_Order,
    SigKeySize_Order,
    ReverseSigKeySize_Order,
    Creation_Order,
    ReverseCreation_Order,
    Expiration_Order,
    ReverseExpiration_Order
  );

  {$IFNDEF _PGP2COMP_BCPPB_}
  // BCPPB has ambiguity issues in method declarations if these are included
  TTrustLevel = KeyPropTypes.TTrustLevel;
  TValidityLevel = KeyPropTypes.TValidityLevel;
  TKeyAlgorithm = KeyPropTypes.TKeyAlgorithm;
  TCipherAlgorithm = KeyPropTypes.TCipherAlgorithm;
  TADKType = KeyPropTypes.TADKType;
  {$ENDIF}

  // Key server ----------------------------------------------------------------

  TServerType = (
    KeyServerType_Invalid,
    KeyServerType_LDAP,
    KeyServerType_HTTP,
    KeyServerType_LDAPS,
    KeyServerType_HTTPS
  );

  TServerFlag = (
    KeyServerListed,
    KeyServerIsRoot
  );

  TServerFlags = Set of TServerFlag;

  // Key import ----------------------------------------------------------------

  TIgnoreKnownKeys = Set of (Ignore_ByHexID, Ignore_ByUserID);

  // Key generation ------------------------------------------------------------

  {$IFNDEF _PGP2COMP_BCPPB_}
  // BCPPB has ambiguity issues in method declarations if these are included
  TKeySize = PGPKeyGenerate.TKeySize;
  TSubKeySize = PGPKeyGenerate.TSubKeySize;
  TMinPassLen = PGPKeyGenerate.TMinPassLen;
  TMinPassQual = PGPKeyGenerate.TMinPassQual;
  {$ENDIF}

  // Encoding ------------------------------------------------------------------

  {$IFNDEF _PGP2COMP_BCPPB_}
  // BCPPB has ambiguity issues in method declarations if these are included
  TSignAlgorithm = PGPEncode.TSignAlgorithm;
  TConventionalAlgorithm = PGPEncode.TConventionalAlgorithm;
  {$ENDIF}

  // Decoding ------------------------------------------------------------------

  {$IFNDEF _PGP2COMP_BCPPB_}
  // BCPPB has ambiguity issues in method declarations if these are included
  TSigStatus = PGPDecode.TSigStatus;
  TSigPropsRec = PGPDecode.TSigPropsRec;
  {$ENDIF}

  // Encoding & Decoding -------------------------------------------------------
  TFileOutput = (FileOutput_Disabled, FileOutput_DefaultQuery, FileOutput_AlwaysQuery);

// Common Constants ------------------------------------------------------------

const
  KeyProp_HexID = KeyPropTypes.KeyProp_HexID;
  KeyProp_UserID = KeyPropTypes.KeyProp_UserID;
  KeyProp_Fingerprint = KeyPropTypes.KeyProp_Fingerprint;
  KeyProp_CreaTimeStr = KeyPropTypes.KeyProp_CreaTimeStr;
  KeyProp_ExpTimeStr = KeyPropTypes.KeyProp_ExpTimeStr;
  KeyProp_Size = KeyPropTypes.KeyProp_Size;
  KeyProp_X509Cert = KeyPropTypes.KeyProp_X509Cert;
  KeyProp_KeyAlgorithm = KeyPropTypes.KeyProp_KeyAlgorithm;
  KeyProp_HashAlgorithm = KeyPropTypes.KeyProp_HashAlgorithm;
  KeyProp_Trust = KeyPropTypes.KeyProp_Trust;
  KeyProp_Validity = KeyPropTypes.KeyProp_Validity;
  KeyProp_CreaTimeNum = KeyPropTypes.KeyProp_CreaTimeNum;
  KeyProp_ExpTimeNum = KeyPropTypes.KeyProp_ExpTimeNum;
  KeyProp_Corrupt = KeyPropTypes.KeyProp_Corrupt;
  KeyProp_Secret = KeyPropTypes.KeyProp_Secret;
  KeyProp_ImplicitTrust = KeyPropTypes.KeyProp_ImplicitTrust;
  KeyProp_Revoked = KeyPropTypes.KeyProp_Revoked;
  KeyProp_Disabled = KeyPropTypes.KeyProp_Disabled;
  KeyProp_Expired = KeyPropTypes.KeyProp_Expired;
  KeyProp_SecretShared = KeyPropTypes.KeyProp_SecretShared;
  KeyProp_CanEncrypt = KeyPropTypes.KeyProp_CanEncrypt;
  KeyProp_CanDecrypt = KeyPropTypes.KeyProp_CanDecrypt;
  KeyProp_CanSign = KeyPropTypes.KeyProp_CanSign;
  KeyProp_CanVerify = KeyPropTypes.KeyProp_CanVerify;
  KeyProp_HasRevoker = KeyPropTypes.KeyProp_HasRevoker;
  KeyProp_HasADK = KeyPropTypes.KeyProp_HasADK;
  KeyProp_HasSubKey = KeyPropTypes.KeyProp_HasSubKey;
  KeyProp_LegacyKey = KeyPropTypes.KeyProp_LegacyKey;
  KeyProp_IncludeUserIDs = KeyPropTypes.KeyProp_IncludeUserIDs;
  KeyProp_IncludeSignerIDs = KeyPropTypes.KeyProp_IncludeSignerIDs;
  KeyProp_IncludeGroupsList = KeyPropTypes.KeyProp_IncludeGroupsList;

  KeyProps_IDFlags = [KeyProp_HexID, KeyProp_UserID];
  KeyProps_IDComplete = KeyProps_IDFlags + [KeyProp_Size, KeyProp_KeyAlgorithm];


// Common Procs ----------------------------------------------------------------

type
  TOnFailure = procedure(ErrorCode: Longint; const ErrorMsg: String) of Object;


// TPGPPreferences -------------------------------------------------------------

type
  TOnGetPreferences = procedure(const Preferences: TPreferenceRec) of Object;
  TOnSetPreferences = procedure(var Preferences: TPreferenceRec) of Object;

  TPGPPreferences = class(TComponent)
  private
    FAltPubringFile: RawByteString;
    FAltSecringFile: RawByteString;
    FAltGroupsFile: RawByteString;
    FParentHandle: THandle;
    FPassCacheSeconds: TPassCacheSeconds;
    FPassCacheShare: Longbool;
    FPreferences: TPreferenceRec;
    FPGPPrefs: TPGPPrefs;
    FPrefsPage: TPrefsPage;
    FOnFailure: TOnFailure;
    FOnGetPreferences: TOnGetPreferences;
    FOnSetPreferences: TOnSetPreferences;
    function GetFlags: Longint;
    function PrefsError: Longbool;
    function GetAltPubKeyring: AnsiString;
    function GetAltPrivKeyring: AnsiString;
    function GetAltGroupsFile: AnsiString;
    function GetPGPVersion: AnsiString;
    function GetVersionString: AnsiString;
    procedure SetAltPubKeyring(const Value: AnsiString);
    procedure SetAltPrivKeyring(const Value: AnsiString);
    procedure SetAltGroupsFile(const Value: AnsiString);
    procedure SetVersionString(const Value: AnsiString);
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function ShowPrefsDialog: Longint;
    function DoGetPreferences: Longint;
    function DoSetPreferences: Longint;
    property ParentHandle: THandle
      read FParentHandle
      write FParentHandle;
    property PassCacheSeconds: TPassCacheSeconds
      read FPassCacheSeconds
      write FPassCacheSeconds
      stored FALSE;
    property PassCacheShare: Longbool
      read FPassCacheShare
      write FPassCacheShare
      stored FALSE;
    property PGPVersion: AnsiString
      read GetPGPVersion
      stored FALSE;
  published
    property AltPubKeyring: AnsiString
      read GetAltPubKeyring
      write SetAltPubKeyring;
    property AltPrivKeyring: AnsiString
      read GetAltPrivKeyring
      write SetAltPrivKeyring;
    property AltGroupsFile: AnsiString
      read GetAltGroupsFile
      write SetAltGroupsFile;
    property PrefsPage: TPrefsPage
      read FPrefsPage
      write FPrefsPage;
    property DefaultKeyHexID: AnsiString
      read FPreferences.DefaultKeyHexID
      write FPreferences.DefaultKeyHexID;
    property PublicKeyring: AnsiString
      read FPreferences.PublicKeyring
      write FPreferences.PublicKeyring;
    property PrivateKeyring: AnsiString
      read FPreferences.PrivateKeyring
      write FPreferences.PrivateKeyring;
    property RandomSeedFile: AnsiString
      read FPreferences.RandomSeedFile
      write FPreferences.RandomSeedFile;
    property GroupsFile: AnsiString
      read FPreferences.GroupsFile
      write FPreferences.GroupsFile;
    property PGPPrefs: TPGPPrefs
      read FPGPPrefs
      write FPGPPrefs;
    property VersionString: AnsiString
      read GetVersionString
      write SetVersionString;
    property OnFailure: TOnFailure
      read FOnFailure
      write FOnFailure;
    property OnGetPreferences: TOnGetPreferences
      read FOnGetPreferences
      write FOnGetPreferences;
    property OnSetPreferences: TOnSetPreferences
      read FOnSetPreferences
      write FOnSetPreferences;
  end;


// TPGPKeyServer ---------------------------------------------------------------

type
  TServerEntry = Record
    ServerProtocol: TServerType;
    ServerDNS: AnsiString;
    ServerPort: Word;
    ServerFlags: TServerFlags;
  end;

  TOnGetDefaultServer = procedure(const ServerEntry: TServerEntry) of Object;
  TOnSetDefaultServer = procedure(var ServerEntry: TServerEntry) of Object;
  TOnServerResults = procedure(const KeyPropsList: TKeyPropsList) of Object;

  TPGPKeyServer = class(TComponent)
  private
    FParentHandle: THandle;
    FKeyDlgPrompt: AnsiString;
    FKeyProps: TKeyProps;
    FServerEntry: TServerEntry;
    FOnFailure: TOnFailure;
    FOnGetDefaultServer: TOnGetDefaultServer;
    FOnSetDefaultServer: TOnSetDefaultServer;
    FOnServerResults: TOnServerResults;
    procedure SetKeyProps(Value: TKeyProps);
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function DoGetKeysFromServer(const KeyData: AnsiString): Longint;
    function DoSendKeysToServer: Longint;
    function DoGetDefaultServer: Longint;
    function DoSetDefaultServer: Longint;
    property ParentHandle: THandle
      read FParentHandle
      write FParentHandle;
  published
    property KeyDlgPrompt: AnsiString
      read FKeyDlgPrompt
      write FKeyDlgPrompt;
    property KeyProps: TKeyProps
      read FKeyProps
      write SetKeyProps;
    property ServerProtocol: TServerType
      read FServerEntry.ServerProtocol
      write FServerEntry.ServerProtocol;
    property ServerDNS: AnsiString
      read FServerEntry.ServerDNS
      write FServerEntry.ServerDNS;
    property ServerPort: Word
      read FServerEntry.ServerPort
      write FServerEntry.ServerPort;
    property ServerFlags: TServerFlags
      read FServerEntry.ServerFlags
      write FServerEntry.ServerFlags;
    property OnFailure: TOnFailure
      read FOnFailure
      write FOnFailure;
    property OnGetDefaultServer: TOnGetDefaultServer
      read FOnGetDefaultServer
      write FOnGetDefaultServer;
    property OnSetDefaultServer: TOnSetDefaultServer
      read FOnSetDefaultServer
      write FOnSetDefaultServer;
    property OnServerResults: TOnServerResults
      read FOnServerResults
      write FOnServerResults;
  end;


// TPGPGetKeyProps -------------------------------------------------------------

type
  TKeyRingProps = class(TComponent)
  private
    FComboBox: TComboBox;
    FListBox: TListBox;
    FPrimaryIDs: TRawByteStrings;
    FRingAlgorithmFilter: TAlgorithmKeyFilter;
    FRingBoolFilter: TBooleanKeyFilter;
    FRingKeyOrdering: TKeyOrdering;
    FRingProps: TKeyProps;
    FRingPropsList: TKeyPropsList;
    FOnFailure: TOnFailure;
    function GetPrimaryIDs: Longint;
    procedure SetRingProps(Value: TKeyProps);
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Update: Longint;
    property RingPropsList: TKeyPropsList
      read FRingPropsList;
  published
    property RingAlgorithmFilter: TAlgorithmKeyFilter
      read FRingAlgorithmFilter
      write FRingAlgorithmFilter;
    property RingBoolFilter: TBooleanKeyFilter
      read FRingBoolFilter
      write FRingBoolFilter;
    property RingKeyOrdering: TKeyOrdering
      read FRingKeyOrdering
      write FRingKeyOrdering;
    property RingProps: TKeyProps
      read FRingProps
      write SetRingProps;
    property ComboBox: TComboBox
      read FComboBox
      write FComboBox;
    property ListBox: TListBox
      read FListBox
      write FListBox;
    property OnFailure: TOnFailure
      read FOnFailure
      write FOnFailure;
  end;

  TOnGetKeyID = procedure(var KeyID: AnsiString) of Object;
  TOnGetKeyProps = procedure(const KeyPropsList: TKeyPropsList) of Object;

  TPGPGetKeyProps = class(TKeyRingProps)
  private
    FAlgorithmFilter: TAlgorithmKeyFilter;
    FBoolFilter: TBooleanKeyFilter;
    FKeyOrdering: TKeyOrdering;
    FKeyID: AnsiString;
    FKeyProps: TKeyProps;
    FOnGetKeyID: TOnGetKeyID;
    FOnGetKeyProps: TOnGetKeyProps;
    procedure SetKeyProps(Value: TKeyProps);
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function DoGetKeyProps: Longint;
    function KeyIsOnRing: Longint;
    function KeyRingCount: Longint;
    function GetKeyProp(KeyPropsList: TKeyPropsList; KeyIndex: Longint; KeyProp: TKeyProp): Variant;
    function GetUserIDList(KeyPropsList: TKeyPropsList; KeyIndex: Longint): TRawByteStrings;
    function GetSignerIDList(KeyPropsList: TKeyPropsList; KeyIndex: Longint): TRawByteStrings;
    function GetRevKeyIDList(KeyPropsList: TKeyPropsList; KeyIndex: Longint): TRawByteStrings;
    function GetADKeyIDList(KeyPropsList: TKeyPropsList; KeyIndex: Longint): TRawByteStrings;
    function GetGroupHexIDs(KeyPropsList: TKeyPropsList; GroupIndex: Longint; var GroupName: AnsiString): AnsiString;
  published
    property AlgorithmFilter: TAlgorithmKeyFilter
      read FAlgorithmFilter
      write FAlgorithmFilter;
    property BoolFilter: TBooleanKeyFilter
      read FBoolFilter
      write FBoolFilter;
    property KeyOrdering: TKeyOrdering
      read FKeyOrdering
      write FKeyOrdering;
    property KeyID: AnsiString
      read FKeyID
      write FKeyID;
    property KeyProps: TKeyProps
      read FKeyProps
      write SetKeyProps;
    property OnGetKeyID: TOnGetKeyID
      read FOnGetKeyID
      write FOnGetKeyID;
    property OnGetKeyProps: TOnGetKeyProps
      read FOnGetKeyProps
      write FOnGetKeyProps;
  end;


// TPGPSetKeyProps -------------------------------------------------------------

type
  TOnKeySelected = function(const KeyProps: TKeyPropsRec): Longbool of Object;
  TOnEnterPassphrase = procedure(const Passphrase: PAnsiChar; var Cancel: Longbool) of Object;
  TOnChangePassphrase = procedure(const OldPassphrase, NewPassphrase: PAnsiChar;
				  MinPassLen: TMinPassLen;
				  MinPassQual: TMinPassQual;
				  var Cancel: Longbool) of Object;

  TPGPSetKeyProps = class(TComponent)
  private
    FParentHandle: THandle;
    FKeyDlgPrompt: AnsiString;
    FKeyHexID: AnsiString;
    FKeyProps: TKeyProps;
    FMinPassLen: TMinPassLen;
    FMinPassQual: TMinPassQual;
    FPassDlgPrompt: AnsiString;
    FPassDlgNewPrompt: AnsiString;
    FPassDlgOldPrompt: AnsiString;
    FOnFailure: TOnFailure;
    FOnKeySelected: TOnKeySelected;
    FOnEnterPassphrase: TOnEnterPassphrase;
    FOnChangePassphrase: TOnChangePassphrase;
    function GetKeyHexID(FilterFlags: DWord): Longint;
    procedure SetKeyProps(Value: TKeyProps);
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function DoKeyDisable: Longint;
    function DoKeyEnable: Longint;
    function DoKeyRemove: Longint;
    function DoKeyRevoke: Longint;
    function DoKeyPassChange: Longint;
    property ParentHandle: THandle
      read FParentHandle
      write FParentHandle;
  published
    property KeyID: AnsiString
      read FKeyHexID
      write FKeyHexID;
    property KeyProps: TKeyProps
      read FKeyProps
      write SetKeyProps;
    property KeyDlgPrompt: AnsiString
      read FKeyDlgPrompt
      write FKeyDlgPrompt;
    property PassDlgPrompt: AnsiString
      read FPassDlgPrompt
      write FPassDlgPrompt;
    property PassDlgNewPrompt: AnsiString
      read FPassDlgNewPrompt
      write FPassDlgNewPrompt;
    property PassDlgOldPrompt: AnsiString
      read FPassDlgOldPrompt
      write FPassDlgOldPrompt;
    property MinPassLen: TMinPassLen
      read FMinPassLen
      write FMinPassLen;
    property MinPassQual: TMinPassQual
      read FMinPassQual
      write FMinPassQual;
    property OnFailure: TOnFailure
      read FOnFailure
      write FOnFailure;
    property OnKeySelected: TOnKeySelected
      read FOnKeySelected
      write FOnKeySelected;
    property OnEnterPassphrase: TOnEnterPassphrase
      read FOnEnterPassphrase
      write FOnEnterPassphrase;
    property OnChangePassphrase: TOnChangePassphrase
      read FOnChangePassphrase
      write FOnChangePassphrase;
  end;


// TPGPKeysGenerate ------------------------------------------------------------

type
  TPGPKeysGenerate = class(TPGPKeysGenerateCustom)
  private
    FOnFailure: TOnFailure;
  protected
  public
    constructor Create(AOwner: TComponent); override;
    function DHDSSKeyGenerate: Longint; override;
    function DHSubKeyGenerate: Longint; override;
    function DSAKeyGenerate: Longint; override;
    function RSASubKeyGenerate: Longint; override;
    function RSAKeyGenerate(Legacy: Longbool): Longint; override;
  published
    property OnFailure: TOnFailure
      read FOnFailure
      write FOnFailure;
  end;


// TPGPKeyImport ---------------------------------------------------------------

type
  TOnGetFileIn = procedure(var FileIn: AnsiString) of Object;
  TOnKeyImported = procedure(const KeyPropsList: TKeyPropsList; KeysImported: Longint) of Object;

  TPGPKeyImport = class(TComponent)
  private
    FParentHandle: THandle;
    FFileIn: AnsiString;
    FIgnoreKnownKeys: TIgnoreKnownKeys;
    FKeyData: AnsiString;
    FKeyDlgPrompt: AnsiString;
    FKeyProps: TKeyProps;
    FOnFailure: TOnFailure;
    FOnGetFileIn: TOnGetFileIn;
    FOnKeyImported: TOnKeyImported;
    procedure SetKeyProps(Value: TKeyProps);
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function DoKeyImport: Longint;
    function DoKeyImportFile: Longint;
    property ParentHandle: THandle
      read FParentHandle
      write FParentHandle;
  published
    property FileIn: AnsiString
      read FFileIn
      write FFileIn;
    property IgnoreKnownKeys: TIgnoreKnownKeys
      read FIgnoreKnownKeys
      write FIgnoreKnownKeys;
    property KeyData: AnsiString
      read FKeyData
      write FKeyData;
    property KeyDlgPrompt: AnsiString
      read FKeyDlgPrompt
      write FKeyDlgPrompt;
    property KeyProps: TKeyProps
      read FKeyProps
      write SetKeyProps;
    property OnFailure: TOnFailure
      read FOnFailure
      write FOnFailure;
    property OnGetFileIn: TOnGetFileIn
      read FOnGetFileIn
      write FOnGetFileIn;
    property OnKeyImported: TOnKeyImported
      read FOnKeyImported
      write FOnKeyImported;
  end;


// TPGPKeyExport ---------------------------------------------------------------

type
  TOnGetFileOut = procedure(var FileOut: AnsiString) of Object;
  TOnKeyExported = procedure(const KeyPropsList: TKeyPropsList; const KeyData, FileOut: AnsiString) of Object;

  TPGPKeyExport = class(TComponent)
  private
    FParentHandle: THandle;
    FExportCompatible: Longbool;
    FExportPrivate: Longbool;
    FFileOut: AnsiString;
    FKeyData: AnsiString;
    FKeyIDs: TRawByteStrings;
    FKeyDlgPrompt: AnsiString;
    FKeyProps: TKeyProps;
    FOnFailure: TOnFailure;
    FOnGetFileOut: TOnGetFileOut;
    FOnKeyExported: TOnKeyExported;
    procedure SetKeyIDs(const Value: TRawByteStrings);
    procedure SetKeyProps(Value: TKeyProps);
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function DoKeyExport: Longint;
    function DoKeyExportFile: Longint;
    property ParentHandle: THandle
      read FParentHandle
      write FParentHandle;
  published
    property ExportCompatible: Longbool
      read FExportCompatible
      write FExportCompatible;
    property ExportPrivate: Longbool
      read FExportPrivate
      write FExportPrivate;
    property FileOut: AnsiString
      read FFileOut
      write FFileOut;
    property KeyData: AnsiString
      read FKeyData
      write FKeyData;
    property KeyIDs: TRawByteStrings
      read FKeyIDs
      write SetKeyIDs;
    property KeyDlgPrompt: AnsiString
      read FKeyDlgPrompt
      write FKeyDlgPrompt;
    property KeyProps: TKeyProps
      read FKeyProps
      write SetKeyProps;
    property OnFailure: TOnFailure
      read FOnFailure
      write FOnFailure;
    property OnGetFileOut: TOnGetFileOut
      read FOnGetFileOut
      write FOnGetFileOut;
    property OnKeyExported: TOnKeyExported
      read FOnKeyExported
      write FOnKeyExported;
  end;


// TPGPEncode ------------------------------------------------------------------

type
  TOnEncoded = procedure(const BufferOut, FileOut: AnsiString) of Object;

  TPGPEncode = class(TPGPEncodeCustom)
  private
    FFileOutput: TFileOutput;
    FOnEncoded: TOnEncoded;
    FOnFailure: TOnFailure;
    procedure SetFileOutput(Value: TFileOutput);
  protected
  public
    constructor Create(AOwner: TComponent); override;
    function KeyEncryptBuffer(const DataBuffer: RawByteString; Sign, SMime: Longbool): Longint; override;
    function KeyEncryptFile(const FileName: RawByteString; Sign, SMime: Longbool): Longint; override;
    function ConventionalEncryptBuffer(const DataBuffer: RawByteString): Longint; override;
    function ConventionalEncryptFile(const FileName: RawByteString): Longint; override;
    function ClearSignBuffer(const DataBuffer: RawByteString; SMime: Longbool): Longint; override;
    function ClearSignFile(const FileName: RawByteString; SMime: Longbool): Longint; override;
    function ArmorBuffer(const DataBuffer: RawByteString): Longint; override;
    function ArmorFile(const FileName: RawByteString): Longint; override;
    function DetachedSignBuffer(const DataBuffer: RawByteString): Longint; override;
    function DetachedSignFile(const FileName: RawByteString): Longint; override;
  published
    property FileOutput: TFileOutput
      read FFileOutput
      write SetFileOutput;
    property OnEncoded: TOnEncoded
      read FOnEncoded
      write FOnEncoded;
    property OnFailure: TOnFailure
      read FOnFailure
      write FOnFailure;
  end;


// TPGPDecode ------------------------------------------------------------------

type
  TOnDecoded = procedure(const BufferOut, FileOut: AnsiString;
			 const SigProps: TSigPropsRec;
			 const KeyPropsList: TKeyPropsList) of Object;

  TPGPDecode = class(TPGPDecodeCustom)
  private
    FFileOutput: TFileOutput;
    FIgnoreKnownKeys: TIgnoreKnownKeys;
    FOnDecoded: TOnDecoded;
    FOnFailure: TOnFailure;
    procedure SetFileOutput(Value: TFileOutput);
  protected
  public
    constructor Create(AOwner: TComponent); override;
    function AnalyseBuffer(const DataBuffer: RawByteString): Longint; override;
    function AnalyseFile(const FileName: RawByteString): Longint; override;
    function AnalyseMimeBuffer(const DataBuffer: RawByteString): Longint; override;
    function AnalyseMimeFile(const FileName: RawByteString): Longint; override;
    function DecodeBuffer(const DataBuffer: RawByteString): Longint; override;
    function DecodeFile(const FileName: RawByteString): Longint; override;
    function DecodeMimeBuffer(const DataBuffer: RawByteString): Longint; override;
    function DecodeMimeFile(const FileName: RawByteString): Longint; override;
  published
    property FileOutput: TFileOutput
      read FFileOutput
      write SetFileOutput;
    property IgnoreKnownKeys: TIgnoreKnownKeys
      read FIgnoreKnownKeys
      write FIgnoreKnownKeys;
    property OnDecoded: TOnDecoded
      read FOnDecoded
      write FOnDecoded;
    property OnFailure: TOnFailure
      read FOnFailure
      write FOnFailure;
  end;


// Utilities -------------------------------------------------------------------

type
  EFailException = class(Exception);

function WipeFiles(const FileNames: TRawByteStrings; ParentHandle: THandle): Longbool;
function ValidPassphrase(const KeyID: AnsiString; Passphrase: PAnsiChar): Longbool;
function PassphraseCacheEmpty: Longbool;
function PurgePassphraseCache: Longbool;
function GetShortHexID(const HexID: AnsiString): AnsiString;
function ValidSecMem(Alloc: PAnsiChar): Longbool;
function SecAlloc(Size: Cardinal): PAnsiChar;
function SecFree(var Alloc: PAnsiChar): Longint;
function SecRealloc(var Alloc: PAnsiChar; Size: Cardinal; Copy: Longbool): Longint;
function GetKeyPropsFlags(const setKeyProps: TKeyProps): DWord;
function GetKeyFilter(AlgoFilter: TAlgorithmKeyFilter; BoolFilter: TBooleanKeyFilter; CertFilter: TCertKeyFilter): Longint;
function ShowError(const FOnFailure: TOnFailure; ErrorCode: Longint; ErrorStr: String): Longint;


implementation

{$IFDEF UNICODE}
uses
  AnsiStrings;
{$ENDIF}

// String constants

const
  HexFormError = 'Hex format error: ';
  PathStrError = 'Path RawByteString error: ';
  PGPInitError = 'PGP initialization failure';


// Utility Funcs ---------------------------------------------------------------

function WipeFiles(const FileNames: TRawByteStrings; ParentHandle: THandle): Longbool;
var
  WarnPref	: Longbool;
  PGPsc		: pPGPsc;
  PGPtls	: pPGPtls;
  WipeCount	: PGPUInt32;
  FileIndex	: Longint;
  FileName	: RawByteString;
begin
  PGPsc := NIL;
  PGPtls := NIL;
  Result := FALSE;
  if FileNames <> NIL then with FileNames do begin
    if GetPrefWarnOnWipe(WarnPref) = 0 then SetPrefWarnOnWipe(FALSE);
    try
      if InitPGPsc(ParentHandle, PGPsc, PGPtls, 0) then begin
	try
	  GetPrefWipeCount(WipeCount);
	  for FileIndex := 0 to pred(Count) do begin
	    Result := FALSE;
	    FileName := Strings[FileIndex];
	    if FileExists(FileName) then begin
	      Result := WipeFileList(ParentHandle, PGPsc, CmdLineToFileList(PAnsiChar(FileName)),
				     succ(Ord(SDKVersion >= PGP7XX)), WipeCount);
	    end;
	    if not Result then Break;
	  end;
	finally
	  UninitPGPsc(ParentHandle, PGPsc, PGPtls);
	end;
      end;
    finally
      SetPrefWarnOnWipe(WarnPref);
    end;
  end;
end;

function ValidPassphrase(const KeyID: AnsiString; Passphrase: PAnsiChar): Longbool;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  Utf8Passphrase: PAnsiChar;
begin
  Result := FALSE;
  Context := NIL;
  KeySetMain := NIL;
  if KeyRings.InitKeyRings(Context, KeySetMain) = 0 then begin
    Utf8Passphrase := NIL;
    try
      GetUtf8Passphrase(Passphrase, Utf8Passphrase);
      Result := PassphraseIsValid(Context, KeySetMain, PAnsiChar(KeyID), Utf8Passphrase);
    finally
      PGPFreeData(Utf8Passphrase);
      KeyRings.FreeKeyRings;
    end;
  end;
end;

function PassphraseCacheEmpty: Longbool;
var
  Context: pPGPContext;
begin
  Result := FALSE;
  if PGPNewContext(kPGPsdkAPIVersion, Context) = 0 then begin
    try
      Result := KeyFuncs.IsPassphraseCacheEmpty(Context);
    finally
      PGPFreeContext(Context);
    end;
  end;
end;

function PurgePassphraseCache: Longbool;
var
  Context: pPGPContext;
begin
  Result := FALSE;
  if PGPNewContext(kPGPsdkAPIVersion, Context) = 0 then begin
    try
      Result := KeyFuncs.PurgePassphraseCache(Context);
    finally
      PGPFreeContext(Context);
    end;
  end;
end;

function GetShortHexID(const HexID: AnsiString): AnsiString;
begin
  Result := ShortHexID(HexID);
end;

function ValidSecMem(Alloc: PAnsiChar): Longbool;
begin
  Result := (PGPGetMemoryMgrDataInfo(Alloc) = kPGPMemoryMgrBlockInfo_Valid or
					      kPGPMemoryMgrBlockInfo_Secure or
  					      kPGPMemoryMgrBlockInfo_NonPageable);
end;

function SecAlloc(Size: Cardinal): PAnsiChar;
begin
  Result := PGPNewSecureData(PGPGetDefaultMemoryMgr, Size, kPGPMemoryMgrFlags_Clear);
end;

function SecFree(var Alloc: PAnsiChar): Longint;
begin
  Result := kPGPError_BadMemAddress;
  if Alloc <> NIL then begin
    Result := PGPFreeData(Alloc);
    if Result = kPGPError_NoErr then Alloc := NIL;
  end;
end;

function SecRealloc(var Alloc: PAnsiChar; Size: Cardinal; Copy: Longbool): Longint;
begin
  Result := PGPReallocData(PGPGetDefaultMemoryMgr, Pointer(Alloc), Size, kPGPMemoryMgrFlags_Clear);
  // kPGPMemoryMgrFlags_Clear only clears excess bytes
  if (Result = kPGPError_NoErr) and not Copy then FillChar(Alloc^, Size, 0);
end;

function GetKeyPropsFlags(const setKeyProps: TKeyProps): DWord;
begin
  Result:=
    DWord(KeyProp_HexID in setKeyProps) * spgpKeyPropFlag_KeyID or
    DWord(KeyProp_UserID in setKeyProps) * spgpKeyPropFlag_UserID or
    DWord(KeyProp_Fingerprint in setKeyProps) * spgpKeyPropFlag_Fingerprint or
    DWord(KeyProp_CreaTimeStr in setKeyProps) * spgpKeyPropFlag_CreationTimeStr or
    DWord(KeyProp_ExpTimeStr in setKeyProps) * spgpKeyPropFlag_ExpirationTimeStr or
    DWord(KeyProp_Size in setKeyProps) * spgpKeyPropFlag_KeyBits or
    DWord(KeyProp_X509Cert in setKeyProps) * spgpKeyPropFlag_X509Cert or
    DWord(KeyProp_KeyAlgorithm in setKeyProps) * spgpKeyPropFlag_KeyAlg or
    DWord(KeyProp_HashAlgorithm in setKeyProps) * spgpKeyPropFlag_HashAlg or
    DWord(KeyProp_Trust in setKeyProps) * spgpKeyPropFlag_Trust or
    DWord(KeyProp_Validity in setKeyProps) * spgpKeyPropFlag_Validity or
    DWord(KeyProp_CreaTimeNum in setKeyProps) * spgpKeyPropFlag_CreationTime or
    DWord(KeyProp_ExpTimeNum in setKeyProps) * spgpKeyPropFlag_ExpirationTime or
    DWord(KeyProp_Corrupt in setKeyProps) * spgpKeyPropFlag_IsCorrupt or
    DWord(KeyProp_Secret in setKeyProps) * spgpKeyPropFlag_IsSecret or
    DWord(KeyProp_ImplicitTrust in setKeyProps) * spgpKeyPropFlag_IsAxiomatic or
    DWord(KeyProp_Revoked in setKeyProps) * spgpKeyPropFlag_IsRevoked or
    DWord(KeyProp_Disabled in setKeyProps) * spgpKeyPropFlag_IsDisabled or
    DWord(KeyProp_Expired in setKeyProps) * spgpKeyPropFlag_IsExpired or
    DWord(KeyProp_SecretShared in setKeyProps) * spgpKeyPropFlag_IsSecretShared or
    DWord(KeyProp_CanEncrypt in setKeyProps) * spgpKeyPropFlag_CanEncrypt or
    DWord(KeyProp_CanDecrypt in setKeyProps) * spgpKeyPropFlag_CanDecrypt or
    DWord(KeyProp_CanSign in setKeyProps) * spgpKeyPropFlag_CanSign or
    DWord(KeyProp_CanVerify in setKeyProps) * spgpKeyPropFlag_CanVerify or
    DWord(KeyProp_HasRevoker in setKeyProps) * spgpKeyPropFlag_HasRevoker or
    DWord(KeyProp_HasADK in setKeyProps) * spgpKeyPropFlag_HasADK or
    DWord(KeyProp_HasSubKey in setKeyProps) * spgpKeyPropFlag_HasSubKey or
    DWord(KeyProp_LegacyKey in setKeyProps) * spgpKeyPropFlag_LegacyKey or
    DWord(KeyProp_IncludeUserIDs in setKeyProps) * spgpKeyPropFlag_IncludeUserIDs or
    DWord(KeyProp_IncludeSignerIDs in setKeyProps) * spgpKeyPropFlag_IncludeSignerIDs or
    DWord(KeyProp_IncludeGroupsList in setKeyProps) * spgpKeyPropFlag_IncludeGroupsList;
end;

function GetKeyFilter(AlgoFilter: TAlgorithmKeyFilter; BoolFilter: TBooleanKeyFilter; CertFilter: TCertKeyFilter): Longint;
begin
  Result := 0;
  case AlgoFilter of
    AlgorithmFilter_DHDSS: Result := KeyFilterFlag_DHDSS;
    AlgorithmFilter_RSA: Result := KeyFilterFlag_RSA;
    AlgorithmFilter_RSALegacy: Result := KeyFilterFlag_V3;
  end;
  case BoolFilter of
    BoolFilter_CanEncrypt: Result := Result or KeyFilterFlag_CanEncrypt;
    BoolFilter_CanDecrypt: Result := Result or KeyFilterFlag_CanDecrypt;
    BoolFilter_CanSign: Result := Result or KeyFilterFlag_CanSign;
    BoolFilter_CanVerify: Result := Result or KeyFilterFlag_CanVerify;
  end;
  case CertFilter of
    CertFilter_X509Cert: Result := Result or KeyFilterFlag_X509Cert;
    CertFilter_X509Root: Result := Result or KeyFilterFlag_X509Root;
    CertFilter_X509CanEncrypt: Result := Result or KeyFilterFlag_X509CanEncrypt;
    CertFilter_X509CanSign: Result := Result or KeyFilterFlag_X509CanSign;
  end;
end;

function ShowError(const FOnFailure: TOnFailure; ErrorCode: Longint; ErrorStr: String): Longint;
var
  aErrorStr: Array[0..255] of AnsiChar;
begin
  Result := ErrorCode;
  if PGPInitErrorCode = ieNone then begin
    if ErrorStr = E then begin
      PGPGetErrorString(ErrorCode, 255, aErrorStr);
      ErrorStr := String(Utf8OrAnsi(aErrorStr));
    end;
  end
  else ErrorStr := PGPInitError;
  if Assigned(FOnFailure) then
    FOnFailure(ErrorCode, ErrorStr)
  else Raise EFailException.Create(ErrorStr);
end;

function ShowHexError(const FOnFailure: TOnFailure; const IDStr: RawByteString): Longbool;
begin
  if IsHexID(IDStr) = 0 then begin
    {$IFDEF UNICODE}
    ShowError(FOnFailure, -1, String(HexFormError + QU + IDStr + QU));
    {$ELSE}
    ShowError(FOnFailure, -1, HexFormError + QU + IDStr + QU);
    {$ENDIF}
    Result := TRUE;
  end
  else Result := FALSE;
end;

function ShowPathStrError(const FOnFailure: TOnFailure; const PathStr: RawByteString): Longbool;
begin
  if (PathStr = E) or (Length(PathStr) > MAX_PATH) then begin
    {$IFDEF UNICODE}
    ShowError(FOnFailure, -1, String(PathStrError + QU + PathStr + QU));
    {$ELSE}
    ShowError(FOnFailure, -1, PathStrError + QU + PathStr + QU);
    {$ENDIF}
    Result := TRUE;
  end
  else Result := FALSE;
end;

procedure AdjustKeyProps(BasicProps, NewProps: TKeyProps; var PropsToAdjust: TKeyProps);
const
  X509Props = [KeyProp_X509Cert, KeyProp_X509Key];
begin
  if (KeyProp_X509Cert in NewProps) and not (KeyProp_X509Cert in PropsToAdjust)
  or (KeyProp_X509Key in NewProps) and not (KeyProp_X509Key in PropsToAdjust) then begin
    NewProps := NewProps + X509Props;
  end
  else if not (KeyProp_X509Cert in NewProps) and (KeyProp_X509Cert in PropsToAdjust)
  or not (KeyProp_X509Key in NewProps) and (KeyProp_X509Key in PropsToAdjust) then begin
    NewProps := NewProps - X509Props;
  end;
  if (KeyProp_IncludeSignerIDs in NewProps) and not (KeyProp_IncludeUserIDs in PropsToAdjust) then
    Include(NewProps, KeyProp_IncludeUserIDs)
  else if not (KeyProp_IncludeUserIDs in NewProps) then Exclude(NewProps, KeyProp_IncludeSignerIDs);
  PropsToAdjust := BasicProps + NewProps;
end;


// TPGPGetKeyProps -------------------------------------------------------------

constructor TKeyRingProps.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FRingKeyOrdering := UserID_Order;
  FPrimaryIDs := TRawByteStrings.Create;
  FRingPropsList := TKeyPropsList.Create(spgpKeyPropFlag_All);
end;

destructor TKeyRingProps.Destroy;
begin
  FRingPropsList.Free;
  FPrimaryIDs.Free;
  inherited Destroy;
end;

function TKeyRingProps.GetPrimaryIDs: Longint;
var
  KeyIndex	: Longint;
begin
  Result := 0;
  if (FPrimaryIDs <> NIL) and (FRingPropsList <> NIL) and (FRingPropsList.Count <> 0) then begin
    FPrimaryIDs.Clear;
    with FRingPropsList do begin
      for KeyIndex := 0 to pred(Count) do with PKeyPropsRec(Objects[KeyIndex])^ do begin
	case kKeyAlgorithm of
	  KeyAlgorithm_RSA: FPrimaryIDs.Add(kUserID + ' (RSA  - ' + kSize + ')');
	  KeyAlgorithm_RSAEncryptOnly: FPrimaryIDs.Add(kUserID + ' (RSA encrypt only - ' + kSize + ')');
	  KeyAlgorithm_RSASignOnly: FPrimaryIDs.Add(kUserID + ' (RSA sign only - ' + kSize + ')');
	  KeyAlgorithm_DH: FPrimaryIDs.Add(kUserID + ' (DH - ' + kSize + ')');
	  KeyAlgorithm_DSS: FPrimaryIDs.Add(kUserID + ' (DSS - ' + kSize + ')');
	  KeyAlgorithm_DHDSS: FPrimaryIDs.Add(kUserID + ' (DH/DSS - ' + kSize + ')');
	  {KeyAlgorithm_ElGamalEncSign: FPrimaryIDs.Add(kUserID + ' (ElGamal E/S - ' + kSize + ')'); // GnuPG only}
	end;
	inc(Result);
      end;
    end;
    if Assigned(FComboBox) then FComboBox.Items.Assign(FPrimaryIDs);
    if Assigned(FListBox) then FListBox.Items.Assign(FPrimaryIDs);
  end;
end;

procedure TKeyRingProps.SetRingProps(Value: TKeyProps);
begin
  AdjustKeyProps(KeyProps_IDComplete + [KeyProp_Secret, KeyProp_IncludeGroupsList], Value, FRingProps);
end;

function TKeyRingProps.Update: Longint;
begin
  try
    FRingPropsList.Clear;
    Result := FindKeyProps(E, FRingPropsList,
			   GetKeyPropsFlags(FRingProps),
			   GetKeyFilter(FRingAlgorithmFilter, FRingBoolFilter, CertFilter_AllKeys),
			   TPGPKeyOrdering(succ(FRingKeyOrdering)));
    if Result >= 0 then
      Result := GetPrimaryIDs
    else begin
      if Assigned(FComboBox) then FComboBox.Items.Clear;
      if Assigned(FListBox) then FListBox.Items.Clear;
      ShowError(FOnFailure, Result, E);
    end;
  except
    on EFailException do Raise;
    on E: Exception do begin
      if Assigned(FComboBox) then FComboBox.Items.Clear;
      if Assigned(FListBox) then FListBox.Items.Clear;
      Result := ShowError(FOnFailure, -1, E.Message);
    end;
  end;
end;

constructor TPGPGetKeyProps.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
end;

destructor TPGPGetKeyProps.Destroy;
begin
  inherited Destroy;
end;

procedure TPGPGetKeyProps.SetKeyProps(Value: TKeyProps);
begin
  AdjustKeyProps(KeyProps_IDFlags, Value, FKeyProps);
end;

function TPGPGetKeyProps.DoGetKeyProps: Longint;
var
  KeyPropsList	: TKeyPropsList;
begin
  Result := kPGPError_PublicKeyNotFound;
  try
    KeyPropsList := NIL;
    if Assigned(FOnGetKeyID) then FOnGetKeyID(FKeyID);
    if FKeyID <> E then begin
      Result := FindKeyProps(FKeyID, KeyPropsList,
			     GetKeyPropsFlags(FKeyProps),
			     GetKeyFilter(FAlgorithmFilter, FBoolFilter, CertFilter_AllKeys),
			     TPGPKeyOrdering(succ(FKeyOrdering)));
      try
	if Result >= 0 then begin
	  if Assigned(FOnGetKeyProps) then FOnGetKeyProps(KeyPropsList);
	end
	else ShowError(FOnFailure, Result, E);
      finally
	KeyPropsList.Free;
      end;
    end;
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPGetKeyProps.KeyIsOnRing: Longint;
var
  KeyPropsList	: TKeyPropsList; // dummy
begin
  Result := kPGPError_PublicKeyNotFound;
  try
    KeyPropsList := NIL;
    if Assigned(FOnGetKeyID) then FOnGetKeyID(FKeyID);
    if FKeyID <> E then begin
      Result := FindKeyProps(FKeyID, KeyPropsList,
			     spgpKeyPropFlag_None,
			     KeyFilterFlag_AllKeys,
			     Any_Ordering);
      if Result < 0 then ShowError(FOnFailure, Result, E);
    end;
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPGetKeyProps.KeyRingCount: Longint;
var
  KeyPropsList	: TKeyPropsList; // dummy
begin
  try
    KeyPropsList := NIL;
    Result := FindKeyProps(E, KeyPropsList,
			   spgpKeyPropFlag_None,
			   KeyFilterFlag_AllKeys,
			   Any_Ordering);
    if Result < 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPGetKeyProps.GetKeyProp(KeyPropsList: TKeyPropsList; KeyIndex: Longint; KeyProp: TKeyProp): Variant;
begin
  Result := 0;
  if (KeyPropsList <> NIL) and (KeyIndex >= 0) and (KeyIndex < KeyPropsList.Count) then begin
    with PKeyPropsRec(KeyPropsList.Objects[KeyIndex])^ do begin
      case KeyProp of
	KeyProp_HexID: if KeyProp_HexID in KeyPropsList.ValidProps then begin
	  Result := kHexID;
	  Exit;
	end;
	KeyProp_UserID: if KeyProp_UserID in KeyPropsList.ValidProps then begin
	  Result := kUserID;
	  Exit;
	end;
	KeyProp_Fingerprint: if KeyProp_Fingerprint in KeyPropsList.ValidProps then begin
	  Result := kFingerprint;
	  Exit;
	end;
	KeyProp_CreaTimeStr: if KeyProp_CreaTimeStr in KeyPropsList.ValidProps then begin
	  Result := kCreaTimeStr;
	  Exit;
	end;
	KeyProp_ExpTimeStr: if KeyProp_ExpTimeStr in KeyPropsList.ValidProps then begin
	  Result := kExpTimeStr;
	  Exit;
	end;
	KeyProp_Size: if KeyProp_Size in KeyPropsList.ValidProps then begin
	  Result := kSize;
	  Exit;
	end;
	KeyProp_X509Cert: if KeyProp_X509Cert in KeyPropsList.ValidProps then begin
	  Result := kX509Cert;
	  Exit;
	end;
	KeyProp_KeyAlgorithm: if KeyProp_KeyAlgorithm in KeyPropsList.ValidProps then begin
	  Result := kKeyAlgorithm;
	  Exit;
	end;
	KeyProp_HashAlgorithm: if KeyProp_HashAlgorithm in KeyPropsList.ValidProps then begin
	  Result := kHashAlgorithm;
	  Exit;
	end;
	KeyProp_Trust: if KeyProp_Trust in KeyPropsList.ValidProps then begin
	  Result := kTrust;
	  Exit;
	end;
	KeyProp_Validity: if KeyProp_Validity in KeyPropsList.ValidProps then begin
	  Result := kValidity;
	  Exit;
	end;
	KeyProp_CreaTimeNum: if KeyProp_CreaTimeNum in KeyPropsList.ValidProps then begin
	  Result := kCreaTimeNum;
	  Exit;
	end;
	KeyProp_ExpTimeNum: if KeyProp_ExpTimeNum in KeyPropsList.ValidProps then begin
	  Result := kExpTimeNum;
	  Exit;
	end;
	KeyProp_Corrupt: if KeyProp_Corrupt in KeyPropsList.ValidProps then begin
	  Result := kCorrupt;
	  Exit;
	end;
	KeyProp_Secret: if KeyProp_Secret in KeyPropsList.ValidProps then begin
	  Result := kPrivate;
	  Exit;
	end;
	KeyProp_ImplicitTrust: if KeyProp_ImplicitTrust in KeyPropsList.ValidProps then begin
	  Result := kImplicitTrust;
	  Exit;
	end;
	KeyProp_Revoked: if KeyProp_Revoked in KeyPropsList.ValidProps then begin
	  Result := kRevoked;
	  Exit;
	end;
	KeyProp_Disabled: if KeyProp_Disabled in KeyPropsList.ValidProps then begin
	  Result := kDisabled;
	  Exit;
	end;
	KeyProp_Expired: if KeyProp_Expired in KeyPropsList.ValidProps then begin
	  Result := kExpired;
	  Exit;
	end;
	KeyProp_SecretShared: if KeyProp_SecretShared in KeyPropsList.ValidProps then begin
	  Result := kSecShared;
	  Exit;
	end;
	KeyProp_CanEncrypt: if KeyProp_CanEncrypt in KeyPropsList.ValidProps then begin
	  Result := kCanEncrypt;
	  Exit;
	end;
	KeyProp_CanDecrypt: if KeyProp_CanDecrypt in KeyPropsList.ValidProps then begin
	  Result := kCanDecrypt;
	  Exit;
	end;
	KeyProp_CanSign: if KeyProp_CanSign in KeyPropsList.ValidProps then begin
	  Result := kCanSign;
	  Exit;
	end;
	KeyProp_CanVerify: if  KeyProp_CanVerify in KeyPropsList.ValidProps then begin
	  Result := kCanVerify;
	  Exit;
	end;
	KeyProp_HasRevoker: if KeyProp_HasRevoker in KeyPropsList.ValidProps then begin
	  Result := kHasRevoker;
	  Exit;
	end;
	KeyProp_HasADK: if KeyProp_HasADK in KeyPropsList.ValidProps then begin
	  Result := kHasADK;
	  Exit;
	end;
	KeyProp_HasSubKey: if KeyProp_HasSubKey in KeyPropsList.ValidProps then begin
	  Result := kHasSubKey;
	  Exit;
	end;
	KeyProp_LegacyKey: if KeyProp_LegacyKey in KeyPropsList.ValidProps then begin
	  Result := kLegacyKey;
	  Exit;
	end;
      end;
      ShowError(FOnFailure, kPGPError_BadParams, E);
    end;
  end
  else ShowError(FOnFailure, kPGPError_ItemNotFound, E);
end;

function TPGPGetKeyProps.GetUserIDList(KeyPropsList: TKeyPropsList; KeyIndex: Longint): TRawByteStrings;
begin
  Result := NIL;
  if (KeyPropsList <> NIL) and (KeyIndex >= 0) and (KeyIndex < KeyPropsList.Count) then
    Result := PKeyPropsRec(KeyPropsList.Objects[KeyIndex])^.kUserIDList
  else ShowError(FOnFailure, kPGPError_ItemNotFound, E);
end;

function TPGPGetKeyProps.GetSignerIDList(KeyPropsList: TKeyPropsList; KeyIndex: Longint): TRawByteStrings;
begin
  Result := NIL;
  if (KeyPropsList <> NIL) and (KeyIndex >= 0) and (KeyIndex < KeyPropsList.Count) then
    Result := PKeyPropsRec(KeyPropsList.Objects[KeyIndex])^.kSignerIDList
  else ShowError(FOnFailure, kPGPError_ItemNotFound, E);
end;

function TPGPGetKeyProps.GetRevKeyIDList(KeyPropsList: TKeyPropsList; KeyIndex: Longint): TRawByteStrings;
begin
  Result := NIL;
  if (KeyPropsList <> NIL) and (KeyIndex >= 0) and (KeyIndex < KeyPropsList.Count) then
    Result := PKeyPropsRec(KeyPropsList.Objects[KeyIndex])^.kRevKeyIDList
  else ShowError(FOnFailure, kPGPError_ItemNotFound, E);
end;

function TPGPGetKeyProps.GetADKeyIDList(KeyPropsList: TKeyPropsList; KeyIndex: Longint): TRawByteStrings;
begin
  Result := NIL;
  if (KeyPropsList <> NIL) and (KeyIndex >= 0) and (KeyIndex < KeyPropsList.Count) then
    Result := PKeyPropsRec(KeyPropsList.Objects[KeyIndex])^.kADKeyIDList
  else ShowError(FOnFailure, kPGPError_ItemNotFound, E);
end;

function TPGPGetKeyProps.GetGroupHexIDs(KeyPropsList: TKeyPropsList; GroupIndex: Longint; var GroupName: AnsiString): AnsiString;
var
  TabPos	: Longint;
begin
  Result := E;
  GroupName := E;
  if (KeyPropsList <> NIL) and (GroupIndex >= 0) and (GroupIndex < KeyPropsList.GroupsList.Count) then begin
    Result := KeyPropsList.GroupsList.Strings[GroupIndex];
    TabPos := FirstCharPos(TAB, Result);
    GroupName := Copy(Result, 1, pred(TabPos));
    Delete(Result, 1, TabPos);
  end
  else ShowError(FOnFailure, kPGPError_ItemNotFound, E);
end;


// TPGPSetKeyProps -------------------------------------------------------------

constructor TPGPSetKeyProps.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FMinPassLen := 10;
  FMinPassQual := 100;
  if AOwner is TForm then FParentHandle := TForm(AOwner).Handle;
end;

destructor TPGPSetKeyProps.Destroy;
begin
  inherited Destroy;
end;

function TPGPSetKeyProps.GetKeyHexID(FilterFlags: DWord): Longint;
var
  KeyPropsList	: TKeyPropsList;
  KeyPropsRec	: TKeyPropsRec;
begin
  Result := 0;
  KeyPropsList := NIL;
  try
    if FindKeyProps(FKeyHexID, KeyPropsList,
		    GetKeyPropsFlags(FKeyProps),
		    FilterFlags, Any_Ordering) <> 1 then begin
      if KeyPropsList <> NIL then KeyPropsList.Clear;
      Result := SelectKeysDialog(FKeyDlgPrompt, KeyPropsList, TRUE,
				 GetKeyPropsFlags(FKeyProps),
				 FilterFlags, FParentHandle);
      if Result = 0 then begin
	KeyPropsRec := KeyPropsList.KeyProps[0];
	FKeyHexID := KeyPropsList[0];
      end;
    end
    else begin
      KeyPropsRec := KeyPropsList.KeyProps[0];
      FKeyHexID := KeyPropsList[0];
    end;
    if Result = 0 then begin
      if not ShowHexError(FOnFailure, FKeyHexID) then begin
	if Assigned(FOnKeySelected) and not FOnKeySelected(KeyPropsRec) then Result := kPGPError_UserAbort;
      end
      else Result := kPGPError_BadParams;
    end;
  finally
    KeyPropsList.Free;
  end;
end;

procedure TPGPSetKeyProps.SetKeyProps(Value: TKeyProps);
begin
  AdjustKeyProps(KeyProps_IDFlags, Value, FKeyProps);
end;

function TPGPSetKeyProps.DoKeyDisable: Longint;
begin
  try
    Result := GetKeyHexID(KeyFilterFlag_Enabled);
    if Result = 0 then Result := KeyDisable(PAnsiChar(FKeyHexID));
    if Result < 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPSetKeyProps.DoKeyEnable: Longint;
begin
  try
    Result := GetKeyHexID(KeyFilterFlag_Disabled);
    if Result = 0 then Result := KeyEnable(PAnsiChar(FKeyHexID));
    if Result < 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPSetKeyProps.DoKeyRemove: Longint;
begin
  try
    Result := GetKeyHexID(KeyFilterFlag_AllKeys);
    if Result = 0 then Result := KeyRemove(PAnsiChar(FKeyHexID));
    if Result < 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPSetKeyProps.DoKeyRevoke: Longint;
var
  Passphrase	: PAnsiChar;
  Cancel	: Longbool;
begin
  try
    Result := GetKeyHexID(KeyFilterFlag_CanSign);
    if Result = 0 then begin
      if Assigned(FOnEnterPassphrase) then begin
	Passphrase := PGPNewSecureData(PGPGetDefaultMemoryMgr, succ(MaxPassphraseLength) * UTF8Factor,
				       kPGPMemoryMgrFlags_Clear);
	if Passphrase <> NIL then begin
	  try
	    Cancel := FALSE;
	    Result := kPGPError_UserAbort;
	    FOnEnterPassphrase(Passphrase, Cancel);
	    if not Cancel then begin
	      if SDKVersion >= PGP8XX then begin
		SecureAnsiToUtf8PChar(Passphrase, Passphrase, succ(MaxPassphraseLength) * UTF8Factor);
	      end;
	      Result := KeyRevoke(PAnsiChar(FKeyHexID), Passphrase);
	    end;
	  finally
	    PGPFreeData(Passphrase);
	  end;
	end
	else Result := kPGPError_OutOfMemory;
      end
      else Result := KeyRevokeDialog(FKeyHexID, FPassDlgPrompt, FParentHandle);
    end;
    if Result < 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPSetKeyProps.DoKeyPassChange: Longint;
var
  OldPassphrase	: PAnsiChar;
  NewPassphrase	: PAnsiChar;
  PassLength	: Longint;
  Cancel	: Longbool;
begin
  try
    Result := GetKeyHexID(KeyFilterFlag_CanSign);
    if Result = 0 then begin
      if Assigned(FOnChangePassphrase) then begin
	OldPassphrase := PGPNewSecureData(PGPGetDefaultMemoryMgr, succ(MaxPassphraseLength) * UTF8Factor,
					  kPGPMemoryMgrFlags_Clear);
	if OldPassphrase <> NIL then begin
	  try
	    NewPassphrase := PGPNewSecureData(PGPGetDefaultMemoryMgr, succ(MaxPassphraseLength) * UTF8Factor,
					      kPGPMemoryMgrFlags_Clear);
	    if NewPassphrase <> NIL then begin
	      try
		Cancel := FALSE;
		Result := kPGPError_UserAbort;
		repeat
		  if SDKVersion >= PGP8XX then begin
		    SecureUtf8ToAnsiPChar(OldPassphrase, OldPassphrase, succ(MaxPassphraseLength) * UTF8Factor);
		    SecureUtf8ToAnsiPChar(NewPassphrase, NewPassphrase, succ(MaxPassphraseLength) * UTF8Factor);
		  end;
		  FOnChangePassphrase(OldPassphrase, NewPassphrase, FMinPassLen, FMinPassQual, Cancel);
		  PassLength := StrLen(NewPassphrase);
		  if SDKVersion >= PGP8XX then begin
		    SecureAnsiToUtf8PChar(OldPassphrase, OldPassphrase, succ(MaxPassphraseLength) * UTF8Factor);
		    SecureAnsiToUtf8PChar(NewPassphrase, NewPassphrase, succ(MaxPassphraseLength) * UTF8Factor);
		  end;
		until Cancel
		or (PassLength >= FMinPassLen) and (PGPEstimatePassphraseQuality(NewPassphrase) >= FMinPassQual);
		if not Cancel then Result := ChangePassphrase(PAnsiChar(FKeyHexID), OldPassphrase, NewPassphrase);
	      finally
		PGPFreeData(NewPassphrase);
	      end;
	    end
	    else Result := kPGPError_OutOfMemory;
	  finally
	    PGPFreeData(OldPassphrase);
	  end;
	end
	else Result := kPGPError_OutOfMemory;
      end
      else begin
	Result := KeyPassChangeDialog(FKeyHexID, FPassDlgOldPrompt, FPassDlgNewPrompt,
				      FMinPassLen, FMinPassQual, FParentHandle);
      end;
    end;
    if Result < 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;


// TPGPPreferences -------------------------------------------------------------

constructor TPGPPreferences.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if hPGPclLib <> 0 then begin
    GetPrefShareCache(FPassCacheShare);
    GetPrefCacheTime(Longint(FPassCacheSeconds));
  end;
  if AOwner is TForm then FParentHandle := TForm(AOwner).Handle;
end;

destructor TPGPPreferences.Destroy;
begin
  inherited Destroy;
end;

function TPGPPreferences.ShowPrefsDialog: Longint;
begin
  Result := PreferencesDialog(Ord(FPrefsPage), FParentHandle);
end;

function TPGPPreferences.GetFlags: Longint;
begin
  Result := Ord(Pref_PublicKeyring in FPGPPrefs) * PrefsFlag_PublicKeyring or
	    Ord(Pref_PrivateKeyring in FPGPPrefs) * PrefsFlag_PrivateKeyring or
	    Ord(Pref_RandomSeedFile in FPGPPrefs) * PrefsFlag_RandomSeedFile or
	    Ord(Pref_GroupsFile in FPGPPrefs) * PrefsFlag_GroupsFile or
	    Ord(Pref_DefaultKeyID in FPGPPrefs) * PrefsFlag_DefaultKeyID;
end;

function TPGPPreferences.PrefsError: Longbool;
begin
  with FPreferences do begin
    Result := (Pref_PublicKeyring in FPGPPrefs) and ShowPathStrError(FOnFailure, PublicKeyring) or
	      (Pref_PrivateKeyring in FPGPPrefs) and ShowPathStrError(FOnFailure, PrivateKeyring) or
	      (Pref_RandomSeedFile in FPGPPrefs) and ShowPathStrError(FOnFailure, RandomSeedFile) or
	      (Pref_GroupsFile in FPGPPrefs) and ShowPathStrError(FOnFailure, GroupsFile) or
	      (Pref_DefaultKeyID in FPGPPrefs) and ShowHexError(FOnFailure, DefaultKeyHexID);
  end;
end;

function TPGPPreferences.GetAltPubKeyring: AnsiString;
begin
  Result := FAltPubringFile;
end;

function TPGPPreferences.GetAltPrivKeyring: AnsiString;
begin
  Result := FAltSecringFile;
end;

function TPGPPreferences.GetAltGroupsFile: AnsiString;
begin
  Result := FAltGroupsFile;
end;

function TPGPPreferences.GetPGPVersion: AnsiString;
begin
  Result := pgpBase.PGPVersion;
end;

function TPGPPreferences.GetVersionString: AnsiString;
begin
  Result := MyVersion;
end;

procedure TPGPPreferences.SetAltPubKeyring(const Value: AnsiString);
begin
  FAltPubringFile := Value;
  KeyRings.PubringFile := Value;
end;

procedure TPGPPreferences.SetAltPrivKeyring(const Value: AnsiString);
begin
  FAltSecringFile := Value;
  KeyRings.SecringFile := Value;
end;

procedure TPGPPreferences.SetAltGroupsFile(const Value: AnsiString);
begin
  FAltGroupsFile := Value;
  KeyRings.GroupsFile := Value;
end;

procedure TPGPPreferences.SetVersionString(const Value: AnsiString);
begin
  if Value <> E then
    StrPLCopy(MyVersion, Value, pred(SizeOf(TVersionString)))
  else if PGPGetSDKString(MyVersion) <> 0 then MyVersion := 'N/A';
end;

function TPGPPreferences.DoGetPreferences: Longint;
var PrefsFlag: Longint;
begin
  Result := kPGPError_BadParams;
  try
    PrefsFlag := GetFlags;
    if PrefsFlag <> 0 then begin
      Result := GetPreferences(FPreferences, PrefsFlag);
      if Result < 0 then
	ShowError(FOnFailure, Result, E)
      else if Assigned(FOnGetPreferences) then FOnGetPreferences(FPreferences);
    end;
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPPreferences.DoSetPreferences: Longint;
var PrefsFlag: Longint;
begin
  Result := kPGPError_BadParams;
  try
    if Assigned(FOnSetPreferences) then FOnSetPreferences(FPreferences);
    PrefsFlag := GetFlags;
    if (PrefsFlag <> 0) and not PrefsError then begin
      Result := SetPreferences(FPreferences, PrefsFlag);
      if Result < 0 then ShowError(FOnFailure, Result, E);
    end;
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;


// TPGPKeyServer ---------------------------------------------------------------

constructor TPGPKeyServer.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if AOwner is TForm then FParentHandle := TForm(AOwner).Handle;
end;

destructor TPGPKeyServer.Destroy;
begin
  inherited Destroy;
end;

procedure TPGPKeyServer.SetKeyProps(Value: TKeyProps);
begin
  AdjustKeyProps(KeyProps_IDComplete, Value, FKeyProps);
end;

function TPGPKeyServer.DoGetKeysFromServer(const KeyData: AnsiString): Longint;
var
  KeyPropsList	 : TKeyPropsList;
begin
  KeyPropsList := NIL;
  try
    Result := GetKeyFromServerDialog(FKeyDlgPrompt, KeyData, KeyPropsList, GetKeyPropsFlags(FKeyProps), FParentHandle);
    try
      if Result = 0 then begin
	if Assigned(FOnServerResults) then FOnServerResults(KeyPropsList);
      end
      else ShowError(FOnFailure, Result, E);
    finally
      KeyPropsList.Free;
    end;
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPKeyServer.DoSendKeysToServer: Longint;
var
  KeyPropsList	 : TKeyPropsList;
begin
  KeyPropsList := NIL;
  try
    Result := SendKeyToServerDialog(FKeyDlgPrompt, KeyPropsList, GetKeyPropsFlags(FKeyProps), FParentHandle);
    try
      if Result = 0 then begin
	if Assigned(FOnServerResults) then FOnServerResults(KeyPropsList);
      end
      else ShowError(FOnFailure, Result, E);
    finally
      KeyPropsList.Free;
    end;
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPKeyServer.DoGetDefaultServer: Longint;
var
  Entry		 : TRawByteStrings;
  Index		 : Integer;
  Pref		 : pPGPPref;
  pKeyServerEntry: pPGPKeyServerEntry;
  ServerCount	 : PGPUInt32;
  FlagValue	 : PGPUInt32;
begin
  Pref := NIL;
  pKeyServerEntry := NIL;
  try
    if SDKVersion >= PGP7XX then begin
      Result := kPGPError_PrefNotFound;
      if PrefsFile <> NIL then with PrefsFile do if LoadPrefs then begin
	try
	  Entry := TRawByteStrings.Create;
	  if Entry <> NIL then with Entry do begin
	    try
	      if SDKVersion >= PGP9XX then begin
		Text := GetPrefValue(KeyServers9, Index);
		if Count > 0 then CommaText := Strings[0];
	      end
	      else CommaText := PrefsData.Values[KeyServers7 + '1'];
	      if Count >= 8 then with FServerEntry do begin
		if SDKVersion >= PGP9XX then begin
		  ServerProtocol := TServerType(StrToIntDef(Values[kPGPPrefKeyServerProtocol], 0));
		  ServerDNS := Values[kPGPPrefKeyServerHostname];
		  ServerPort := StrToIntDef(Values[kPGPPrefKeyServerPort], 0);
		  FlagValue := StrToIntDef(Values[kPGPPrefKeyServerFlags], 0);
		end
		else begin
		  ServerProtocol := TServerType(StrToIntDef(Strings[0], 0));
		  ServerDNS := Strings[3];
		  ServerPort := StrToIntDef(Strings[4], 0);
		  FlagValue := StrToIntDef(Strings[7], 0);
		end;
		FServerEntry.ServerFlags := [];
		if FlagValue and kKeyServerListed <> 0 then Include(ServerFlags, KeyServerListed);
		if FlagValue and kKeyServerIsRoot <> 0 then Include(ServerFlags, KeyServerIsRoot);
	      end;
	    finally
	      Entry.Free;
	    end;
	    Result := 0;
	  end;
	except
	end;
	PrefsData.Clear;
      end;
    end
    else begin
      Result := PGPclOpenClientPrefs(PGPGetDefaultMemoryMgr, Pref);
      if Result = 0 then begin
	try
	  Result := PGPGetKeyServerPrefs(Pref, pKeyServerEntry, ServerCount);
	  try
	    if Result = 0 then begin
	      if ServerCount <> 0 then begin
		with FServerEntry do begin
		  ServerProtocol := TServerType(pKeyServerEntry^[0].Protocol);
		  ServerDNS := pKeyServerEntry^[0].ServerDNS;
		  ServerPort := pKeyServerEntry^[0].ServerPort;
		  FServerEntry.ServerFlags := [];
		  if pKeyServerEntry^[0].Flags and kKeyServerListed <> 0 then Include(ServerFlags, KeyServerListed);
		  if pKeyServerEntry^[0].Flags and kKeyServerIsRoot <> 0 then Include(ServerFlags, KeyServerIsRoot);
		end;
		if Assigned(FOnGetDefaultServer) then FOnGetDefaultServer(FServerEntry);
	      end
	      else Result := kPGPError_PrefNotFound;
	    end;
	  finally
	    PGPFreeData(pKeyServerEntry);
	  end;
	finally
	  PGPclCloseClientPrefs(Pref, FALSE);
	end;
      end;
    end;
    if Result < 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPKeyServer.DoSetDefaultServer: Longint;
var
  Entry		 : TRawByteStrings;
  Index		 : Integer;
  Pref		 : pPGPPref;
  pKeyServerEntry: pPGPKeyServerEntry;
  ServerCount	 : PGPUInt32;
  FlagValue	 : PGPUint32;
begin
  Pref := NIL;
  pKeyServerEntry := NIL;
  try
    if SDKVersion >= PGP7XX then begin
      Result := kPGPError_PrefNotFound;
      if PrefsFile <> NIL then with PrefsFile do if LoadPrefs then begin
	try
	  Entry := TRawByteStrings.Create;
	  if Entry <> NIL then with Entry do begin
	    try
	      if SDKVersion >= PGP9XX then begin
		Text := GetPrefValue(KeyServers9, Index);
		if Count > 0 then CommaText := Strings[0];
	      end
	      else CommaText := PrefsData.Values[KeyServers7 + '1'];
	      if Count >= 8 then with FServerEntry do begin
		FlagValue := Ord(KeyServerListed in ServerFlags) * kKeyServerListed or
			     Ord(KeyServerIsRoot in ServerFlags) * kKeyServerIsRoot;
		if SDKVersion >= PGP9XX then begin
		  Values[kPGPPrefKeyServerTitle] := BS;
		  Values[kPGPPrefKeyServerDomain] := BS;
		  Values[kPGPPrefKeyServerProtocol] := IntToStr(Ord(ServerProtocol));
		  Values[kPGPPrefKeyServerHostname] := ServerDNS;
		  Values[kPGPPrefKeyServerPort] := IntToStr(ServerPort);
		  Values[kPGPPrefKeyServerFlags] := IntToStr(FlagValue);
		  case Ord(ServerProtocol) of
		    kPGPKeyServerProtocol_LDAP: begin
		      Values[kPGPPrefKeyServerKeyserverType] := IntToStr(kPGPKeyServerType_PGPLDAP);
		    end;
		    kPGPKeyServerProtocol_LDAPS: begin
		      Values[kPGPPrefKeyServerKeyserverType] := IntToStr(kPGPKeyServerType_PGPLDAPS);
		    end;
		    kPGPKeyServerProtocol_HTTP: begin
		      Values[kPGPPrefKeyServerKeyserverType] := IntToStr(kPGPKeyServerType_PGPHTTP);
		    end;
		    kPGPKeyServerProtocol_HTTPS: begin
		      Values[kPGPPrefKeyServerKeyserverType] := IntToStr(kPGPKeyServerType_PGPHTTP);
		    end;
		  end;
		  Values[kPGPPrefKeyServerAuthAlgorithm] := '-1';
		  Values[kPGPPrefKeyServerTrusted] := 'FALSE';
		  SetPrefValue(KeyServers9, CommaText, Index);
		end
		else begin
		  Text := IntToStr(Ord(ServerProtocol)) + CM +
			  Strings[1] + CM +
			  QU + Strings[2] + QU + CM +
			  QU + ServerDNS + QU + CM +
			  IntToStr(ServerPort) + CM +
			  QU + Strings[5] + QU + CM +
			  Strings[6] + CM +
			  IntToStr(FlagValue);
		end;
		PrefsData.Values[KeyServers7 + '1'] := {$IFDEF UNICODE}AnsiStrings.{$ENDIF}TrimRight(Text);
	      end;
	    finally
	      Entry.Free;
	    end;
	    if SavePrefs then Result := 0;
	  end;
	except
	end;
	PrefsData.Clear;
      end;
    end
    else begin
      Result := PGPclOpenClientPrefs(PGPGetDefaultMemoryMgr, Pref);
      if Result = 0 then begin
	try
	  Result := PGPGetKeyServerPrefs(Pref, pKeyServerEntry, ServerCount);
	  try
	    if Result = 0 then begin
	      if ServerCount <> 0 then begin
		if Assigned(FOnSetDefaultServer) then FOnSetDefaultServer(FServerEntry);
		with FServerEntry do begin
		  pKeyServerEntry^[0].StructSize := SizeOf(TPGPKeyServerEntry);
		  if ServerProtocol <> KeyServerType_Invalid then
		    pKeyServerEntry^[0].Protocol := PGPKeyServerProtocol(ServerProtocol)
		  else begin
		    Result := kPGPError_BadParams;
		    Exit;
		  end;
		  if ServerDNS <> E then
		    StrPLCopy(pKeyServerEntry^[0].ServerDNS, ServerDNS, kMaxServerNameLength)
		  else begin
		    Result := kPGPError_BadParams;
		    Exit;
		  end;
		  pKeyServerEntry^[0].ServerPort := ServerPort;
		  pKeyServerEntry^[0].Flags := Ord(KeyServerListed in ServerFlags) * kKeyServerListed or
					    Ord(KeyServerIsRoot in ServerFlags) * kKeyServerIsRoot;
		end;
		Result := PGPSetKeyServerPrefs(Pref, pKeyServerEntry, ServerCount);
		if Result = 0 then PGPclNotifyPrefsChanges(GetCurrentProcessID);
	      end
	      else Result := kPGPError_PrefNotFound;
	    end;
	  finally
	    PGPFreeData(pKeyServerEntry);
	  end;
	finally
	  PGPclCloseClientPrefs(Pref, TRUE);
	end;
      end;
    end;
    if Result < 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;


// TPGPKeysGenerate ------------------------------------------------------------

constructor TPGPKeysGenerate.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if AOwner is TForm then ParentHandle := TForm(AOwner).Handle;
end;

function TPGPKeysGenerate.DHDSSKeyGenerate: Longint;
begin
  try
    Result := inherited DHDSSKeyGenerate;
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPKeysGenerate.DHSubKeyGenerate: Longint;
begin
  try
    Result := inherited DHSubKeyGenerate;
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPKeysGenerate.DSAKeyGenerate: Longint;
begin
  try
    Result := inherited DSAKeyGenerate;
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPKeysGenerate.RSASubKeyGenerate: Longint;
begin
  try
    Result := inherited RSASubKeyGenerate;
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPKeysGenerate.RSAKeyGenerate(Legacy: Longbool): Longint;
begin
  try
    Result := inherited RSAKeyGenerate(Legacy);
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;


// TPGPKeyImport ---------------------------------------------------------------

constructor TPGPKeyImport.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if AOwner is TForm then FParentHandle := TForm(AOwner).Handle;
end;

destructor TPGPKeyImport.Destroy;
begin
  inherited Destroy;
end;

procedure TPGPKeyImport.SetKeyProps(Value: TKeyProps);
begin
  AdjustKeyProps(KeyProps_IDComplete, Value, FKeyProps);
end;

function TPGPKeyImport.DoKeyImport: Longint;
var
  KeyPropsList	: TKeyPropsList;
  InputData	: RawByteString;
begin
  Result := kPGPError_ItemNotFound;
  KeyPropsList := NIL;
  try
    if FKeyData <> E then begin
      InputData := FKeyData;
      Result := KeyImportDialog(FKeyDlgPrompt, E, InputData, NIL, KeyPropsList, FALSE,
				Ord(Ignore_ByHexID in FIgnoreKnownKeys) * IgnoreFlag_ByHexID or
				Ord(Ignore_ByUserID in FIgnoreKnownKeys) * IgnoreFlag_ByUserID,
				GetKeyPropsFlags(FKeyProps), FParentHandle);
      FKeyData := InputData;
      try
	if (Result = 0) or ((Result = kPGPError_FilePermissions) and (KeyPropsList.Count > 0)) then begin
	  if Assigned(FOnKeyImported) then FOnKeyImported(KeyPropsList, KeyPropsList.Count);
	end;
      finally
	KeyPropsList.Free;
      end;
    end;
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPKeyImport.DoKeyImportFile: Longint;
var
  KeyPropsList	: TKeyPropsList;
  InputFile	: RawByteString;
begin
  Result := kPGPError_FileNotFound;
  KeyPropsList := NIL;
  FKeyProps := FKeyProps;
  try
    if Assigned(FOnGetFileIn) then FOnGetFileIn(FFileIn);
    if FFileIn <> E then begin
      InputFile := FFileIn;
      Result := KeyImportDialog(FKeyDlgPrompt, E, InputFile, NIL, KeyPropsList, TRUE,
				Ord(Ignore_ByHexID in FIgnoreKnownKeys) * IgnoreFlag_ByHexID or
				Ord(Ignore_ByUserID in FIgnoreKnownKeys) * IgnoreFlag_ByUserID,
				GetKeyPropsFlags(FKeyProps), FParentHandle);
      FFileIn := InputFile;
      try
	if (Result = 0) or ((Result = kPGPError_FilePermissions) and (KeyPropsList.Count > 0)) then begin
	  if Assigned(FOnKeyImported) then FOnKeyImported(KeyPropsList, KeyPropsList.Count);
	end;
      finally
	KeyPropsList.Free;
      end;
    end;
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;


// TPGPKeyExport ---------------------------------------------------------------

constructor TPGPKeyExport.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FExportCompatible := TRUE;
  FKeyIDs := TRawByteStrings.Create;
  if AOwner is TForm then FParentHandle := TForm(AOwner).Handle;
end;

destructor TPGPKeyExport.Destroy;
begin
  FKeyIDs.Free;
  inherited Destroy;
end;

procedure TPGPKeyExport.SetKeyIDs(const Value: TRawByteStrings);
begin
  FKeyIDs.Assign(Value);
end;

procedure TPGPKeyExport.SetKeyProps(Value: TKeyProps);
begin
  AdjustKeyProps(KeyProps_IDComplete, Value, FKeyProps);
end;

function TPGPKeyExport.DoKeyExport: Longint;
var
  KeyPropsList	: TKeyPropsList;
  OutputData	: RawByteString;
begin
  Result := kPGPError_ItemNotFound;
  FKeyData := E;
  KeyPropsList := NIL;
  try
    if (FKeyIDs <> NIL) or (FKeyDlgPrompt <> E) then begin
      OutputData := FKeyData;
      Result := KeyExportDialog(FKeyDlgPrompt, FKeyIDs, OutputData,
				KeyPropsList, FALSE,
				FExportCompatible, FExportPrivate,
				GetKeyPropsFlags(FKeyProps), FParentHandle);
      FKeyData := OutputData;
    end;
    try
      if Result = 0 then begin
	if Assigned(FOnKeyExported) then FOnKeyExported(KeyPropsList, FKeyData, E);
      end
      else ShowError(FOnFailure, Result, E);
    finally
      KeyPropsList.Free;
    end;
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPKeyExport.DoKeyExportFile: Longint;
var
  KeyPropsList	: TKeyPropsList;
  OutputFile	: RawByteString;
begin
  Result := kPGPError_FileNotFound;
  FKeyData := E;
  KeyPropsList := NIL;
  try
    if Assigned(FOnGetFileOut) then FOnGetFileOut(FFileOut);
    if FFileOut <> E then begin
      Result := kPGPError_ItemNotFound;
      if (FKeyIDs <> NIL) or (FKeyDlgPrompt <> E) then begin
	OutputFile := FFileOut;
	Result := KeyExportDialog(FKeyDlgPrompt, FKeyIDs, OutputFile,
				  KeyPropsList, TRUE,
				  FExportCompatible, FExportPrivate,
				  GetKeyPropsFlags(FKeyProps), FParentHandle);
	FFileOut := OutputFile;
      end;
      try
	if Result = 0 then begin
	  if Assigned(FOnKeyExported) then FOnKeyExported(KeyPropsList, E, FFileOut);
	end
	else ShowError(FOnFailure, Result, E);
      finally
	KeyPropsList.Free;
      end;
    end;
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;


// TPGPEncode ------------------------------------------------------------------

procedure TPGPEncode.SetFileOutput(Value: TFileOutput);
begin
  FFileOutput := Value;
  inherited FileOutput := (FFileOutput <> FileOutput_Disabled);
  inherited AlwaysQueryOutputFileName := (FFileOutput = FileOutput_AlwaysQuery);
end;

constructor TPGPEncode.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if AOwner is TForm then ParentHandle := TForm(AOwner).Handle;
end;

function TPGPEncode.KeyEncryptBuffer(const DataBuffer: RawByteString; Sign, SMime: Longbool): Longint;
begin
  try
    Result := inherited KeyEncryptBuffer(DataBuffer, Sign, SMime);
    if Result = 0 then begin
      if Assigned(FOnEncoded) then FOnEncoded(OutputBuffer, OutputFileName);
    end
    else ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPEncode.KeyEncryptFile(const FileName: RawByteString; Sign, SMime: Longbool): Longint;
begin
  try
    Result := inherited KeyEncryptFile(FileName, Sign, SMime);
    if Result = 0 then begin
      if Assigned(FOnEncoded) then FOnEncoded(OutputBuffer, OutputFileName);
    end
    else ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPEncode.ConventionalEncryptBuffer(const DataBuffer: RawByteString): Longint;
begin
  try
    Result := inherited ConventionalEncryptBuffer(DataBuffer);
    if Result = 0 then begin
      if Assigned(FOnEncoded) then FOnEncoded(OutputBuffer, OutputFileName);
    end
    else ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPEncode.ConventionalEncryptFile(const FileName: RawByteString): Longint;
begin
  try
    Result := inherited ConventionalEncryptFile(FileName);
    if Result = 0 then begin
      if Assigned(FOnEncoded) then FOnEncoded(OutputBuffer, OutputFileName);
    end
    else ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPEncode.ClearSignBuffer(const DataBuffer: RawByteString; SMime: Longbool): Longint;
begin
  try
    Result := inherited ClearSignBuffer(DataBuffer, SMime);
    if Result = 0 then begin
      if Assigned(FOnEncoded) then FOnEncoded(OutputBuffer, OutputFileName);
    end
    else ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPEncode.ClearSignFile(const FileName: RawByteString; SMime: Longbool): Longint;
begin
  try
    Result := inherited ClearSignFile(FileName, SMime);
    if Result = 0 then begin
      if Assigned(FOnEncoded) then FOnEncoded(OutputBuffer, OutputFileName);
    end
    else ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPEncode.ArmorBuffer(const DataBuffer: RawByteString): Longint;
begin
  try
    Result := inherited ArmorBuffer(DataBuffer);
    if Result = 0 then begin
      if Assigned(FOnEncoded) then FOnEncoded(OutputBuffer, OutputFileName);
    end
    else ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPEncode.ArmorFile(const FileName: RawByteString): Longint;
begin
  try
    Result := inherited ArmorFile(FileName);
    if Result = 0 then begin
      if Assigned(FOnEncoded) then FOnEncoded(OutputBuffer, OutputFileName);
    end
    else ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPEncode.DetachedSignBuffer(const DataBuffer: RawByteString): Longint;
begin
  try
    Result := inherited DetachedSignBuffer(DataBuffer);
    if Result = 0 then begin
      if Assigned(FOnEncoded) then FOnEncoded(OutputBuffer, OutputFileName);
    end
    else ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPEncode.DetachedSignFile(const FileName: RawByteString): Longint;
begin
  try
    Result := inherited DetachedSignFile(FileName);
    if Result = 0 then begin
      if Assigned(FOnEncoded) then FOnEncoded(OutputBuffer, OutputFileName);
    end
    else ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;


// TPGPDecode ------------------------------------------------------------------

procedure TPGPDecode.SetFileOutput(Value: TFileOutput);
begin
  FFileOutput := Value;
  inherited FileOutput := (FFileOutput <> FileOutput_Disabled);
  inherited AlwaysQueryOutputFileName := (FFileOutput = FileOutput_AlwaysQuery);
end;

constructor TPGPDecode.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if AOwner is TForm then ParentHandle := TForm(AOwner).Handle;
end;

function TPGPDecode.AnalyseBuffer(const DataBuffer: RawByteString): Longint;
begin
  try
    Result := inherited AnalyseBuffer(DataBuffer);
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPDecode.AnalyseFile(const FileName: RawByteString): Longint;
begin
  try
    Result := inherited AnalyseFile(FileName);
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPDecode.AnalyseMimeBuffer(const DataBuffer: RawByteString): Longint;
begin
  try
    Result := inherited AnalyseMimeBuffer(DataBuffer);
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPDecode.AnalyseMimeFile(const FileName: RawByteString): Longint;
begin
  try
    Result := inherited AnalyseMimeFile(FileName);
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPDecode.DecodeBuffer(const DataBuffer: RawByteString): Longint;
begin
  IgnoreKnownFlag := Ord(Ignore_ByHexID in FIgnoreKnownKeys) * IgnoreFlag_ByHexID or
		     Ord(Ignore_ByUserID in FIgnoreKnownKeys) * IgnoreFlag_ByUserID;
  try
    Result := inherited DecodeBuffer(DataBuffer);
    if Result = 0 then begin
      if Assigned(FOnDecoded) then FOnDecoded(OutputBuffer, OutputFileName, SigPropsRec, KeyPropsList);
    end
    else ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPDecode.DecodeFile(const FileName: RawByteString): Longint;
begin
  IgnoreKnownFlag := Ord(Ignore_ByHexID in FIgnoreKnownKeys) * IgnoreFlag_ByHexID or
		     Ord(Ignore_ByUserID in FIgnoreKnownKeys) * IgnoreFlag_ByUserID;
  try
    Result := inherited DecodeFile(FileName);
    if Result = 0 then begin
      if Assigned(FOnDecoded) then FOnDecoded(OutputBuffer, OutputFileName, SigPropsRec, KeyPropsList);
    end
    else ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPDecode.DecodeMimeBuffer(const DataBuffer: RawByteString): Longint;
begin
  IgnoreKnownFlag := Ord(Ignore_ByHexID in FIgnoreKnownKeys) * IgnoreFlag_ByHexID or
		     Ord(Ignore_ByUserID in FIgnoreKnownKeys) * IgnoreFlag_ByUserID;
  try
    Result := inherited DecodeMimeBuffer(DataBuffer);
    if Result = 0 then begin
      if Assigned(FOnDecoded) then FOnDecoded(OutputBuffer, OutputFileName, SigPropsRec, KeyPropsList);
    end
    else ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TPGPDecode.DecodeMimeFile(const FileName: RawByteString): Longint;
begin
  IgnoreKnownFlag := Ord(Ignore_ByHexID in FIgnoreKnownKeys) * IgnoreFlag_ByHexID or
		     Ord(Ignore_ByUserID in FIgnoreKnownKeys) * IgnoreFlag_ByUserID;
  try
    Result := inherited DecodeMimeFile(FileName);
    if Result = 0 then begin
      if Assigned(FOnDecoded) then FOnDecoded(OutputBuffer, OutputFileName, SigPropsRec, KeyPropsList);
    end
    else ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;


// Register --------------------------------------------------------------------

procedure Register;
begin
  RegisterComponents('PGP', [TPGPPreferences, TPGPKeyServer, TPGPGetKeyProps, TPGPSetKeyProps,
			     TPGPKeysGenerate, TPGPKeyImport, TPGPKeyExport, TPGPEncode, TPGPDecode]);
end;


begin
  PGPInitErrorCode := pgpBase.PGPInitErrorCode;

  // initialize notification messages ------------------------------------------

  if PGPInitErrorCode = ieNone then begin
    WM_PGP_ReloadPrefs := WM_PGPReloadPrefs;
    WM_PGP_ReloadKeyring := WM_PGPReloadKeyring;
    WM_PGP_ReloadKeyserverPrefs := WM_PGPReloadKeyserverPrefs;
  end;
end.

