{------------------------------------------------------------------------------}
{                                                                              }
{                 This unit is partly based on Steve Heller's                  }
{         SPGP sources available from http://www.oz.net/~srheller/spgp/        }
{                                                                              }
{                Portions created by Michael in der Wiesche are                }
{              Copyright (C) 2001-2010 by Michael in der Wiesche               }
{                                                                              }
{------------------------------------------------------------------------------}
{$J+,Z4}
unit PGPDialogs;

interface

uses
  Windows, Classes, SysUtils, UTF8, PGPStrTools, PGPStringList, KeyPropTypes, X509Types, PrefFuncs, KeyFuncs,
  pgpBase, pgpErrors, pgpPubTypes, pgpUtilities, pgpOptionList, pgpMemoryMgr,
  pgpKeyServer, pgpGroups, pgpKeys, pgpTLS, pgpUI, pgpCL;

type
  TSignOption = (soNo, soEnc, soClear, soDetached);
  TPassphraseCallback = procedure(const Passphrase: PAnsiChar; const UserValue);

{ Tools }
function GetClassName(Window: THandle): RawByteString;
function GetWindowText(Window: THandle): RawByteString;
function IsStandardDialog(Window: THandle): Longbool;
procedure CenterDialog(Dialog, Owner: THandle);
procedure RedrawWindow(Window: THandle);

{ "public" functions }
function PreferencesDialog(PrefsPage, WinHandle: Integer): Integer;
function SelectKeysDialog(const Prompt: RawByteString; var KeyPropsList: TKeyPropsList;
			  SingleKey: Longbool; PropertyFlags, FilterFlags, WinHandle: Integer): Integer;
function SendKeyToServerDialog(const Prompt: RawByteString; var KeyPropsList: TKeyPropsList;
			       KeyPropFlags, WinHandle: Integer): Integer;
function GetKeyFromServerDialog(const Prompt, KeyData: RawByteString; var KeyPropsList: TKeyPropsList;
				KeyPropFlags, WinHandle: Integer): Integer;
function KeyImportDialog(const KeyPrompt, CertPassPrompt, KeyData: RawByteString; CertPassphrase: PAnsiChar;
			 var KeyPropsList: TKeyPropsList; FromFile: Longbool; IgnoreKnownFlag: Integer;
			 KeyPropFlags, WinHandle: Integer): Integer;
function KeyExportDialog(Prompt: RawByteString; const KeyIDs: TRawByteStrings; var KeyData: RawByteString; var KeyPropsList: TKeyPropsList;
			 ToFile, ExportCompatible, ExportPrivate: Longbool; KeyPropFlags, WinHandle: Integer): Integer;
function KeyRevokeDialog(const KeyHexID, PassPrompt: RawByteString; WinHandle: Integer): Integer;
function KeyPassChangeDialog(const KeyHexID, OldPrompt, NewPrompt: RawByteString;
			     MinPassLength, MinPassQuality, WinHandle: Integer): Integer;
function SimplePassphraseDialog(const Title, Prompt: RawByteString; WinHandle: Integer;
				PassphraseCallback: TPassphraseCallback; const CallbackParam): Integer;

{ "private" functions }
function SelectX509CertifiedKeysDialog(Context: pPGPContext; KeySetX509CertifiedKeys: pPGPKeySet;
				       var KeySetSelected: pPGPKeySet; var CertSelected: pPGPKeyDBObj;
				       const Prompt: RawByteString; CertsOnly, SingleKey: Longbool;
				       WinHandle: Integer): Integer;
function SelectAddKeysToKeyRing(Context: pPGPContext; KeysToAdd: pPGPKeySet; KeySetMain: pPGPKeySet;
				var KeyPropsList: TKeyPropsList; const Prompt: RawByteString; SingleKey: Longbool;
				KeyPropFlags, WinHandle: Integer): PGPError;
function RecipientsDialog(Context: pPGPContext; AllKeys: pPGPKeySet; AlwaysDisplayDialog, SMIME: Longbool;
			  var Armor, Textmode, EyesOnly, MIME: PGPUInt32; const DefaultKey: RawByteString;
			  var RecipientKeys: Pointer; const Prompt: RawByteString; WinHandle: Integer): PGPError;
function KeyServerDialog(Context: pPGPContext; KeySetMain: pPGPKeySet; const KeyData: RawByteString;
			 var KeysFound: Pointer; WinHandle: Integer): PGPError;
function CollectRandomDataDialog(Context: pPGPContext; NeededEntropyBits: PGPUInt32; WinHandle: Integer): PGPError;
function PassphraseDialog(Context: pPGPContext; const Title, Prompt: RawByteString;
			  var Passphrase: PAnsiChar; WinHandle: Integer): Integer;
function ConvEncPassphraseDialog(Context: pPGPContext; var Passphrase: PAnsiChar;
				 const Prompt: RawByteString; WinHandle: Integer): PGPError;
function ConvDecPassphraseDialog(Context: pPGPContext; var Passphrase: PAnsiChar;
				 const Prompt: RawByteString; WinHandle: Integer): PGPError;
function KeyPassphraseDialog(Context: pPGPContext; KeySet: pPGPKeySet; var Passphrase: PAnsiChar;
			     const Prompt: RawByteString; WinHandle: Integer): PGPError;
function SigningPassphraseDialog(Context: pPGPContext; KeySetMain: pPGPKeySet;
				 var SigningKey: pPGPKey; var SignKeyPass: PAnsiChar;
				 const DefaultKeyID: RawByteString; FindMatchingKey: Longbool; SignOption: TSignOption;
				 var Armor, MIME, RipeMD: PGPUInt32; const Prompt: RawByteString; WinHandle: Integer): PGPError;
function DecryptionPassphraseDialog(Context: pPGPContext; RecipientKeys: pPGPKeySet;
				    KeyIDCount: PGPUInt32; const KeyIDList: pPGPKeyID7;
				    var DecryptionKey: pPGPKey; var DecryptKeyPass: PAnsiChar;
				    FindMatchingKey: Longbool; const Prompt: RawByteString; WinHandle: Integer): PGPError;
function ConfirmationPassphraseDialog(Context: pPGPContext; var Passphrase: PAnsiChar;
				      MinPassLength, MinPassQuality: Integer;
				      ShowPassphraseQuality: Longbool;
				      const Prompt: RawByteString; WinHandle: Integer): PGPError;


implementation

const
  SYS_DLG = '#32770';
  PGP_DLG = 'PGP';

function GetClassName(Window: THandle): RawByteString;
begin
  SetString(Result, NIL, $FFFF);
  Delete(Result, succ(GetClassNameA(Window, Pointer(Result), $FFFF)), MAXINT);
end;

function GetWindowText(Window: THandle): RawByteString;
var
  iLen: Integer;
begin
  iLen := succ(GetWindowTextLengthA(Window));
  SetString(Result, NIL, iLen);
  iLen := GetWindowTextA(Window, Pointer(Result), iLen);
  if (iLen > 0) and (Result[iLen] = DOT) then dec(iLen);
  Delete(Result, succ(iLen), MAXINT);
end;

function IsStandardDialog(Window: THandle): Longbool;
var
  sClassName: RawByteString;
begin
  sClassName := GetClassName(Window);
  Result := (sClassName = SYS_DLG) or (sClassName = PGP_DLG);
end;

procedure CenterDialog(Dialog, Owner: THandle);
var
  rOwner, rDialog: TRect;
begin
  GetWindowRect(Owner, rOwner);
  with rOwner do begin
    dec(Right, pred(Left));
    dec(Bottom, pred(Top));
  end;
  GetWindowRect(Dialog, rDialog);
  with rDialog do begin
    dec(Right, pred(Left));
    dec(Bottom, pred(Top));
    Left := rOwner.Left + (rOwner.Right - Right) div 2;
    Top := rOwner.Top + (rOwner.Bottom - Bottom) div 2;
    MoveWindow(Dialog, Left, Top, Right, Bottom, FALSE);
  end;
end;

procedure RedrawWindow(Window: THandle);
begin
  Windows.RedrawWindow(Window, NIL, 0, RDW_UPDATENOW or RDW_ALLCHILDREN);
end;

function PreferencesDialog(PrefsPage, WinHandle: Integer): Integer;
var
  Context	 : pPGPContext;
  KeySetMain	 : pPGPKeySet;
begin
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  Result := kPGPError_PrefNotFound;
  try
    if SDKVersion >= PGP95X then begin
      case PrefsPage of
	kPGPclGeneralPrefs	: ;
	kPGPclKeyringPrefs	: PrefsPage := kPGPclKeysPrefs95;
	kPGPclEmailPrefs	: PrefsPage := kPGPclMessagingPrefs95;
	kPGPclHotkeyPrefs	: PrefsPage := kPGPclHotkeyPrefs95;
	kPGPclAdvancedPrefs	: PrefsPage := kPGPclDiskPrefs95;
      else
	Exit;
      end;
    end
    else if SDKVersion >= PGP9XX then begin
      case PrefsPage of
	kPGPclGeneralPrefs	: ;
	kPGPclKeyringPrefs	: PrefsPage := kPGPclKeysPrefs90;
	kPGPclEmailPrefs	: PrefsPage := kPGPclMessagingPrefs90;
	kPGPclHotkeyPrefs	: PrefsPage := kPGPclHotkeysPrefs90;
	kPGPclAdvancedPrefs	: PrefsPage := kPGPclWipingPrefs90;
      else
	Exit;
      end;
    end;
    Result := PGPclPreferences(Context, WinHandle, PrefsPage, KeySetMain);
    RedrawWindow(WinHandle);
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function SelectKeysDialog(const Prompt: RawByteString; var KeyPropsList: TKeyPropsList;
			  SingleKey: Longbool; PropertyFlags, FilterFlags, WinHandle: Integer): Integer;
var
  Context	 : pPGPContext;
  KeySetMain	 : pPGPKeySet;
  KeyFilter	 : pPGPFilter;
  KeySetFiltered : pPGPKeySet;
  KeySetSelected : pPGPKeySet;
  KeyCount	 : PGPUInt32;
begin
  KeyFilter := NIL;
  KeySetFiltered := NIL;
  KeySetSelected := NIL;
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  try
    if FilterFlags <> KeyFilterFlag_AllKeys then begin
      Result := GetKeyFilter(Context, FilterFlags, KeyFilter);
      if Result <> 0 then Exit;
      try
	Result := PGPFilterKeySet(KeySetMain, KeyFilter, KeySetFiltered);
      finally
	PGPFreeFilter(KeyFilter);
      end;
      if Result <> 0 then Exit;
    end
    else KeySetFiltered := KeySetMain;
    try
      Result := PGPclSelectKeys(Context, WinHandle, Prompt, KeySetFiltered, KeySetMain,
				Ord(SingleKey = TRUE) * PGPCL_SINGLESELECTION, KeySetSelected);
      RedrawWindow(WinHandle);
      if Result <> 0 then Exit;
    finally
      if KeySetFiltered <> KeySetMain then PGPFreeKeySet(KeySetFiltered);
    end;
    try
      Result := PGPCountKeys(KeySetSelected, KeyCount);
      if Result <> 0 then Exit;
      if KeyCount <> 0 then begin
	Result := GetKeySetProps(Context, KeySetSelected, KeyPropsList, PropertyFlags,
				 KeyFilterFlag_AllKeys, UserID_Ordering);
	if Result > 0 then Result := 0;
      end;
    finally
      PGPFreeKeySet(KeySetSelected);
    end;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function InitKeyServer(Context: pPGPContext; var ServerThreadStorage: pPGPKeyServerThreadStorage;
		       var KeyServerSpecs: pPGPKeyServerSpec; var KeyServerCount: PGPUInt32): PGPError;
var
  KeyServerList	 : TRawByteStrings;
  KeyServerIndex : Integer;
begin
  ServerThreadStorage := NIL;
  KeyServerSpecs := NIL;
  KeyServerCount := 0;
  KeyServerList := NIL;
  Result := PGPKeyServerInit;
  if Result <> 0 then Exit;
  Result := PGPKeyServerCreateThreadStorage(ServerThreadStorage);
  if Result <> 0 then Exit;
  try
    Result := GetServerList(KeyServerList);
    if (Result = 0) and (KeyServerList <> NIL) then with KeyServerList do begin
      KeyServerSpecs := AllocMem(SizeOf(TPGPKeyServerSpec) * Count);
      if KeyServerSpecs <> NIL then begin
	for KeyServerIndex := 0 to pred(Count) do begin
	  if SDKVersion >= PGP7XX then begin
	    Result := PGPNewKeyServer(Context, kPGPKeyServerClass_PGP, KeyServerSpecs.Spec.PGP7[KeyServerIndex].Server,
				      PGPONetURL(Context, PAnsiChar(Strings[KeyServerIndex])), PGPOLastOption(Context));
	  end
	  else begin
	    Result := PGPNewKeyServer(Context, kPGPKeyServerClass_PGP, KeyServerSpecs.Spec.PGP6[KeyServerIndex].Server,
				      PGPONetURL(Context, PAnsiChar(Strings[KeyServerIndex])), PGPOLastOption(Context));
	  end;
	  if Result <> 0 then Exit;
	  inc(KeyServerCount);
	end;
      end
      else Result := kPGPError_OutOfMemory;
    end;
  finally
    KeyServerList.Free;
  end;
end;

procedure FreeKeyServer(ServerThreadStorage: pPGPKeyServerThreadStorage;
			KeyServerSpecs: pPGPKeyServerSpec; KeyServerCount: PGPUInt32);
begin
  try
    for KeyServerCount := pred(KeyServerCount) downto 0 do begin
      if SDKVersion >= PGP7XX then
	PGPFreeKeyServer(KeyServerSpecs.Spec.PGP7[KeyServerCount].Server)
      else PGPFreeKeyServer(KeyServerSpecs.Spec.PGP6[KeyServerCount].Server);
    end;
  finally
    if KeyServerSpecs <> NIL then FreeMem(KeyServerSpecs);
    PGPKeyServerDisposeThreadStorage(ServerThreadStorage);
    PGPKeyServerCleanup;
  end;
end;

function SendKeyToServerDialog(const Prompt: RawByteString; var KeyPropsList: TKeyPropsList;
			       KeyPropFlags, WinHandle: Integer): Integer;
var
  Context	 : pPGPContext;
  KeySetMain	 : pPGPKeySet;
  tlsContext	 : pPGPtlsContext;
  KeySetToSend	 : pPGPKeySet;
  ThreadStorage	 : pPGPKeyServerThreadStorage;
  KeyServerSpecs : pPGPKeyServerSpec;
  KeyServerCount : PGPUInt32;
  KeySetFailed	 : pPGPKeySet;
begin
  KeySetToSend := NIL;
  KeySetFailed := NIL;
  tlsContext := NIL;
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  try
    Result := PGPclSelectKeys(Context, WinHandle, Prompt, KeySetMain, KeySetMain, 0, KeySetToSend);
    RedrawWindow(WinHandle);
    if Result <> 0 then Exit;
    try
      Result := PGPNewTLSContext(Context, tlsContext);
      if Result <> 0 then Exit;
      try
	Result := InitKeyServer(Context, ThreadStorage, KeyServerSpecs, KeyServerCount);
	if Result <> 0 then Exit;
	try
	  Result := PGPSendToKeyServerDialog(Context, KeyServerSpecs^, tlsContext, KeySetToSend, KeySetFailed,
					     PGPOUIParentWindowHandle(Context, WinHandle), PGPOLastOption(Context));
	  RedrawWindow(WinHandle);
	finally
	  FreeKeyServer(ThreadStorage, KeyServerSpecs, KeyServerCount);
	end;
	if (KeySetFailed <> NIL) and (KeySetToSend <> NIL) then PGPRemoveKeys(KeySetFailed, KeySetToSend);
	GetKeySetProps(Context, KeySetToSend, KeyPropsList, KeyPropFlags, KeyFilterFlag_AllKeys, UserID_Ordering);
      finally
	PGPFreeTLSContext(tlsContext);
      end;
    finally
      PGPFreeKeySet(KeySetToSend);
      PGPFreeKeySet(KeySetFailed);
    end;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function SearchKeyServerOptions(Context: pPGPContext; const KeyData: RawByteString; IncludeSubKeys, IncludeUTF8: Longbool;
				WinHandle: Integer; var OptionList: pPGPOptionList): Integer;
var
  KeyFilter	 : pPGPFilter;
begin
  OptionList := NIL;
  Result := GetKeyFilterByAnyID(Context, KeyData, TRUE, IncludeUTF8, KeyFilter);
  if Result = 0 then begin
    try
      Result := PGPBuildOptionList(Context, OptionList,
	[
	  PGPOKeyServerSearchFilter(Context, KeyFilter),
	  PGPOUIParentWindowHandle(Context, WinHandle)
	]);
    finally
      PGPFreeFilter(KeyFilter);
    end;
  end;
end;

function GetKeyFromServerDialog(const Prompt, KeyData: RawByteString; var KeyPropsList: TKeyPropsList;
				KeyPropFlags, WinHandle: Integer): Integer;
var
  Context	 : pPGPContext;
  KeySetMain	 : pPGPKeySet;
  tlsContext	 : pPGPtlsContext;
  ThreadStorage	 : pPGPKeyServerThreadStorage;
  KeyServerSpecs : pPGPKeyServerSpec;
  KeyServerCount : PGPUInt32;
  OptionList	 : pPGPOptionList;
  KeysFound	 : Pointer;
begin
  tlsContext := NIL;
  OptionList := NIL;
  KeysFound := NIL;
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  try
    Result := PGPNewTLSContext(Context, tlsContext);
    if Result <> 0 then Exit;
    try
      Result := InitKeyServer(Context, ThreadStorage, KeyServerSpecs, KeyServerCount);
      if Result <> 0 then Exit;
      try
	Result := SearchKeyServerOptions(Context, KeyData, FALSE, FALSE, WinHandle, OptionList);
	if Result <> 0 then Exit;
	try
	  Result := PGPSearchKeyServerDialog(Context, KeyServerCount, KeyServerSpecs, tlsContext, PGPFalse,
					     KeysFound, OptionList, PGPOLastOption(Context));
	  RedrawWindow(WinHandle);
	finally
	  PGPFreeOptionList(OptionList);
	end;
	if Result <> 0 then Exit;
	if SDKVersion >= PGP7XX then begin
	  try
	    Result := SelectAddKeysToKeyRing(Context, PGPPeekKeyDBRootKeySet(pPGPKeyDB(KeysFound)),
					     KeySetMain, KeyPropsList, Prompt, FALSE, KeyPropFlags, WinHandle);
	  finally
	    PGPFreeKeyDB(pPGPKeyDB(KeysFound));
	  end;
	end
	else begin
	  try
	    Result := SelectAddKeysToKeyRing(Context, pPGPKeySet(KeysFound), KeySetMain,
					     KeyPropsList, Prompt, FALSE, KeyPropFlags, WinHandle);
	  finally
	    PGPFreeKeySet(pPGPKeySet(KeysFound));
	  end;
	end;
      finally
	FreeKeyServer(ThreadStorage, KeyServerSpecs, KeyServerCount);
      end;
    finally
      PGPFreeTLSContext(tlsContext);
    end;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function KeyImportDialog(const KeyPrompt, CertPassPrompt, KeyData: RawByteString; CertPassphrase: PAnsiChar;
			 var KeyPropsList: TKeyPropsList; FromFile: Longbool; IgnoreKnownFlag: Integer;
			 KeyPropFlags, WinHandle: Integer): Integer;
type
  TInputFormat = (ifNone, ifPKCS_12, ifPKCS_7, ifPEM, ifKEY);
var
  Context	 : pPGPContext;
  KeySetMain	 : pPGPKeySet;
  FileHandle	 : THandle;
  InputBuffer	 : RawByteString;
  InputSize	 : Longint;
  InputFormat	 : TInputFormat;
  OptionList	 : pPGPOptionList;
  PosKeyData	 : Longint;
  LenKeyData	 : Longint;
  KeysToImport	 : Pointer;
  KeySetSelected : pPGPKeySet;
  OptionListCopy : pPGPOptionList;
  KeysToAdd	 : pPGPKeyDB;
  CertsToImport	 : pPGPKeySet;
begin
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  OptionList := NIL;
  KeysToImport := NIL;
  KeySetSelected := NIL;
  try
    if FromFile then begin
      FileHandle := FileOpen(String(KeyData), fmOpenReadWrite);
      if FileHandle <> INVALID_HANDLE_VALUE then begin
	try
	  InputSize := FileSeek(FileHandle, 0, FILE_END);
	  SetString(InputBuffer, NIL, InputSize);
	  FileSeek(FileHandle, 0, FILE_BEGIN);
	  FileRead(FileHandle, Pointer(InputBuffer)^, InputSize);
	finally
	  FileClose(FileHandle);
	end;
      end
      else InputSize := 0;
    end
    else begin
      InputSize := Length(KeyData);
      InputBuffer := KeyData;
    end;
    if SDKVersion >= PGP7XX then
      InputFormat := ifNone
    else InputFormat := ifKEY;
    repeat
      PosKeyData := 1;
      LenKeyData := InputSize;
      case InputFormat of
	ifNone: begin
	  InputFormat := ifPKCS_12;
	  Result := PGPBuildOptionList(Context, OptionList, [PGPOInputFormat(Context, kPGPInputFormat_PKCS12)]);
	end;
	ifPKCS_12: begin
	  InputFormat := ifPKCS_7;
	  Result := PGPBuildOptionList(Context, OptionList, [PGPOInputFormat(Context, kPGPInputFormat_X509DataInPKCS7)]);
	end;
	ifPKCS_7 : begin
	  InputFormat := ifPEM;
	  PosKeyData := FirstStrPos(X509Cert_Beg, InputBuffer);
	  if PosKeyData > 0 then
	    LenKeyData := InputSize - pred(PosKeyData)
	  else Continue;
	  Result := PGPBuildOptionList(Context, OptionList, [PGPOInputFormat(Context, kPGPInputFormat_PEMEncodedX509Cert)]);
	end
      else
	InputFormat := ifKEY;
	Result := PGPBuildOptionList(Context, OptionList, [PGPONullOption(Context)]);
      end;
      if Result <> 0 then Exit;
      try
	Result := PGPAppendOptionList(OptionList, [PGPOInputBuffer(Context, @InputBuffer[PosKeyData], LenKeyData)]);
	if Result <> 0 then Exit;
	if SDKVersion >= PGP7XX then begin
	  Result := PGPImport(Context, KeysToImport, OptionList, PGPOLastOption(Context));
	  if Result = kPGPError_BadPassphrase then begin
	    repeat
	      Result := PGPCopyOptionList(OptionList, OptionListCopy);
	      if Result <> 0 then Exit;
	      try
		if CertPassPrompt <> '' then begin
		  CertPassphrase := NIL;
		  Result := PassphraseDialog(Context, PGP_DLG, CertPassPrompt, CertPassphrase, WinHandle);
		  if Result <> 0 then Exit;
		end;
		try
		  Result := PGPAppendOptionList(OptionListCopy, [PGPOPassphrase(Context, CertPassphrase)]);
		  if Result <> 0 then Exit;
		  Result := PGPImport(Context, KeysToImport, OptionListCopy, PGPOLastOption(Context));
		finally
		  if CertPassPrompt <> '' then PGPFreeData(CertPassphrase);
		end;
	      finally
		PGPFreeOptionList(OptionListCopy);
	      end;
	    until (Result <> kPGPError_BadPassphrase) or (CertPassPrompt = '');
	  end
	  else if InputFormat = ifPEM then begin
	    KeySetSelected := PGPPeekKeyDBRootKeySet(pPGPKeyDB(KeysToImport));
	    repeat
	      PosKeyData := NextStrPos(X509Cert_Beg, InputBuffer, PosKeyData + Length(X509Cert_Beg) + Length(CRLF));
	      if PosKeyData > 0 then begin
		LenKeyData := InputSize - pred(PosKeyData);
		PGPFreeOptionList(OptionList);
		OptionList := NIL;
		Result := PGPBuildOptionList(Context, OptionList,
		  [
		    PGPOInputFormat(Context, kPGPInputFormat_PEMEncodedX509Cert),
		    PGPOInputBuffer(Context, @InputBuffer[PosKeyData], LenKeyData)
		  ]);
		if Result <> 0 then Exit;
		KeysToAdd := NIL;
		Result := PGPImport(Context, KeysToAdd, OptionList, PGPOLastOption(Context));
		if Result <> 0 then Exit;
		try
		  Result := PGPCopyKeys(PGPPeekKeyDBRootKeySet(KeysToAdd), KeysToImport, KeySetSelected);
		  if Result <> 0 then Exit;
		finally
		  PGPFreeKeyDB(KeysToAdd);
		end;
	      end;
	    until PosKeyData = 0;
	  end;
	end
	else Result := PGPImportKeySet(Context, KeysToImport, OptionList, PGPOLastOption(Context));
	if Result = 0 then begin
	  try
	    if SDKVersion >= PGP7XX then begin
	      if KeySetSelected = NIL then KeySetSelected := PGPPeekKeyDBRootKeySet(pPGPKeyDB(KeysToImport));
	      Result := GetExclusiveKeySet(KeySetSelected, KeySetMain, Context, IgnoreKnownFlag);
	      if Result <> 0 then Exit;
	      if KeyPrompt = '' then begin
		Result := AddKeysToKeyRing(Context, KeySetMain, KeySetSelected, KeyPropsList, KeyPropFlags);
		if Result = 0 then
		  Result := kPGPError_InvalidInputFormat
		else if Result > 0 then Result := 0;
	      end
	      else begin
		CertsToImport := NIL;
		if InputFormat = ifKEY then
		  CertsToImport := KeySetSelected
		else begin
		  Result := GetCertKeySetFromKeySet(KeySetSelected, FALSE, CertsToImport);
		  if Result <> 0 then Exit;
		end;
		try
		  Result := SelectAddKeysToKeyRing(Context, CertsToImport, KeySetMain, KeyPropsList,
						   KeyPrompt, FALSE, KeyPropFlags, WinHandle);
		finally
		  if InputFormat <> ifKEY then PGPFreeKeySet(CertsToImport);
		end;
	      end;
	    end
	    else begin
	      Result := GetExclusiveKeySet(pPGPKeySet(KeysToImport), KeySetMain, Context, IgnoreKnownFlag);
	      if Result <> 0 then Exit;
	      if KeyPrompt = '' then begin
		Result := AddKeysToKeyRing(Context, KeySetMain, KeysToImport, KeyPropsList, KeyPropFlags);
		if Result > 0 then Result := 0;
	      end
	      else begin
		Result := SelectAddKeysToKeyRing(Context, pPGPKeySet(KeysToImport), KeySetMain,
						 KeyPropsList, KeyPrompt, FALSE, KeyPropFlags, WinHandle);
	      end;
	    end;
	  finally
	    if SDKVersion >= PGP7XX then
	      PGPFreeKeyDB(pPGPKeyDB(KeysToImport))
	    else PGPFreeKeySet(pPGPKeySet(KeysToImport));
	    KeysToImport := NIL;
	  end;
	end
	else if (InputFormat < ifKEY) and (Result <> kPGPError_BadPassphrase) then Result := kPGPError_InvalidInputFormat;
      finally
	PGPFreeOptionList(OptionList);
	OptionList := NIL;
      end;
    until (Result <> kPGPError_InvalidInputFormat) or (InputFormat = ifKEY);
    if (Result < 0) and (Result >= -40) then Result := kPGPError_InvalidInputFormat;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function KeyExportDialog(Prompt: RawByteString; const KeyIDs: TRawByteStrings; var KeyData: RawByteString; var KeyPropsList: TKeyPropsList;
			 ToFile, ExportCompatible, ExportPrivate: Longbool; KeyPropFlags, WinHandle: Integer): Integer;
var
  Context	 : pPGPContext;
  KeySetMain	 : pPGPKeySet;
  KeySetSelected : pPGPKeySet;
  OptionList	 : pPGPOptionList;
  FileSpec	 : pPGPFileSpec;
  KeyCount	 : PGPUInt32;
  ExportFormat	 : PGPExportFormat;
  KeyOutput	 : PAnsiChar;
  KeyLength	 : PGPUInt32;
begin
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  KeySetSelected := NIL;
  OptionList := NIL;
  FileSpec := NIL;
  try
    if Prompt = '' then
      Result := GetKeySetByAnyIDs(Context, KeySetMain, KeyIDs.CommaText, KeySetSelected)
    else begin
      Result := PGPclSelectKeys(Context, WinHandle, Prompt, KeySetMain, KeySetMain, 0, KeySetSelected);
      RedrawWindow(WinHandle);
    end;
    if Result <> 0 then Exit;
    try
      Result := PGPCountKeys(KeySetSelected, KeyCount);
      if Result <> 0 then Exit;
      if KeyCount <> 0 then begin
	if ExportCompatible then
	  ExportFormat := kPGPExportFormat_Basic
	else ExportFormat := kPGPExportFormat_Complete;
	Result := PGPBuildOptionList(Context, OptionList,
	  [
	    PGPOVersionString(Context, MyVersion),
	    PGPOExportFormat(Context, ExportFormat),
	    PGPOExportPrivateKeys(Context, PGPBoolean(ExportPrivate = TRUE)),
	    PGPOExportPrivateSubKeys(Context, PGPBoolean(ExportPrivate = TRUE))
	  ]);
	if Result <> 0 then Exit;
	try
	  if ToFile then begin
	    Result := PGPNewFileSpecFromFullPath(Context, PAnsiChar(KeyData), FileSpec);
	    if Result <> 0 then Exit;
	    try
	      if Result <> 0 then Exit;
	      Result := PGPAppendOptionList(OptionList, [PGPOOutputFile(Context, FileSpec)]);
	      if Result <> 0 then Exit;
	      if SDKVersion >= PGP7XX then begin
		Result := PGPAppendOptionList(OptionList, [PGPOExportKeySet(Context, KeySetSelected)]);
		if Result = 0 then Result := PGPExport(Context, OptionList, PGPOLastOption(Context));
	      end
	      else Result := PGPExportKeySet(KeySetSelected, OptionList, PGPOLastOption(Context));
	      if Result = 0 then Result := GetKeySetProps(Context, KeySetSelected, KeyPropsList, KeyPropFlags,
							  KeyFilterFlag_AllKeys, UserID_Ordering);
	      if Result > 0 then Result := 0;
	    finally
	      PGPFreeFileSpec(FileSpec);
	    end;
	  end
	  else begin
	    Result := PGPAppendOptionList(OptionList, [PGPOAllocatedOutputBuffer(Context, KeyOutput, $FFFFFFFF, KeyLength)]);
	    if Result <> 0 then Exit;
	    if SDKVersion >= PGP7XX then begin
	      Result := PGPAppendOptionList(OptionList, [PGPOExportKeySet(Context, KeySetSelected)]);
	      if Result = 0 then Result := PGPExport(Context, OptionList, PGPOLastOption(Context));
	    end
	    else Result := PGPExportKeySet(KeySetSelected, OptionList, PGPOLastOption(Context));
	    if (Result = 0) and (KeyLength <> 0) then begin
	      try
		SetString(KeyData, NIL, KeyLength);
		StrPLCopy(PAnsiChar(KeyData), KeyOutput, KeyLength);
	      finally
		PGPFreeData(KeyOutput);
	      end;
	    end
	    else KeyData := '';
	    if Result = 0 then Result := GetKeySetProps(Context, KeySetSelected, KeyPropsList, KeyPropFlags,
							KeyFilterFlag_AllKeys, UserID_Ordering);
	    if Result > 0 then Result := 0;
	  end;
	finally
	  PGPFreeOptionList(OptionList);
	end;
      end;
    finally
      PGPFreeKeySet(KeySetSelected);
    end;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function KeyRevokeDialog(const KeyHexID, PassPrompt: RawByteString; WinHandle: Integer): Integer;
var
  Context	 : pPGPContext;
  KeySetMain	 : pPGPKeySet;
  KeySetFound	 : pPGPKeySet;
  Passphrase	 : PAnsiChar;
begin
  KeySetFound := NIL;
  Passphrase := NIL;
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  try
    Result := GetKeySetByAnyIDs(Context, KeySetMain, KeyHexID, KeySetFound);
    if Result <> 0 then Exit;
    try
      Result := KeyPassphraseDialog(Context, KeySetFound, Passphrase, PassPrompt, WinHandle);
      RedrawWindow(WinHandle);
      if Result <> 0 then Exit;
    finally
      PGPFreeKeySet(KeySetFound);
    end;
    try
      KeyRevoke(PAnsiChar(KeyHexID), Passphrase);
    finally
      PGPFreeData(Passphrase);
    end;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function KeyPassChangeDialog(const KeyHexID, OldPrompt, NewPrompt: RawByteString;
			     MinPassLength, MinPassQuality, WinHandle: Integer): Integer;
var
  Context	 : pPGPContext;
  KeySetMain	 : pPGPKeySet;
  KeySetFound	 : pPGPKeySet;
  OldPassphrase	 : PAnsiChar;
  NewPassphrase  : PAnsiChar;
begin
  KeySetFound := NIL;
  OldPassphrase := NIL;
  NewPassphrase := NIL;
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  try
    Result := GetKeySetByAnyIDs(Context, KeySetMain, KeyHexID, KeySetFound);
    if Result <> 0 then Exit;
    try
      Result := KeyPassphraseDialog(Context, KeySetFound, OldPassphrase, OldPrompt, WinHandle);
      RedrawWindow(WinHandle);
      if Result <> 0 then Exit;
    finally
      PGPFreeKeySet(KeySetFound);
    end;
    try
      Result := ConfirmationPassphraseDialog(Context, NewPassphrase,
					     MinPassLength, MinPassQuality, TRUE,
					     NewPrompt, WinHandle);
      RedrawWindow(WinHandle);
      if Result <> 0 then Exit;
      try
	ChangePassphrase(PAnsiChar(KeyHexID), OldPassphrase, NewPassphrase);
      finally
	PGPFreeData(NewPassphrase);
      end;
    finally
      PGPFreeData(OldPassphrase);
    end;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function SimplePassphraseDialog(const Title, Prompt: RawByteString; WinHandle: Integer;
				PassphraseCallback: TPassphraseCallback; const CallbackParam): Integer;
var
  Context	 : pPGPContext;
  Passphrase	 : PAnsiChar;
begin
  Result := kPGPError_BadParams;
  if not Assigned(PassphraseCallback) then Exit;
  Result := PGPNewContext(kPGPsdkAPIVersion, Context);
  if Result <> 0 then Exit;
  Passphrase := NIL;
  try
    Result := PassphraseDialog(Context, Title, Prompt, Passphrase, WinHandle);
    if Result = 0 then PassphraseCallback(Passphrase, CallbackParam);
  finally
    PGPFreeData(Passphrase);
    PGPFreeContext(Context);
  end;
end;

function SelectX509CertifiedKeysDialog(Context: pPGPContext; KeySetX509CertifiedKeys: pPGPKeySet;
				       var KeySetSelected: pPGPKeySet; var CertSelected: pPGPKeyDBObj;
				       const Prompt: RawByteString; CertsOnly, SingleKey: Longbool;
				       WinHandle: Integer): Integer;
begin
  KeySetSelected := NIL;
  if (SDKVersion >= PGP81X) and CertsOnly and SingleKey then begin
    Result := PGPclSelectX509Cert(Context, WinHandle, Prompt, KeySetX509CertifiedKeys,
				  kPGPclDefaultX509SelectionFlags, CertSelected);
    if Result = 0 then Result := PGPNewSingletonKeySet(CertSelected, KeySetSelected);
  end
  else begin
    Result := PGPclSelectKeys(Context, WinHandle, Prompt, KeySetX509CertifiedKeys, NIL,
			      Ord(SingleKey = TRUE) * PGPCL_SINGLESELECTION, KeySetSelected);
  end;
  RedrawWindow(WinHandle);
end;

function SelectAddKeysToKeyRing(Context: pPGPContext; KeysToAdd: pPGPKeySet; KeySetMain: pPGPKeySet;
				var KeyPropsList: TKeyPropsList; const Prompt: RawByteString; SingleKey: Longbool;
				KeyPropFlags, WinHandle: Integer): PGPError;
var
  KeyCount	 : PGPUInt32;
  KeyDBTemp	 : pPGPKeyDB;
  KeySetTemp	 : pPGPKeySet;
  KeySetSelected : pPGPKeySet;
  KeySetSelToAdd : pPGPKeySet;
begin
  KeySetSelected := NIL;
  KeySetSelToAdd := NIL;
  Result := PGPCountKeys(KeysToAdd, KeyCount);
  if Result <> 0 then Exit;
  if KeyCount <> 0 then begin
    if SDKVersion >= PGP7XX then begin
      KeyDBTemp := NIL;
      Result := PGPNewKeyDB(Context, KeyDBTemp);
      if Result <> 0 then Exit;
      try
	KeySetTemp := NIL;
	Result := PGPCopyKeys(KeySetMain, KeyDBTemp, KeySetTemp);
	if Result <> 0 then Exit;
	PGPFreeKeySet(KeySetTemp);
	Result := PGPCopyKeys(KeysToAdd, KeyDBTemp, KeySetTemp);
	if Result <> 0 then Exit;
	try
	  // KeyDBTemp and KeySetTemp are needed for displaying "web of trust" validities
	  Result := PGPclSelectKeys(Context, WinHandle, Prompt, KeySetTemp, NIL,
	  			    Ord(SingleKey = TRUE) * PGPCL_SINGLESELECTION, KeySetSelected);
	  RedrawWindow(WinHandle);
	  if Result <> 0 then Exit;
	  try
	    Result := PGPCopyKeys(KeySetSelected, PGPPeekKeySetKeyDB(KeySetMain), KeySetSelToAdd);
	  finally
	    PGPFreeKeySet(KeySetSelected);
	  end;
	finally
	  PGPFreeKeySet(KeySetTemp);
	end;
      finally
	PGPFreeKeyDB(KeyDBTemp);
      end;
    end
    else begin
      Result := PGPclSelectKeys(Context, WinHandle, Prompt, KeysToAdd, KeySetMain,
      				Ord(SingleKey = TRUE) * PGPCL_SINGLESELECTION, KeySetSelected);
      RedrawWindow(WinHandle);
      if Result <> 0 then Exit;
      KeySetSelToAdd := KeySetSelected;
      Result := PGPAddKeys(KeySetSelToAdd, KeySetMain);
    end;
    if Result <> 0 then Exit;
    try
      Result := KeyRings.UpdateKeyRings;
      if Result <> 0 then Exit;
      Result := GetKeySetProps(Context, KeySetSelToAdd, KeyPropsList, KeyPropFlags,
			       KeyFilterFlag_AllKeys, UserID_Ordering);
      if Result > 0 then Result := 0;
    finally
      PGPFreeKeySet(KeySetSelToAdd);
    end;
  end
  else Result := kPGPError_PublicKeyNotFound;
end;

function RecipientsDialog(Context: pPGPContext; AllKeys: pPGPKeySet; AlwaysDisplayDialog, SMime: Longbool;
			  var Armor, Textmode, EyesOnly, MIME: PGPUInt32; const DefaultKey: RawByteString;
			  var RecipientKeys: Pointer; const Prompt: RawByteString; WinHandle: Integer): PGPError;
var
  Utf8Prompt	 : RawByteString;
  GroupFileSpec	 : pPGPFileSpec;
  GroupSet	 : pPGPGroupSet;
  GroupOption	 : pPGPOptionList;
  FormOptions	 : pPGPOptionList;
  OptionList	 : pPGPOptionList;
  RecipientSpec	 : TPGPRecipientSpec;
begin
  if SDKVersion >= PGP9XX then
    Utf8Prompt := AnsiToUtf8(Prompt)
  else Utf8Prompt := Prompt;
  RecipientKeys := NIL;
  GroupFileSpec := NIL;
  GroupSet := NIL;
  OptionList := NIL;
  FormOptions := NIL;
  if SMime then
    Result := kPGPError_FeatureNotAvailable
  else Result := PGPNewFileSpecFromFullPath(Context, PAnsiChar(KeyRings.GroupsFile), GroupFileSpec);
  try
    if Result = 0 then Result := PGPNewGroupSetFromFile(Context, GroupFileSpec, GroupSet);
    try
      if Result = 0 then
	GroupOption := PGPOUIRecipientGroups(Context, GroupSet)
      else GroupOption := PGPONullOption(Context);
      if SMime then
	FormOptions := PGPONullOption(Context)
      else begin
	Result := PGPBuildOptionList(Context, FormOptions,
	  [
	    PGPOUICheckbox(Context, 801, '&Armor', NIL, Armor, Armor, PGPOLastOption(Context)),
	    PGPOUICheckbox(Context, 802, '&MIME', NIL, MIME, MIME, PGPOLastOption(Context)),
	    PGPOUICheckbox(Context, 803, '&Textmode', NIL, Textmode, Textmode, PGPOLastOption(Context)),
	    PGPOUICheckbox(Context, 804, 'For your &eyes only', NIL, EyesOnly, EyesOnly, PGPOLastOption(Context))
	  ]);
	if Result <> 0 then Exit;
      end;
      try
	Result := PGPBuildOptionList(Context, OptionList,
	  [
	    PGPOUIDialogOptions(Context, FormOptions, PGPOLastOption(Context)),
	    PGPOUIDisplayMarginalValidity(Context, PGPTrue),
	    GroupOption,
	    PGPOUIWindowTitle(Context, PGP_DLG),
	    PGPOUIDialogPrompt(Context, PAnsiChar(Utf8Prompt)),
	    PGPOUIParentWindowHandle(Context, WinHandle)
	  ]);
	if Result <> 0 then Exit;
	if DefaultKey <> '' then begin
	  FillChar(RecipientSpec, SizeOf(TPGPRecipientSpec), 0);
	  RecipientSpec.AType := kPGPRecipientSpecType_Key;
	  Result := GetKeyByHexID(AllKeys, DefaultKey, RecipientSpec.ID.Key);
	  if Result = 0 then PGPAppendOptionList(OptionList, [PGPOUIDefaultRecipients(Context, 1, @RecipientSpec)]);
	end;
	try
	  if SDKVersion >= PGP7XX then begin
	    Result := PGPRecipientDialog(Context, PGPPeekKeySetKeyDB(AllKeys), PGPBoolean(Ord(AlwaysDisplayDialog = TRUE)),
					 pPGPKeyDB(RecipientKeys), OptionList, PGPOLastOption(Context));
	  end
	  else begin
	    Result := PGPRecipientDialog(Context, AllKeys, PGPBoolean(Ord(AlwaysDisplayDialog = TRUE)),
					 pPGPKeySet(RecipientKeys), OptionList, PGPOLastOption(Context));
	  end;
	  RedrawWindow(WinHandle);
	finally
	  PGPFreeOptionList(OptionList);
	end;
      finally
	if not SMime then PGPFreeOptionList(FormOptions);
      end;
    finally
      if GroupSet <> NIL then PGPFreeGroupSet(GroupSet);
    end;
  finally
    PGPFreeFileSpec(GroupFileSpec);
  end;
end;

function KeyServerDialog(Context: pPGPContext; KeySetMain: pPGPKeySet; const KeyData: RawByteString;
			 var KeysFound: Pointer; WinHandle: Integer): PGPError;
var
  tlsContext	 : pPGPtlsContext;
  ThreadStorage	 : pPGPKeyServerThreadStorage;
  KeyServerSpecs : pPGPKeyServerSpec;
  KeyServerCount : PGPUInt32;
  OptionList	 : pPGPOptionList;
begin
  KeysFound := NIL;
  tlsContext := NIL;
  OptionList := NIL;
  Result := PGPNewTLSContext(Context, tlsContext);
  if Result <> 0 then Exit;
  try
    Result := InitKeyServer(Context, ThreadStorage, KeyServerSpecs, KeyServerCount);
    if Result <> 0 then Exit;
    try
      Result := SearchKeyServerOptions(Context, KeyData, FALSE, FALSE, WinHandle, OptionList);
      if Result <> 0 then Exit;
      try
	Result := PGPSearchKeyServerDialog(Context, KeyServerCount, KeyServerSpecs, tlsContext, PGPFalse,
					   KeysFound, OptionList, PGPOLastOption(Context));
	RedrawWindow(WinHandle);
      finally
	PGPFreeOptionList(OptionList);
      end;
    finally
      FreeKeyServer(ThreadStorage, KeyServerSpecs, KeyServerCount);
    end;
  finally
    PGPFreeTLSContext(tlsContext);
  end;
end;

function CollectRandomDataDialog(Context: pPGPContext; NeededEntropyBits: PGPUInt32; WinHandle: Integer): PGPError;
begin
  Result := PGPCollectRandomDataDialog(Context,
				       NeededEntropyBits,
				       PGPOUIParentWindowHandle(Context, WinHandle),
				       PGPOLastOption(Context));
  RedrawWindow(WinHandle);
end;

function PassphraseDialog(Context: pPGPContext; const Title, Prompt: RawByteString;
			  var Passphrase: PAnsiChar; WinHandle: Integer): Integer;
var
  Utf8Title	 : RawByteString;
  Utf8Prompt	 : RawByteString;
  OptionList	 : pPGPOptionList;
begin
  if Utf8Title = '' then Utf8Title := PGP_DLG;
  if SDKVersion >= PGP9XX then begin
    Utf8Title := AnsiToUtf8(Title);
    Utf8Prompt := AnsiToUtf8(Prompt);
  end
  else begin
    Utf8Title := Title;
    Utf8Prompt := Prompt;
  end;
  Passphrase := NIL;
  OptionList := NIL;
  Result := PGPBuildOptionList(Context, OptionList,
    [
      PGPOUIWindowTitle(Context, PAnsiChar(Utf8Title)),
      PGPOUIDialogPrompt(Context, PAnsiChar(Utf8Prompt)),
      PGPOUIOutputPassphrase(Context, Passphrase),
      PGPOUIParentWindowHandle(Context, WinHandle)
    ]);
  if Result <> 0 then Exit;
  try
    Result := PGPPassphraseDialog(Context, OptionList, PGPOLastOption(Context));
    RedrawWindow(WinHandle);
  finally
    PGPFreeOptionList(OptionList);
  end;
end;

function ConvEncPassphraseDialog(Context: pPGPContext; var Passphrase: PAnsiChar;
				 const Prompt: RawByteString; WinHandle: Integer): PGPError;
var
  Utf8Prompt	 : RawByteString;
  OptionList	 : pPGPOptionList;
begin
  if SDKVersion >= PGP9XX then
    Utf8Prompt := AnsiToUtf8(Prompt)
  else Utf8Prompt := Prompt;
  Passphrase := NIL;
  OptionList := NIL;
  Result := PGPBuildOptionList(Context, OptionList,
    [
      PGPOUIWindowTitle(Context, PGP_DLG),
      PGPOUIDialogPrompt(Context, PAnsiChar(Utf8Prompt)),
      PGPOUIOutputPassphrase(Context, Passphrase),
      PGPOUIParentWindowHandle(Context, WinHandle)
    ]);
  if Result <> 0 then Exit;
  try
    Result := PGPConventionalEncryptionPassphraseDialog(Context, OptionList, PGPOLastOption(Context));
    RedrawWindow(WinHandle);
  finally
    PGPFreeOptionList(OptionList);
  end;
end;

function ConvDecPassphraseDialog(Context: pPGPContext; var Passphrase: PAnsiChar;
				 const Prompt: RawByteString; WinHandle: Integer): PGPError;
var
  Utf8Prompt	 : RawByteString;
  OptionList	 : pPGPOptionList;
begin
  if SDKVersion >= PGP9XX then
    Utf8Prompt := AnsiToUtf8(Prompt)
  else Utf8Prompt := Prompt;
  Passphrase := NIL;
  OptionList := NIL;
  Result := PGPBuildOptionList(Context, OptionList,
    [
      PGPOUIWindowTitle(Context, PGP_DLG),
      PGPOUIDialogPrompt(Context, PAnsiChar(Utf8Prompt)),
      PGPOUIOutputPassphrase(Context, Passphrase),
      PGPOUIParentWindowHandle(Context, WinHandle)
    ]);
  if Result <> 0 then Exit;
  try
    Result := PGPConventionalDecryptionPassphraseDialog(Context, OptionList, PGPOLastOption(Context));
    RedrawWindow(WinHandle);
  finally
    PGPFreeOptionList(OptionList);
  end;
end;

function KeyPassphraseDialog(Context: pPGPContext; KeySet: pPGPKeySet; var Passphrase: PAnsiChar;
			     const Prompt: RawByteString; WinHandle: Integer): PGPError;
var
  Utf8Prompt	 : RawByteString;
  Key		 : pPGPKey;
  OptionList	 : pPGPOptionList;
begin
  if SDKVersion >= PGP9XX then
    Utf8Prompt := AnsiToUtf8(Prompt)
  else Utf8Prompt := Prompt;
  Passphrase := NIL;
  OptionList := NIL;
  Result := GetKeyFromKeySet(KeySet, Key);
  if Result <> 0 then Exit;
  Result := PGPBuildOptionList(Context, OptionList,
    [
      PGPOUIWindowTitle(Context, PGP_DLG),
      PGPOUIDialogPrompt(Context, PAnsiChar(Utf8Prompt)),
      PGPOUIOutputPassphrase(Context, Passphrase),
      PGPOUIParentWindowHandle(Context, WinHandle)
    ]);
  if Result <> 0 then Exit;
  try
    Result := PGPKeyPassphraseDialog(Context, Key, OptionList, PGPOLastOption(Context));
    RedrawWindow(WinHandle);
  finally
    PGPFreeOptionList(OptionList);
  end;
end;

function SigningPassphraseDialog(Context: pPGPContext; KeySetMain: pPGPKeySet;
				 var SigningKey: pPGPKey; var SignKeyPass: PAnsiChar;
				 const DefaultKeyID: RawByteString; FindMatchingKey: Longbool; SignOption: TSignOption;
				 var Armor, MIME, RipeMD: PGPUInt32; const Prompt: RawByteString; WinHandle: Integer): PGPError;
var
  Utf8Prompt	 : RawByteString;
  DefaultKey	 : pPGPKey;
  OptionList	 : pPGPOptionList;
  DialogOptions	 : pPGPOptionList;
begin
  if SDKVersion >= PGP9XX then
    Utf8Prompt := AnsiToUtf8(Prompt)
  else Utf8Prompt := Prompt;
  SigningKey := NIL;
  SignKeyPass := NIL;
  OptionList := NIL;
  if (DefaultKeyID = '') or (GetKeyByHexID(KeySetMain, PAnsiChar(DefaultKeyID), DefaultKey) <> 0) then DefaultKey := NIL;
  Result := PGPBuildOptionList(Context, OptionList,
    [
      PGPOUIWindowTitle(Context, PGP_DLG),
      PGPOUIDialogPrompt(Context, PAnsiChar(Utf8Prompt)),
      PGPOUIDefaultKey(Context, DefaultKey),
      PGPOUIFindMatchingKey(Context, PGPBoolean(FindMatchingKey = TRUE)),
      PGPOUIVerifyPassphrase(Context, PGPTrue),
      PGPOUIOutputPassphrase(Context, SignKeyPass),
      PGPOUIParentWindowHandle(Context, WinHandle)
    ]);
  if Result <> 0 then Exit;
  if SignOption <> soNo then begin
    DialogOptions := NIL;
    try
      case SignOption of
	soEnc: begin
	  Result := PGPBuildOptionList(Context, DialogOptions,
	    [
	      PGPOUICheckbox(Context, 803, '&RipeMD', NIL, RipeMD, RipeMD, PGPOLastOption(Context))
	    ]);
	end;
	soClear: begin
	  Result := PGPBuildOptionList(Context, DialogOptions,
	    [
	      PGPOUICheckbox(Context, 802, '&MIME', NIL, MIME, MIME, PGPOLastOption(Context)),
	      PGPOUICheckbox(Context, 803, '&RipeMD (RSA)', NIL, RipeMD, RipeMD, PGPOLastOption(Context))
	    ]);
	end;
	soDetached: begin
	  Result := PGPBuildOptionList(Context, DialogOptions,
	    [
	      PGPOUICheckbox(Context, 801, '&Armor', NIL, Armor, Armor, PGPOLastOption(Context)),
	      PGPOUICheckbox(Context, 803, '&RipeMD', NIL, RipeMD, RipeMD, PGPOLastOption(Context))
	    ]);
	end;
      end;
      if Result <> 0 then Exit;
      Result := PGPAppendOptionList(OptionList, [PGPOUIDialogOptions(Context, DialogOptions, PGPOLastOption(Context))]);
    finally
      PGPFreeOptionList(DialogOptions);
    end;
    if Result <> 0 then Exit;
  end;
  try
    if SDKVersion >= PGP7XX then begin
      Result := PGPSigningPassphraseDialog(Context, PGPPeekKeySetKeyDB(KeySetMain), SigningKey,
					   OptionList, PGPOLastOption(Context));
    end
    else Result := PGPSigningPassphraseDialog(Context, KeySetMain, SigningKey, OptionList, PGPOLastOption(Context));
    RedrawWindow(WinHandle);
  finally
    PGPFreeOptionList(OptionList);
  end;
end;

function DecryptionPassphraseDialog(Context: pPGPContext; RecipientKeys: pPGPKeySet;
				    KeyIDCount: PGPUInt32; const KeyIDList: pPGPKeyID7;
				    var DecryptionKey: pPGPKey; var DecryptKeyPass: PAnsiChar;
				    FindMatchingKey: Longbool; const Prompt: RawByteString; WinHandle: Integer): PGPError;
var
  Utf8Prompt	 : RawByteString;
  OptionList	 : pPGPOptionList;
begin
  if SDKVersion >= PGP9XX then
    Utf8Prompt := AnsiToUtf8(Prompt)
  else Utf8Prompt := Prompt;
  DecryptionKey := NIL;
  DecryptKeyPass := NIL;
  OptionList := NIL;
  Result := PGPBuildOptionList(Context, OptionList,
    [
      PGPOUIWindowTitle(Context, PGP_DLG),
      PGPOUIDialogPrompt(Context, PAnsiChar(Utf8Prompt)),
      PGPOUIFindMatchingKey(Context, PGPBoolean(FindMatchingKey = TRUE)),
      PGPOUIVerifyPassphrase(Context, PGPTrue),
      PGPOUIOutputPassphrase(Context, DecryptKeyPass),
      PGPOUIParentWindowHandle(Context, WinHandle)
    ]);
  if Result <> 0 then Exit;
  try
    Result := PGPDecryptionPassphraseDialog(Context, RecipientKeys, KeyIDCount, KeyIDList,
					    DecryptionKey, OptionList, PGPOLastOption(Context));
    RedrawWindow(WinHandle);
  finally
    PGPFreeOptionList(OptionList);
  end;
end;

function ConfirmationPassphraseDialog(Context: pPGPContext; var Passphrase: PAnsiChar;
				      MinPassLength, MinPassQuality: Integer;
				      ShowPassphraseQuality: Longbool;
				      const Prompt: RawByteString; WinHandle: Integer): PGPError;
var
  Utf8Prompt	 : RawByteString;
  OptionList	 : pPGPOptionList;
begin
  if SDKVersion >= PGP9XX then
    Utf8Prompt := AnsiToUtf8(Prompt)
  else Utf8Prompt := Prompt;
  Passphrase := NIL;
  OptionList := NIL;
  Result := PGPBuildOptionList(Context, OptionList,
    [
      PGPOUIWindowTitle(Context, PGP_DLG),
      PGPOUIDialogPrompt(Context, PAnsiChar(Utf8Prompt)),
      PGPOUIMinimumPassphraseLength(Context, MinPassLength),
      PGPOUIMinimumPassphraseQuality(Context, MinPassQuality),
      PGPOUIShowPassphraseQuality(Context, PGPBoolean(ShowPassphraseQuality = TRUE)),
      PGPOUIOutputPassphrase(Context, Passphrase),
      PGPOUIParentWindowHandle(Context, WinHandle)
    ]);
  if Result <> 0 then Exit;
  try
    Result := PGPConfirmationPassphraseDialog(Context, OptionList, PGPOLastOption(Context));
    RedrawWindow(WinHandle);
  finally
    PGPFreeOptionList(OptionList);
  end;
end;


end.

