{------------------------------------------------------------------------------}
{                                                                              }
{                                 This code is                                 }
{              Copyright (C) 2010-2010 by Michael in der Wiesche               }
{                                                                              }
{------------------------------------------------------------------------------}
{$J+,Z4}
unit PGPStringList;

interface

uses
  Windows, SysUtils, Classes, pgpBase;

{$IFNDEF UNICODE}
type
  TRawByteStrings = class(TStringList)
  protected
    function GetString(Index: Integer): RawByteString; virtual;
    procedure PutString(Index: Integer; const S: RawByteString); virtual;
  end;

{$ELSE}
type
  TRawByteStrings = Class;

  PStringItem = ^TStringItem;
  TStringItem = Record
    FString: RawByteString;
    FObject: TObject;
  end;

  PStringItemList = ^TStringItemList;
  TStringItemList = Array[0..MaxListSize] of TStringItem;
  TRawByteStringsSortCompare = function(List: TRawByteStrings; Index1, Index2: Integer): Integer;

  TRawByteStrings = class(TStrings)
  private
    FCount: Integer;
    FCapacity: Integer;
    FSorted: Boolean;
    FCaseSensitive: Boolean;
    FDuplicates: TDuplicates;
    FList: PStringItemList;
    FOwnsObjects: Boolean;
    FOnChange: TNotifyEvent;
    FOnChanging: TNotifyEvent;
    procedure Grow;
    procedure QuickSort(L, R: Integer; SCompare: TRawByteStringsSortCompare);
    procedure SetCaseSensitive(const Value: Boolean);
    procedure SetSorted(Value: Boolean);
    function GetCommaText: RawByteString;
    procedure SetCommaText(const Value: RawByteString);
    function GetDelimitedText: RawByteString;
    procedure SetDelimitedText(const Value: RawByteString);
    function GetName(Index: Integer): RawByteString;
    function GetValue(const Name: RawByteString): RawByteString;
    procedure SetValue(const Name, Value: RawByteString);
    function GetValueFromIndex(Index: Integer): RawByteString;
    procedure SetValueFromIndex(Index: Integer; const Value: RawByteString);
  protected
    function ExtractName(const S: RawByteString): RawByteString;
    function Get(Index: Integer): String; override;
    procedure Put(Index: Integer; const S: String); override;
    function GetCapacity: Integer; override;
    procedure SetCapacity(NewCapacity: Integer); override;
    function GetCount: Integer; override;
    function GetObject(Index: Integer): TObject; override;
    procedure PutObject(Index: Integer; AObject: TObject); override;
    function GetString(Index: Integer): RawByteString; virtual;
    procedure PutString(Index: Integer; const S: RawByteString); virtual;
    function GetTextStr: RawByteString; reintroduce; virtual;
    procedure SetTextStr(const Value: RawByteString); reintroduce; virtual;
    function CompareStrings(const S1, S2: RawByteString): Integer; reintroduce; virtual;
    procedure ExchangeItems(Index1, Index2: Integer); virtual;
    procedure InsertItem(Index: Integer; const S: RawByteString; AObject: TObject); virtual;
  public
    constructor Create; overload;
    constructor Create(OwnsObjects: Boolean); overload;
    destructor Destroy; override;
    function GetText: PAnsiChar; reintroduce; virtual;
    procedure SetText(Text: PAnsiChar); reintroduce; virtual;
    function Add(const S: RawByteString): Integer; reintroduce; virtual;
    function AddObject(const S: RawByteString; AObject: TObject): Integer; reintroduce; virtual;
    procedure Append(const S: RawByteString);
    procedure Changed; virtual;
    procedure Changing; virtual;
    procedure Clear; override;
    procedure CustomSort(Compare: TRawByteStringsSortCompare); virtual;
    procedure Delete(Index: Integer); override;
    procedure Exchange(Index1, Index2: Integer); override;
    function Find(const S: RawByteString; var Index: Integer): Boolean; virtual;
    function IndexOf(const S: RawByteString): Integer; reintroduce; virtual;
    function IndexOfName(const Name: RawByteString): Integer; reintroduce; virtual;
    procedure Insert(Index: Integer; const S: RawByteString); reintroduce; virtual;
    procedure InsertObject(Index: Integer; const S: RawByteString; AObject: TObject); reintroduce; virtual;
    procedure Move(CurIndex, NewIndex: Integer); override;
    procedure Sort; virtual;
    procedure LoadFromStream(Stream: TStream; Encoding: TEncoding); override;
    procedure SaveToStream(Stream: TStream; Encoding: TEncoding); override;
    property CaseSensitive: Boolean read FCaseSensitive write SetCaseSensitive;
    property DelimitedText: RawByteString read GetDelimitedText write SetDelimitedText;
    property Duplicates: TDuplicates read FDuplicates write FDuplicates;
    property OwnsObjects: Boolean read FOwnsObjects write FOwnsObjects;
    property Sorted: Boolean read FSorted write SetSorted;
    property Text: RawByteString read GetTextStr write SetTextStr;
    property CommaText: RawByteString read GetCommaText write SetCommaText;
    property Strings[Index: Integer]: RawByteString read GetString write PutString; default;
    property Names[Index: Integer]: RawByteString read GetName;
    property Values[const Name: RawByteString]: RawByteString read GetValue write SetValue;
    property ValueFromIndex[Index: Integer]: RawByteString read GetValueFromIndex write SetValueFromIndex;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnChanging: TNotifyEvent read FOnChanging write FOnChanging;
  end;
{$ENDIF}

implementation

{$IFNDEF UNICODE}
function TRawByteStrings.GetString(Index: Integer): RawByteString;
begin
  Result := Get(Index);
end;

procedure TRawByteStrings.PutString(Index: Integer; const S: RawByteString);
begin
  Put(Index, S);
end;

{$ELSE}
uses
  AnsiStrings, PGPStrTools, RTLConsts;

const
  CTRL_CHARS: Set of AnsiChar = [#1..#32];

function RawByteStringsCompare(List: TRawByteStrings; Index1, Index2: Integer): Integer;
begin
  Result := List.CompareStrings(List.FList^[Index1].FString, List.FList^[Index2].FString);
end;

procedure TRawByteStrings.Grow;
var
  iDelta: Integer;
begin
  if FCapacity > 64 then
    iDelta := FCapacity div 4
  else if FCapacity > 8 then
    iDelta := 16
  else iDelta := 4;
  SetCapacity(FCapacity + iDelta);
end;

procedure TRawByteStrings.QuickSort(L, R: Integer; SCompare: TRawByteStringsSortCompare);
var
  I, J, M: Integer;
begin
  repeat
    I := L;
    J := R;
    M := (L + R) shr 1;
    repeat
      while SCompare(Self, I, M) < 0 do inc(I);
      while SCompare(Self, J, M) > 0 do dec(J);
      if I <= J then begin
	if I <> J then ExchangeItems(I, J);
	if M = I then
	  M := J
	else if M = J then
	  M := I;
	inc(I);
	dec(J);
      end;
    until I > J;
    if L < J then QuickSort(L, J, SCompare);
    L := I;
  until I >= R;
end;

procedure TRawByteStrings.SetCaseSensitive(const Value: Boolean);
begin
  if Value <> FCaseSensitive then begin
    FCaseSensitive := Value;
    if FSorted then begin
      Sorted := FALSE;
      Sorted := TRUE;
    end;
  end;
end;

procedure TRawByteStrings.SetSorted(Value: Boolean);
begin
  if FSorted <> Value then begin
    if Value then Sort;
    FSorted := Value;
  end;
end;

function TRawByteStrings.GetCommaText: RawByteString;
var
  chrDelimiter: Char;
  chrQuoteChar: Char;
begin
  chrDelimiter := Delimiter;
  chrQuoteChar := QuoteChar;
  Delimiter := CM;
  QuoteChar := QU;
  try
    Result := GetDelimitedText;
  finally
    Delimiter := chrDelimiter;
    QuoteChar := chrQuoteChar;
  end;
end;

procedure TRawByteStrings.SetCommaText(const Value: RawByteString);
begin
  Delimiter := CM;
  QuoteChar := QU;
  SetDelimitedText(Value);
end;

function TRawByteStrings.GetDelimitedText: RawByteString;
var
  iIndex: Integer;
  scsDelimiters: TSysCharSet;
  s: RawByteString;
  p: PAnsiChar;
begin
  if (FCount = 1) and (Get(0) = E) then
    Result := AnsiChar(QuoteChar) + AnsiChar(QuoteChar)
  else begin
    Result := E;
    scsDelimiters := [AnsiChar(NUL), AnsiChar(QuoteChar), AnsiChar(Delimiter)];
    if not StrictDelimiter then scsDelimiters := scsDelimiters + CTRL_CHARS;
    for iIndex := 0 to pred(FCount) do begin
      s := GetString(iIndex);
      p := PAnsiChar(s);
      while not (p^ in scsDelimiters) do inc(p);
      if (p^ <> NUL) then s := AnsiQuotedStr(s, AnsiChar(QuoteChar));
      Result := Result + s + AnsiChar(Delimiter);
    end;
    System.Delete(Result, Length(Result), 1);
  end;
end;

procedure TRawByteStrings.SetDelimitedText(const Value: RawByteString);
var
  p1, p2: PAnsiChar;
  s: RawByteString;
begin
  BeginUpdate;
  try
    Clear;
    p1 := PAnsiChar(Value);
    if not StrictDelimiter then while (p1^ in CTRL_CHARS) do inc(p1);
    while p1^ <> NUL do begin
      if p1^ = AnsiChar(QuoteChar) then
	s := AnsiExtractQuotedStr(p1, AnsiChar(QuoteChar))
      else begin
	p2 := p1;
	while ((not StrictDelimiter and (p1^ > SP)) or (StrictDelimiter and (p1^ <> NUL))) and (p1^ <> AnsiChar(Delimiter)) do inc(p1);
	SetString(s, p2, p1 - p2);
      end;
      Add(s);
      if not StrictDelimiter then while (p1^ in CTRL_CHARS) do inc(p1);
      if p1^ = AnsiChar(Delimiter) then begin
	p2 := p1;
	inc(p2);
	if p2^ = NUL then Add(E);
	repeat
	  inc(p1);
	until not (not StrictDelimiter and (p1^ in CTRL_CHARS));
      end;
    end;
  finally
    EndUpdate;
  end;
end;

function TRawByteStrings.GetName(Index: Integer): RawByteString;
begin
  Result := ExtractName(GetString(Index));
end;

function TRawByteStrings.GetValue(const Name: RawByteString): RawByteString;
var
  iIndex: Integer;
begin
  iIndex := IndexOfName(Name);
  if iIndex >= 0 then
    Result := Copy(GetString(iIndex), Length(Name) + 2, MAXINT)
  else Result := E;
end;

procedure TRawByteStrings.SetValue(const Name, Value: RawByteString);
var
  iIndex: Integer;
begin
  iIndex := IndexOfName(Name);
  if Value <> E then begin
    if iIndex < 0 then iIndex := Add(E);
    PutString(iIndex, Name + AnsiChar(NameValueSeparator) + Value);
  end
  else if iIndex >= 0 then Delete(iIndex);
end;

function TRawByteStrings.GetValueFromIndex(Index: Integer): RawByteString;
var
  iLen: Integer;
begin
  if Index >= 0 then begin
    iLen := Length(Names[Index]);
    if iLen > 0 then begin
      Result := Copy(GetString(Index), iLen + 2, MAXINT);
      Exit;
    end;
  end;
  Result := E;
end;

procedure TRawByteStrings.SetValueFromIndex(Index: Integer; const Value: RawByteString);
begin
  if Value <> E then begin
    if Index < 0 then Index := Add(E);
    PutString(Index, Names[Index] + AnsiChar(NameValueSeparator) + Value);
  end
  else if Index >= 0 then Delete(Index);
end;

function TRawByteStrings.ExtractName(const S: RawByteString): RawByteString;
var
  iPos: Integer;
begin
  Result := S;
  iPos := AnsiPos(AnsiChar(NameValueSeparator), Result);
  if iPos <> 0 then
    SetLength(Result, pred(iPos))
  else SetLength(Result, 0);
end;

function TRawByteStrings.Get(Index: Integer): String;
begin
  if Cardinal(Index) >= Cardinal(FCount) then Error(@SListIndexError, Index);
  Result := String(FList^[Index].FString);
end;

procedure TRawByteStrings.Put(Index: Integer; const S: String);
begin
  if FSorted then Error(@SSortedListError, 0);
  if Cardinal(Index) >= Cardinal(FCount) then Error(@SListIndexError, Index);
  Changing;
  FList^[Index].FString := RawByteString(S);
  Changed;
end;

function TRawByteStrings.GetCapacity: Integer;
begin
  Result := FCapacity;
end;

procedure TRawByteStrings.SetCapacity(NewCapacity: Integer);
begin
  if (NewCapacity < FCount) or (NewCapacity > MaxListSize) then Error(@SListCapacityError, NewCapacity);
  if NewCapacity <> FCapacity then begin
    ReallocMem(FList, NewCapacity * SizeOf(TStringItem));
    FCapacity := NewCapacity;
  end;
end;

function TRawByteStrings.GetCount: Integer;
begin
  Result := FCount;
end;

function TRawByteStrings.GetObject(Index: Integer): TObject;
begin
  if Cardinal(Index) >= Cardinal(FCount) then Error(@SListIndexError, Index);
  Result := FList^[Index].FObject;
end;

procedure TRawByteStrings.PutObject(Index: Integer; AObject: TObject);
begin
  if Cardinal(Index) >= Cardinal(FCount) then Error(@SListIndexError, Index);
  Changing;
  FList^[Index].FObject := AObject;
  Changed;
end;

function TRawByteStrings.GetString(Index: Integer): RawByteString;
begin
  if Cardinal(Index) >= Cardinal(FCount) then Error(@SListIndexError, Index);
  Result := FList^[Index].FString;
end;

procedure TRawByteStrings.PutString(Index: Integer; const S: RawByteString);
begin
  if FSorted then Error(@SSortedListError, 0);
  if Cardinal(Index) >= Cardinal(FCount) then Error(@SListIndexError, Index);
  Changing;
  FList^[Index].FString := S;
  Changed;
end;

function TRawByteStrings.GetTextStr: RawByteString;
var
  sLineBreak, sItem: RawByteString;
  iSize, iIndex, iLen: Integer;
  pResult: PAnsiChar;
begin
  sLineBreak := RawByteString(LineBreak);
  iSize := 0;
  for iIndex := 0 to pred(FCount) do inc(iSize, Length(Get(iIndex)) + Length(sLineBreak));
  SetString(Result, NIL, iSize);
  pResult := Pointer(Result);
  for iIndex := 0 to pred(FCount) do begin
    sItem := GetString(iIndex);
    iLen := Length(sItem);
    if iLen <> 0 then begin
      System.Move(Pointer(sItem)^, pResult^, iLen);
      inc(pResult, iLen);
    end;
    iLen := Length(sLineBreak);
    if iLen <> 0 then begin
      System.Move(Pointer(sLineBreak)^, pResult^, iLen);
      inc(pResult, iLen);
    end;
  end;
end;

procedure TRawByteStrings.SetTextStr(const Value: RawByteString);
var
  pLineBreak, pValue, pStart, pLB: PAnsiChar;
  sItem: RawByteString;
  iLen: Integer;
begin
  BeginUpdate;
  try
    Clear;
    pLineBreak := PAnsiChar(AnsiString(LineBreak));
    pValue := Pointer(Value);
    if pValue <> NIL then
      if CompareStr(LineBreak, sLineBreak) = 0 then begin
	while pValue^ <> NUL do begin
	  pStart := pValue;
	  while not (pValue^ in [NUL, LF, CR]) do inc(pValue);
	  SetString(sItem, pStart, pValue - pStart);
	  Add(sItem);
	  if pValue^ = CR then inc(pValue);
	  if pValue^ = LF then inc(pValue);
	end;
      end
      else begin
	iLen := Length(LineBreak);
	while pValue^ <> NUL do begin
	  pStart := pValue;
	  pLB := AnsiStrPos(pValue, pLineBreak);
	  while (pValue^ <> NUL) and (pValue <> pLB) do inc(pValue);
	  SetString(sItem, pStart, pValue - pStart);
	  Add(sItem);
	  if pValue = pLB then inc(pValue, iLen);
	end;
      end;
  finally
    EndUpdate;
  end;
end;

function TRawByteStrings.CompareStrings(const S1, S2: RawByteString): Integer;
begin
  if FCaseSensitive then
    Result := AnsiCompareStr(S1, S2)
  else Result := AnsiCompareText(S1, S2);
end;

procedure TRawByteStrings.ExchangeItems(Index1, Index2: Integer);
var
  si1, si2: PStringItem;
  p: Pointer;
begin
  si1 := @FList^[Index1];
  si2 := @FList^[Index2];
  p := Pointer(si1^.FString);
  Pointer(si1^.FString) := Pointer(si2^.FString);
  Pointer(si2^.FString) := p;
  p := si1^.FObject;
  si1^.FObject := si2^.FObject;
  si2^.FObject := p;
end;

procedure TRawByteStrings.InsertItem(Index: Integer; const S: RawByteString; AObject: TObject);
begin
  Changing;
  if FCount = FCapacity then Grow;
  if Index < FCount then System.Move(FList^[Index], FList^[Index + 1], (FCount - Index) * SizeOf(TStringItem));
  with FList^[Index] do begin
    Pointer(FString) := NIL;
    FObject := AObject;
    FString := S;
  end;
  inc(FCount);
  Changed;
end;

constructor TRawByteStrings.Create;
begin
  inherited Create;
  StrictDelimiter := TRUE;
end;

constructor TRawByteStrings.Create(OwnsObjects: Boolean);
begin
  inherited Create;
  StrictDelimiter := TRUE;
  FOwnsObjects := OwnsObjects;
end;

destructor TRawByteStrings.Destroy;
var
  iIndex: Integer;
begin
  FOnChanging := NIL;
  FOnChange := NIL;
  if FCount <> 0 then begin
    if FOwnsObjects then for iIndex := 0 to pred(FCount) do GetObject(iIndex).Free;
    Finalize(FList^[0], FCount);
    FCount := 0;
    SetCapacity(0);
  end;
  inherited Destroy;
end;

function TRawByteStrings.GetText: PAnsiChar;
begin
  Result := StrNew(PAnsiChar(GetTextStr));
end;

procedure TRawByteStrings.SetText(Text: PAnsiChar);
begin
  SetTextStr(Text);
end;

function TRawByteStrings.Add(const S: RawByteString): Integer;
begin
  Result := AddObject(S, NIL);
end;

function TRawByteStrings.AddObject(const S: RawByteString; AObject: TObject): Integer;
begin
  if not FSorted then
    Result := FCount
  else if Find(S, Result) then begin
    case FDuplicates of
      dupIgnore: Exit;
      dupError: Error(@SDuplicateString, 0);
    end;
  end;
  InsertItem(Result, S, AObject);
end;

procedure TRawByteStrings.Append(const S: RawByteString);
begin
  Add(S);
end;

procedure TRawByteStrings.Changed;
begin
  if (UpdateCount = 0) and Assigned(FOnChange) then FOnChange(Self);
end;

procedure TRawByteStrings.Changing;
begin
  if (UpdateCount = 0) and Assigned(FOnChanging) then FOnChanging(Self);
end;

procedure TRawByteStrings.Clear;
var
  iIndex: Integer;
begin
  if FCount <> 0 then begin
    Changing;
    if FOwnsObjects then for iIndex := 0 to pred(FCount) do GetObject(iIndex).Free;
    Finalize(FList^[0], FCount);
    FCount := 0;
    SetCapacity(0);
    Changed;
  end;
end;

procedure TRawByteStrings.CustomSort(Compare: TRawByteStringsSortCompare);
begin
  if not FSorted and (FCount > 1) then begin
    Changing;
    QuickSort(0, FCount - 1, Compare);
    Changed;
  end;
end;

procedure TRawByteStrings.Delete(Index: Integer);
begin
  if (Index < 0) or (Index >= FCount) then Error(@SListIndexError, Index);
  Changing;
  if FOwnsObjects then GetObject(Index).Free;
  Finalize(FList^[Index]);
  dec(FCount);
  if Index < FCount then System.Move(FList^[Index + 1], FList^[Index], (FCount - Index) * SizeOf(TStringItem));
  Changed;
end;

procedure TRawByteStrings.Exchange(Index1, Index2: Integer);
begin
  if (Index1 < 0) or (Index1 >= FCount) then Error(@SListIndexError, Index1);
  if (Index2 < 0) or (Index2 >= FCount) then Error(@SListIndexError, Index2);
  Changing;
  ExchangeItems(Index1, Index2);
  Changed;
end;

function TRawByteStrings.Find(const S: RawByteString; var Index: Integer): Boolean;
var
  L, H, M, C: Integer;
begin
  Result := FALSE;
  L := 0;
  H := pred(FCount);
  while L <= H do begin
    M := (L + H) shr 1;
    C := CompareStrings(FList^[M].FString, S);
    if C < 0 then
      L := succ(M)
    else begin
      H := pred(M);
      if C = 0 then begin
        Result := TRUE;
	if FDuplicates <> dupAccept then L := M;
      end;
    end;
  end;
  Index := L;
end;

function TRawByteStrings.IndexOf(const S: RawByteString): Integer;
begin
  if FSorted then begin
    if Find(S, Result) then Exit;
  end
  else for Result := 0 to pred(FCount) do if CompareStrings(GetString(Result), S) = 0 then Exit;
  Result := -1;
end;

function TRawByteStrings.IndexOfName(const Name: RawByteString): Integer;
var
  sItem: RawByteString;
  iPos: Integer;
begin
  for Result := 0 to pred(FCount) do begin
    sItem := GetString(Result);
    iPos := AnsiPos(AnsiChar(NameValueSeparator), sItem);
    if (iPos <> 0) and (CompareStrings(Copy(sItem, 1, pred(iPos)), Name) = 0) then Exit;
  end;
  Result := -1;
end;

procedure TRawByteStrings.Insert(Index: Integer; const S: RawByteString);
begin
  InsertObject(Index, S, NIL);
end;

procedure TRawByteStrings.InsertObject(Index: Integer; const S: RawByteString; AObject: TObject);
begin
  if FSorted then Error(@SSortedListError, 0);
  if (Index < 0) or (Index > FCount) then Error(@SListIndexError, Index);
  InsertItem(Index, S, AObject);
end;

procedure TRawByteStrings.Move(CurIndex, NewIndex: Integer);
var
  sTemp: RawByteString;
  objTemp: TObject;
begin
  if CurIndex <> NewIndex then begin
    BeginUpdate;
    try
      sTemp := GetString(CurIndex);
      objTemp := GetObject(CurIndex);
      Delete(CurIndex);
      InsertObject(NewIndex, sTemp, objTemp);
    finally
      EndUpdate;
    end;
  end;
end;

procedure TRawByteStrings.Sort;
begin
  CustomSort(RawByteStringsCompare);
end;

procedure TRawByteStrings.LoadFromStream(Stream: TStream; Encoding: TEncoding);
var
  iSize: Integer;
  sBuffer: RawByteString;
begin
  BeginUpdate;
  try
    with Stream do iSize := Size - Position;
    if iSize > 0 then begin
      SetLength(sBuffer, iSize);
      Stream.ReadBuffer(Pointer(sBuffer)^, iSize);
      SetTextStr(sBuffer);
    end
    else Clear;
  finally
    EndUpdate;
  end;
end;

procedure TRawByteStrings.SaveToStream(Stream: TStream; Encoding: TEncoding);
var
  sBuffer: RawByteString;
begin
  sBuffer := GetTextStr;
  if sBuffer <> E then
    Stream.WriteBuffer(Pointer(sBuffer)^, Length(sBuffer))
  else Stream.Size := 0;
end;
{$ENDIF}

end.
