{$A+,B-,C-,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z4}
{$MINSTACKSIZE $00004000}
{$MAXSTACKSIZE $00100000}
{$IMAGEBASE $00400000}
{$APPTYPE GUI}

// huge strings enabled!

unit TestUnit;

{
  This sample provides code for implementing some of the PGP functions using
  the TPGP-components provided with PGP2comp. It's been written carefully
  to avoid any problems that might compromise running systems, but
  it doesn't contain a sufficient error handling for every possible situation
  as it only should serve for demonstration purposes. So be careful when using
  parts of it in your own applications.

  Michael in der Wiesche, August 4th, 2007
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ComCtrls, ExtCtrls,
  pgpBase, PGP2Comp, PGPKeyGenerate, PGPEncode, PGPDecode, KeyPropTypes, X509Comp, X509Types;

type
  TPGPDemo = class(TForm)
    pnFrame: TPanel;
    reResults: TRichEdit;
    OpenDialog: TOpenDialog;
    SaveDialog: TSaveDialog;
    PGPPreferences: TPGPPreferences;
    PGPKeyServer: TPGPKeyServer;
    PGPGetKeyProps: TPGPGetKeyProps;
    PGPSetKeyProps: TPGPSetKeyProps;
    PGPKeysGenerate: TPGPKeysGenerate;
    PGPKeyImport: TPGPKeyImport;
    PGPKeyExport: TPGPKeyExport;
    PGPEncode: TPGPEncode;
    PGPDecode: TPGPDecode;
    cbxUserIDs: TComboBox;
    cbxCertUserIDs: TComboBox;
    btnKeyGenerate: TButton;
    btnEncode: TButton;
    btnDecode: TButton;
    btnExportKey: TButton;
    btnImportKey: TButton;
    btnFingerprint: TButton;
    btnExportCertificate: TButton;
    btnImportCertificate: TButton;
    btnCertProps: TButton;
    X509RequestHandler: TX509RequestHandler;
    X509GetCertProps: TX509GetCertProps;
    X509SetCertProps: TX509SetCertProps;
    X509CertGenerate: TX509CertGenerate;
    X509CertImport: TX509CertImport;
    X509CertExport: TX509CertExport;
    procedure FormCreate(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure cbxUserIDsChange(Sender: TObject);
    procedure cbxCertUserIDsChange(Sender: TObject);
    procedure PGPFailure(ErrorCode: Longint; const ErrorMsg: String);
    procedure ShowCodingProgress(BytesProcessed, BytesTotal: Longint);
    procedure btnKeyGenerateClick(Sender: TObject);
    procedure PGPKeysGenerateShowState(State: Char; var Cancel: Longbool);
    procedure PGPKeysGenerateKeyGeneration(const NewHexID: AnsiString; const MasterKeyProps: TKeyPropsRec; Aborted: Longbool);
    procedure btnEncodeClick(Sender: TObject);
    procedure PGPEncodeGetInputFileName(var SuggestedName: AnsiString);
    procedure PGPEncodeGetOutputFileName(var SuggestedName: AnsiString);
    procedure PGPEncodeEncoded(const BufferOut, FileOut: AnsiString);
    procedure btnDecodeClick(Sender: TObject);
    procedure PGPDecodeGetInputFileName(var SuggestedName: AnsiString);
    procedure PGPDecodeGetOutputFileName(var SuggestedName: AnsiString);
    procedure PGPDecodeAnalysis(SectionNumber: Longint; SectionType: TAnalysis; var SkipSection: Longbool);
    procedure PGPDecodeDecoded(const BufferOut, FileOut: AnsiString; const SigPropsRec: TSigPropsRec; const KeyPropsList: TKeyPropsList);
    procedure btnExportKeyClick(Sender: TObject);
    procedure PGPKeyExportKeyExported(const KeyPropsList: TKeyPropsList; const KeyData, FileOut: AnsiString);
    procedure btnImportKeyClick(Sender: TObject);
    procedure PGPKeyImportGetFileIn(var FileIn: AnsiString);
    procedure PGPKeyImportKeyImported(const KeyPropsList: TKeyPropsList; KeysImported: Longint);
    procedure btnFingerprintClick(Sender: TObject);
    procedure PGPGetKeyPropsGetKeyProps(const RingPropsList: TKeyPropsList);
    procedure btnExportCertificateClick(Sender: TObject);
    procedure X509CertExportCertExported(const CertPropsList: TX509CertPropsList; CertChain, ToFile: Longbool);
    procedure btnImportCertificateClick(Sender: TObject);
    procedure X509CertImportCertImported(const CertPropsList: TX509CertPropsList; CertsImported: Longint);
    procedure btnCertPropsClick(Sender: TObject);
  private
    sDots: String;
    X509Certificate: Longbool;
    procedure JumpStart;
    procedure UpdateComboBoxes;
    procedure AddLines(const Text: AnsiString);
    procedure DisplayText(const Text: AnsiString);
    procedure SetHint(Ctrl: TControl; const Text: AnsiString);
    procedure AppOnMessage(var Msg: TMsg; var Handled: Boolean);
  public
  end;

var
  PGPDemo: TPGPDemo;

const
  E = '';
  SP = ' ';
  CR = #13;
  LF = #10;
  DOT = '.';
  CRLF = #13#10;
  LFLF = #10#10;
  X509_VERSION_ERROR = 'PGP 6.5.8 cannot process x.509 certificates';

implementation

{$R *.DFM}

// update ID lists on keyring changes

procedure TPGPDemo.JumpStart;
begin
  Activate;
  with reResults do begin
    SetFocus;
    SelStart := 0;
    SelLength := 0;
    Perform(EM_SCROLLCARET, 0, 0);
  end;
end;

procedure TPGPDemo.UpdateComboBoxes;
var
  PreviousCursor: TCursor;
begin
  Application.ProcessMessages;
  PreviousCursor := Screen.Cursor;
  try
    Screen.Cursor := crHourGlass;
    PGPGetKeyProps.Update;
    with cbxUserIDs do if Items.Count <> 0 then begin
      ItemIndex := 0;
      SetHint(cbxUserIDs, PGPGetKeyProps.RingPropsList.KeyProps[ItemIndex].kHexID);
    end;
    if SDKVersion < PGP7XX then Exit;
    X509GetCertProps.Update;
    with cbxCertUserIDs do if Items.Count <> 0 then begin
      ItemIndex := 0;
      SetHint(cbxCertUserIDs, 'Issuer: ' + X509GetCertProps.CertPropsList.X509CertProps[ItemIndex].X509IssuerLongName);
    end;
  finally
    Screen.Cursor := PreviousCursor;
  end;
end;

procedure TPGPDemo.AddLines(const Text: AnsiString);
begin
  reResults.Lines.Add(String(Text));
end;

procedure TPGPDemo.DisplayText(const Text: AnsiString);
begin
  reResults.Lines.Text := String(Text);
end;

procedure TPGPDemo.SetHint(Ctrl: TControl; const Text: AnsiString);
begin
  Ctrl.Hint := String(Text);
end;

procedure TPGPDemo.AppOnMessage(var Msg: TMsg; var Handled: Boolean);
begin
  with Msg do begin
    if (Message = WM_PGP_ReloadPrefs) or (Message = WM_PGP_ReloadKeyring) then
      UpdateComboBoxes
    else inherited;
  end;
end;

// initialize

procedure TPGPDemo.FormCreate(Sender: TObject);
begin
  Application.OnMessage := AppOnMessage;
  Application.HintHidePause := 5000;
end;

// load ID lists

procedure TPGPDemo.FormActivate(Sender: TObject);
begin
  UpdateComboBoxes;
end;

// show hex ID as hint

procedure TPGPDemo.cbxUserIDsChange(Sender: TObject);
begin
  with cbxUserIDs do if ItemIndex <> -1 then
    SetHint(cbxUserIDs, PGPGetKeyProps.RingPropsList.KeyProps[ItemIndex].kHexID)
  else Hint := E;
end;

// show issuer as hint

procedure TPGPDemo.cbxCertUserIDsChange(Sender: TObject);
begin
  if SDKVersion < PGP7XX then Exit;
  with cbxCertUserIDs do if ItemIndex <> -1 then
    SetHint(cbxCertUserIDs, 'Issuer: ' + X509GetCertProps.CertPropsList.X509CertProps[ItemIndex].X509IssuerLongName)
  else Hint := E;
end;

// common FailProc for PGPErrors

procedure TPGPDemo.PGPFailure(ErrorCode: Longint; const ErrorMsg: String);
begin
  MessageDlg(ErrorMsg + ' (' + IntToStr(ErrorCode) + ')', mtError, [mbOK], 0);
end;

// show coding progress

procedure TPGPDemo.ShowCodingProgress(BytesProcessed, BytesTotal: Longint);
begin
  if BytesTotal > 0 then DisplayText('Progress: ' + AnsiString(IntToStr(Round(BytesProcessed / BytesTotal * 100))) + '%');
end;

// generate DH/DSS key

procedure TPGPDemo.btnKeyGenerateClick(Sender: TObject);
begin
  sDots := DOT;
  with reResults do begin
    Clear;
    Refresh;
    Alignment := taLeftJustify;
  end;

  DisplayText('Generating key:' + CRLF);
  PGPKeysGenerate.DHDSSKeyGenerate;
  Invalidate;
end;

// show progress of key generation

procedure TPGPDemo.PGPKeysGenerateShowState(State: Char; var Cancel: Longbool);
begin
  with reResults.Lines do reResults.Lines[1] := sDots;
  sDots := sDots + DOT;
end;

// show results of key generation

procedure TPGPDemo.PGPKeysGenerateKeyGeneration(const NewHexID: AnsiString; const MasterKeyProps: TKeyPropsRec; Aborted: Longbool);
begin
  reResults.Lines.BeginUpdate;
  if not Aborted then with MasterKeyProps do begin
    case kKeyAlgorithm of
      KeyAlgorithm_RSA: DisplayText('New RSA key: ' + kUserID + ' (' + kHexID + ')');
      KeyAlgorithm_DHDSS: DisplayText('New DH/DSS key: ' + kUserID + ' (' + kHexID + ')');
    end;
  end
  else DisplayText('Key generation aborted');
  JumpStart;
  reResults.Lines.EndUpdate;
end;

// start encoding process

procedure TPGPDemo.btnEncodeClick(Sender: TObject);
begin
  with reResults do begin
    Clear;
    Refresh;
    Alignment := taLeftJustify;
  end;
  PGPEncode.PassCacheSeconds := PGPPreferences.PassCacheSeconds;
  PGPEncode.PassCacheShare := PGPPreferences.PassCacheShare;
  PGPEncode.SignKeyID := E;
  PGPEncode.OutputFileName := E;
  PGPEncode.EncryptKeyIDs.Clear;
  PGPEncode.KeyEncryptFile(E, TRUE, FALSE);
end;

// select DataFile for encryption

procedure TPGPDemo.PGPEncodeGetInputFileName(var SuggestedName: AnsiString);
begin
  with OpenDialog do begin
    FileName := String(SuggestedName);
    Title := 'Open file to encode:';
    InitialDir := ExtractFilePath(FileName);
    if InitialDir = E then InitialDir := ExtractFilePath(Application.ExeName);
    Filter := 'All Files (*.*)|*.*';
    if Execute then
      SuggestedName := RawByteString(FileName)
    else SuggestedName := E;
  end;
  Refresh;
end;

// select OutputFile for encrypted/signed data

procedure TPGPDemo.PGPEncodeGetOutputFileName(var SuggestedName: AnsiString);
begin
  with SaveDialog do begin
    FileName := String(SuggestedName);
    Title := 'Select file for encrypted data:';
    InitialDir := ExtractFilePath(FileName);
    if InitialDir = E then InitialDir := ExtractFilePath(Application.ExeName);
    Filter := 'PGP Files (*.pgp)|*.pgp|All Files (*.*)|*.*';
    if Execute then
      SuggestedName := RawByteString(FileName)
    else SuggestedName := E;
  end;
  Refresh;
end;

// show name of file containing encoded data

procedure TPGPDemo.PGPEncodeEncoded(const BufferOut, FileOut: AnsiString);
begin
  reResults.Lines.BeginUpdate;
  with reResults, Lines do begin
    SetFocus;
    DisplayText('Encoded data in file ' + LFLF + '"' + FileOut + '"');
  end;
  PGPEncode.EncryptKeyIDs.Clear;
  JumpStart;
  reResults.Lines.EndUpdate;
end;

// start decoding process

procedure TPGPDemo.btnDecodeClick(Sender: TObject);
begin
  X509Certificate := FALSE;
  with reResults do begin
    Clear;
    Refresh;
    Alignment := taLeftJustify;
  end;
  PGPDecode.PassCacheSeconds := PGPPreferences.PassCacheSeconds;
  PGPDecode.PassCacheShare := PGPPreferences.PassCacheShare;
  PGPDecode.OutputFileName := E;
  PGPDecode.DecodeFile(E);
  Invalidate;
end;

// select DataFile for decoding

procedure TPGPDemo.PGPDecodeGetInputFileName(var SuggestedName: AnsiString);
begin
  with OpenDialog do begin
    FileName := String(SuggestedName);
    Title := 'Open file to decode:';
    FileName := String(SuggestedName);
    InitialDir := ExtractFilePath(FileName);
    if InitialDir = E then InitialDir := ExtractFilePath(Application.ExeName);
    Filter := 'PGP/X509 Files (*.asc, *.pgp, *.sig, *.cer, *.p7b, *.pem)|*.asc;*.pgp;*.sig;*.cer;*.p7b;*.pem|' +
	      'All Files (*.*)|*.*';
    if Execute then
      SuggestedName := RawByteString(FileName)
    else SuggestedName := E;
  end;
  Refresh;
end;

// select OutputFile for decrypted/verified data

procedure TPGPDemo.PGPDecodeGetOutputFileName(var SuggestedName: AnsiString);
begin
  with SaveDialog do begin
    FileName := String(SuggestedName);
    Title := 'Select file for decoded data:';
    FileName := String(SuggestedName);
    InitialDir := ExtractFilePath(FileName);
    if InitialDir = E then InitialDir := ExtractFilePath(Application.ExeName);
    Filter := 'All Files (*.*)|*.*';
    if Execute then
      SuggestedName := RawByteString(FileName)
    else SuggestedName := E;
  end;
  Refresh;
end;

// show decoded data and signature status on success

procedure TPGPDemo.PGPDecodeDecoded(const BufferOut, FileOut: AnsiString; const SigPropsRec: TSigPropsRec; const KeyPropsList: TKeyPropsList);
begin
  reResults.Lines.BeginUpdate;
  if (BufferOut <> E) or (FileOut <> E) then with reResults, Lines do begin
    SetFocus;
    if BufferOut <> E then begin
      if X509Certificate then
	DisplayText('			< ... X509 certificate(s) detected ... >' + CRLF + BufferOut)
      else DisplayText(BufferOut);
    end
    else DisplayText('Decoded data in file ' + LFLF + '"' + FileOut + '"');
  end;
  JumpStart;
  reResults.Lines.EndUpdate;
end;

// show analysis result if X509 data

procedure TPGPDemo.PGPDecodeAnalysis(SectionNumber: Longint; SectionType: TAnalysis; var SkipSection: Longbool);
begin
  X509Certificate := (SectionType in [X509_CERTIFICATE..X509_SECRETKEY]);
end;

// start extracting keys

procedure TPGPDemo.btnExportKeyClick(Sender: TObject);
begin
  with reResults do begin
    Clear;
    Refresh;
    Alignment := taLeftJustify;
  end;
  PGPKeyExport.FileOut := E;
  PGPKeyExport.DoKeyExport;
  Invalidate;
end;

// show number and data of exported keys on success

procedure TPGPDemo.PGPKeyExportKeyExported(const KeyPropsList: TKeyPropsList; const KeyData, FileOut: AnsiString);
begin
  reResults.Lines.BeginUpdate;
  with reResults do begin
    SetFocus;
    DisplayText(AnsiString(IntToStr(KeyPropsList.Count)) + ' exported key(s):' + LFLF + (KeyData));
  end;
  JumpStart;
  reResults.Lines.EndUpdate;
end;

// start importing keys

procedure TPGPDemo.btnImportKeyClick(Sender: TObject);
begin
  with reResults do begin
    Clear;
    Refresh;
    Alignment := taLeftJustify;
  end;
  PGPKeyImport.FileIn := E;
  PGPKeyImport.KeyData := E;
  PGPKeyImport.DoKeyImportFile;
  Invalidate;
end;


// select file containing KeyData

procedure TPGPDemo.PGPKeyImportGetFileIn(var FileIn: AnsiString);
begin
  with OpenDialog do begin
    Title := 'Open a key file:';
    InitialDir := ExtractFilePath(Application.ExeName);
    Filter := 'Key files (*.asc, *.pgp)|*.asc;*.pgp|All Files (*.*)|*.*';
    if Execute then
      FileIn := RawByteString(FileName)
    else FileIn := E;
  end;
  Refresh;
end;

// show UserIDs of imported key(s)

procedure TPGPDemo.PGPKeyImportKeyImported(const KeyPropsList: TKeyPropsList; KeysImported: Longint);
var
  KeyCount: Integer;
begin
  reResults.Lines.BeginUpdate;
  with reResults do begin
    SetFocus;
    DisplayText('The following key(s) have been successfully imported:' + LFLF);
    for KeyCount := 0 to pred(KeyPropsList.Count) do begin
      AddLines(KeyPropsList.KeyProps[KeyCount].kUserID + ' (' + KeyPropsList.KeyProps[KeyCount].kHexID + ')');
    end;
  end;
  JumpStart;
  reResults.Lines.EndUpdate;
end;

// get fingerprint of selected key

procedure TPGPDemo.btnFingerprintClick(Sender: TObject);
begin
  reResults.Clear;
  reResults.Refresh;
  if cbxUserIDs.ItemIndex <> -1 then begin
    Application.ProcessMessages;
    with PGPGetKeyProps do begin
      KeyID := RingPropsList.KeyProps[cbxUserIDs.ItemIndex].kHexID;
      KeyProps := KeyProps + KeyProps_IDFlags + [KeyProp_Fingerprint];
      DoGetKeyProps;
    end;
  end;
end;

// show selected KeyProps on success

procedure TPGPDemo.PGPGetKeyPropsGetKeyProps(const RingPropsList: TKeyPropsList);
begin
  reResults.Lines.BeginUpdate;
  if RingPropsList.Count <> 0 then begin
    with reResults do begin
      Clear;
      Refresh;
      Alignment := taCenter;
    end;
    with RingPropsList, KeyProps[0] do begin
      if ValidProps = KeyProps_IDFlags + [KeyProp_Fingerprint] then DisplayText(kUserID + ' (' + kHexID + ')' + LFLF + kFingerprint);
    end;
  end;
  JumpStart;
  reResults.Lines.EndUpdate;
end;

// export X509 Certificate

procedure TPGPDemo.btnExportCertificateClick(Sender: TObject);
begin
  if SDKVersion < PGP7XX then begin
    PGPFailure(-1, X509_VERSION_ERROR);
    Exit;
  end;
  with reResults do begin
    Clear;
    Refresh;
    Alignment := taLeftJustify;
  end;
  X509CertExport.ExportCertificate(FALSE);
  Invalidate;
end;

// show selected CertProps and exported certificate on success

procedure TPGPDemo.X509CertExportCertExported(const CertPropsList: TX509CertPropsList; CertChain, ToFile: Longbool);
var
  Header: RawByteString;
  ListIndex: Integer;
begin
  reResults.Lines.BeginUpdate;
  with reResults, CertPropsList do begin
    SetFocus;
    for ListIndex := 0 to pred(Count) do begin
      Header := Header + 'Certificate of ' + X509CertProps[ListIndex].x509OwnerLongName + LFLF +
		'issued by: ' + X509CertProps[ListIndex].x509IssuerLongName + LFLF;
    end;
    with X509CertExport do begin
      if ToFile then
	DisplayText(Header + 'Exported to file "' + CertData + '"')
      else if not ArmorCert then
	DisplayText(Header + '				< ... Binary data ... >')
      else DisplayText(Header + CertData);
    end;
  end;
  JumpStart;
  reResults.Lines.EndUpdate;
end;

// import X509 Certificate

procedure TPGPDemo.btnImportCertificateClick(Sender: TObject);
begin
  if SDKVersion < PGP7XX then begin
    PGPFailure(-1, X509_VERSION_ERROR);
    Exit;
  end;
  with reResults do begin
    Clear;
    Refresh;
    Alignment := taLeftJustify;
  end;
  with OpenDialog do begin
    Title := 'Open a certificate file:';
    InitialDir := ExtractFilePath(Application.ExeName);
    Filter := 'Cert files (*.cer, *.p12, *.p7b, *.pem, *.pfx)|*.cer;*.p12;*.p7b;*.pem;*.pfx|All Files (*.*)|*.*';
    if Execute then
      X509CertImport.CertData := RawByteString(FileName)
    else X509CertImport.CertData := E;
  end;
  Refresh;
  X509CertImport.ImportCertificate(TRUE);
  Invalidate;
end;

// show selected CertProps of imported certificate on success

procedure TPGPDemo.X509CertImportCertImported(const CertPropsList: TX509CertPropsList; CertsImported: Longint);
var
  ListIndex: Integer;
begin
  reResults.Lines.BeginUpdate;
  with reResults, CertPropsList do begin
    SetFocus;
    AddLines(AnsiString(IntToStr(CertsImported)) + ' certificate(s)/key(s) imported:' + LFLF);
    for ListIndex := 0 to pred(Count) do begin
      AddLines('Certificate of ' + X509CertProps[ListIndex].x509OwnerLongName + LFLF +
	       'issued by: ' + X509CertProps[ListIndex].x509IssuerLongName + LFLF);
    end;
  end;
  JumpStart;
  reResults.Lines.EndUpdate;
end;

// display certificate properties in Windows' CryptUI dialog

procedure TPGPDemo.btnCertPropsClick(Sender: TObject);
begin
  if SDKVersion < PGP7XX then begin
    PGPFailure(-1, X509_VERSION_ERROR);
    Exit;
  end;
  reResults.Clear;
  reResults.Refresh;
  with reResults do begin
    Clear;
    Refresh;
    Alignment := taLeftJustify;
  end;
  with OpenDialog do begin
    Title := 'Open a certificate file:';
    InitialDir := ExtractFilePath(Application.ExeName);
    Filter := 'Cert files (*.cer, *.p7b, *.pem|*.cer;*.p7b;*.pem;|All Files (*.*)|*.*';
    if Execute then
      X509GetCertProps.CertData := RawByteString(FileName)
    else X509GetCertProps.CertData := E;
  end;
  Refresh;
  X509GetCertProps.DisplayCertProps(TRUE, FALSE);
  Invalidate;
end;

end.

