{------------------------------------------------------------------------------}
{                                                                              }
{                     This code is Copyright (c) 1999-2010                     }
{                           by Michael in der Wiesche                          }
{                                                                              }
{------------------------------------------------------------------------------}
{$A+,B-,C-,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z4}
unit X509Comp;

interface

uses
  Windows, Classes, TypInfo, Forms, StdCtrls, SysUtils, PGPStrTools, PGPStringList, UTF8, Base64, PGP2Comp,
  pgpBase, pgpErrors, pgpPubTypes, pgpMemoryMgr, X509Types, X509Funcs, CertStore;

// Register --------------------------------------------------------------------

procedure Register;


// Common types ----------------------------------------------------------------

type
  TCertAttributes = Class(TPersistent)
  private
    function GetSerialNumber: Longint;
    procedure SetSerialNumber(Value: Longint);
  public
    Values: TX509CertAttributes;
  published
    property CommonName: AnsiString
      read Values.CommonName
      write Values.CommonName;
    property Email: AnsiString
      read Values.Email
      write Values.Email;
    property OrganizationName: AnsiString
      read Values.OrganizationName
      write Values.OrganizationName;
    property OrganizationalUnitName: AnsiString
      read Values.OrganizationalUnitName
      write Values.OrganizationalUnitName;
    property SurName: AnsiString
      read Values.SurName
      write Values.SurName;
    property SerialNumber: Longint
      read GetSerialNumber
      write SetSerialNumber;
    property Country: AnsiString
      read Values.Country
      write Values.Country;
    property Locality: AnsiString
      read Values.Locality
      write Values.Locality;
    property State: AnsiString
      read Values.State
      write Values.State;
    property StreetAddress: AnsiString
      read Values.StreetAddress
      write Values.StreetAddress;
    property Title: AnsiString
      read Values.Title
      write Values.Title;
    property Description: AnsiString
      read Values.Description
      write Values.Description;
    property PostalCode: AnsiString
      read Values.PostalCode
      write Values.PostalCode;
    property POBOX: AnsiString
      read Values.POBOX
      write Values.POBOX;
    property PhysicalDeliveryOfficeName: AnsiString
      read Values.PhysicalDeliveryOfficeName
      write Values.PhysicalDeliveryOfficeName;
    property TelephoneNumber: AnsiString
      read Values.TelephoneNumber
      write Values.TelephoneNumber;
    property X121Address: AnsiString
      read Values.X121Address
      write Values.X121Address;
    property ISDN: AnsiString
      read Values.ISDN
      write Values.ISDN;
    property DestinationIndicator: AnsiString
      read Values.DestinationIndicator
      write Values.DestinationIndicator;
    property Name: AnsiString
      read Values.Name
      write Values.Name;
    property GivenName: AnsiString
      read Values.GivenName
      write Values.GivenName;
    property Initials: AnsiString
      read Values.Initials
      write Values.Initials;
    property HouseIdentifier: AnsiString
      read Values.HouseIdentifier
      write Values.HouseIdentifier;
    property DirectoryManagementDomain: AnsiString
      read Values.DirectoryManagementDomain
      write Values.DirectoryManagementDomain;
    property DomainComponent: AnsiString
      read Values.DomainComponent
      write Values.DomainComponent;
    property UnstructuredName: AnsiString
      read Values.UnstructuredName
      write Values.UnstructuredName;
    property UnstructuredAddress: AnsiString
      read Values.UnstructuredAddress
      write Values.UnstructuredAddress;
    property RFC822Name: AnsiString
      read Values.RFC822Name
      write Values.RFC822Name;
    property DNSName: AnsiString
      read Values.DNSName
      write Values.DNSName;
    property AnotherName: AnsiString
      read Values.AnotherName
      write Values.AnotherName;
    property IPAddress: AnsiString
      read Values.IPAddress
      write Values.IPAddress;
  end;


// Common Procs ----------------------------------------------------------------

type
  TOnFailure = procedure(ErrorCode: Longint; const ErrorMsg: String) of Object;
  TOnEnterPassphrase = procedure(const Passphrase: PAnsiChar; var Cancel: Longbool) of Object;


// TX509RequestHandler ---------------------------------------------------------

type
  TX509RequestHandler = class(TComponent)
  private
    FParentHandle: THandle;
    FArmor: Longbool;
    FCertAttributes: TCertAttributes;
    FCertificate: RawByteString;
    FCertPropsRec: TX509CertPropsRec;
    FKeyAndPassPrompt: AnsiString;
    FKeyHexIDorIASN: AnsiString;
    FKeyProps: TKeyProps;
    FKeyPropsRec: TKeyPropsRec;
    FRequestData: RawByteString;
    FValidDays: Longint;
    FOnFailure: TOnFailure;
    FOnEnterPassphrase: TOnEnterPassphrase;
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function CreateCertRequest: Longint;
    function CreateCertFromRequest: Longint;
    property ParentHandle: THandle
      read FParentHandle
      write FParentHandle;
    property Certificate: RawByteString
      read FCertificate;
    property CertPropsRec: TX509CertPropsRec
      read FCertPropsRec;
    property KeyPropsRec: TKeyPropsRec
      read FKeyPropsRec;
    property RequestData: RawByteString
      read FRequestData
      write FRequestData;
  published
    property Armor: Longbool
      read FArmor
      write FArmor;
    property CertAttributes: TCertAttributes
      read FCertAttributes
      write FCertAttributes;
    property KeyAndPassPrompt: AnsiString
      read FKeyAndPassPrompt
      write FKeyAndPassPrompt;
    property KeyHexIDorIASN: AnsiString
      read FKeyHexIDorIASN
      write FKeyHexIDorIASN;
    property KeyProps: TKeyProps
      read FKeyProps
      write FKeyProps;
    property ValidDays: Longint
      read FValidDays
      write FValidDays;
    property OnFailure: TOnFailure
      read FOnFailure
      write FOnFailure;
    property OnEnterPassphrase: TOnEnterPassphrase
      read FOnEnterPassphrase
      write FOnEnterPassphrase;
  end;


// TX509GetCertProps -----------------------------------------------------------

type
  TCertificateFilter = CertFilter_X509Cert..CertFilter_X509CanSign;

type
  TRingCertProps = class(TComponent)
  private
    FCertIDs: TRawByteStrings;
    FComboBox: TComboBox;
    FListBox: TListBox;
    FRingAlgorithmFilter: TAlgorithmKeyFilter;
    FRingBoolFilter: TBooleanKeyFilter;
    FRingCertFilter: TCertificateFilter;
    FCertPropsList: TX509CertPropsList;
    FOnFailure: TOnFailure;
    function GetCertIDs: Longint;
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Update: Longint;
    property CertPropsList: TX509CertPropsList
      read FCertPropsList;
  published
    property RingAlgorithmFilter: TAlgorithmKeyFilter
      read FRingAlgorithmFilter
      write FRingAlgorithmFilter;
    property RingBoolFilter: TBooleanKeyFilter
      read FRingBoolFilter
      write FRingBoolFilter;
    property RingCertFilter: TCertificateFilter
      read FRingCertFilter
      write FRingCertFilter;
    property ComboBox: TComboBox
      read FComboBox
      write FComboBox;
    property ListBox: TListBox
      read FListBox
      write FListBox;
    property OnFailure: TOnFailure
      read FOnFailure
      write FOnFailure;
  end;

type
  TOnGetCertProps = procedure(const CertPropsList: TX509CertPropsList) of Object;
  TX509GetCertProps = class(TRingCertProps)
  private
    FParentHandle: THandle;
    FAlgorithmFilter: TAlgorithmKeyFilter;
    FBoolFilter: TBooleanKeyFilter;
    FCertFilter: TCertificateFilter;
    FCertData: RawByteString;
    FCertDlgCaption: AnsiString;
    FCertKeyHexID: AnsiString;
    FCertKeyUserID: AnsiString;
    FIncludeSigner: Longbool;
    FSignKeyHexID: AnsiString;
    FOnGetCertProps: TOnGetCertProps;
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function DisplayCertProps(FileInput, B64Decode: Longbool): Longint;
    function VerifyCertChain(FileInput: Longbool): Longint;
    function GetCertProps(FileInput: Longbool): Longint;
    function GetRingCertProps: Longint;
    function GetCertProp(CertPropsList: TX509CertPropsList; CertIndex: Longint; CertProp: TX509CertProp): Variant;
    property ParentHandle: THandle
      read FParentHandle
      write FParentHandle;
    property CertData: RawByteString
      read FCertData
      write FCertData;
  published
    property AlgorithmFilter: TAlgorithmKeyFilter
      read FAlgorithmFilter
      write FAlgorithmFilter;
    property BoolFilter: TBooleanKeyFilter
      read FBoolFilter
      write FBoolFilter;
    property CertFilter: TCertificateFilter
      read FCertFilter
      write FCertFilter;
    property CertDlgCaption: AnsiString
      read FCertDlgCaption
      write FCertDlgCaption;
    property CertKeyHexID: AnsiString
      read FCertKeyHexID
      write FCertKeyHexID;
    property CertKeyUserID: AnsiString
      read FCertKeyUserID
      write FCertKeyUserID;
    property IncludeSigner: Longbool
      read FIncludeSigner
      write FIncludeSigner;
    property SignKeyHexID: AnsiString
      read FSignKeyHexID
      write FSignKeyHexID;
    property OnGetCertProps: TOnGetCertProps
      read FOnGetCertProps
      write FOnGetCertProps;
end;


// TX509SetCertProps -----------------------------------------------------------

type
  TOnCertPropsChanged = procedure(const CertPropsList: TX509CertPropsList) of Object;
  TX509SetCertProps = class(TComponent)
  private
    FParentHandle: THandle;
    FCertKeyPrompt: AnsiString;
    FSignCertIASN: AnsiString;
    FSignKeyAndPassPrompt: AnsiString;
    FTargetCertIASN: AnsiString;
    FOnFailure: TOnFailure;
    FOnCertPropsChanged: TOnCertPropsChanged;
    FOnEnterRevokerPassphrase: TOnEnterPassphrase;
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function RevokeCertificate(ValidDays: Longint): Longint;
    function RemoveCertificate(IncludeUserID: Longbool): Longint;
    property ParentHandle: THandle
      read FParentHandle
      write FParentHandle;
  published
    property CertKeyPrompt: AnsiString
      read FCertKeyPrompt
      write FCertKeyPrompt;
    property SignCertIASN: AnsiString
      read FSignCertIASN
      write FSignCertIASN;
    property SignKeyAndPassPrompt: AnsiString
      read FSignKeyAndPassPrompt
      write FSignKeyAndPassPrompt;
    property TargetCertIASN: AnsiString
      read FTargetCertIASN
      write FTargetCertIASN;
    property OnFailure: TOnFailure
      read FOnFailure
      write FOnFailure;
    property OnCertPropsChanged: TOnCertPropsChanged
      read FOnCertPropsChanged
      write FOnCertPropsChanged;
    property OnEnterRevokerPassphrase: TOnEnterPassphrase
      read FOnEnterRevokerPassphrase
      write FOnEnterRevokerPassphrase;
end;


// TX509CertGenerate -----------------------------------------------------------

type
  TX509CertGenerate = class(TComponent)
  private
    FParentHandle: THandle;
    FCertAttributes: TCertAttributes;
    FCertKeyHexID: AnsiString;
    FCertKeyPrompt: AnsiString;
    FCertPropsList: TX509CertPropsList;
    FSignCertIASN: AnsiString;
    FSignKeyAndPassPrompt: AnsiString;
    FValidDays: Longint;
    FOnFailure: TOnFailure;
    FOnEnterIssuerPassphrase: TOnEnterPassphrase;
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function CreateCertificate(SelfSigned: Longbool): Longint;
    property ParentHandle: THandle
      read FParentHandle
      write FParentHandle;
    property CertPropsList: TX509CertPropsList
      read FCertPropsList;
  published
    property CertAttributes: TCertAttributes
      read FCertAttributes
      write FCertAttributes;
    property CertKeyHexID: AnsiString
      read FCertKeyHexID
      write FCertKeyHexID;
    property CertKeyPrompt: AnsiString
      read FCertKeyPrompt
      write FCertKeyPrompt;
    property SignCertIASN: AnsiString
      read FSignCertIASN
      write FSignCertIASN;
    property SignKeyAndPassPrompt: AnsiString
      read FSignKeyAndPassPrompt
      write FSignKeyAndPassPrompt;
    property ValidDays: Longint
      read FValidDays
      write FValidDays;
    property OnFailure: TOnFailure
      read FOnFailure
      write FOnFailure;
    property OnEnterIssuerPassphrase: TOnEnterPassphrase
      read FOnEnterIssuerPassphrase
      write FOnEnterIssuerPassphrase;
end;


// TX509CertImport -------------------------------------------------------------

type
  TOnCertImported = procedure(const CertPropsList: TX509CertPropsList; CertsImported: Longint) of Object;
  TX509CertImport = class(TComponent)
  private
    FParentHandle: THandle;
    FCertData: RawByteString;
    FCertKeyPrompt: AnsiString;
    FCertPassPrompt: AnsiString;
    FOnFailure: TOnFailure;
    FOnCertImported: TOnCertImported;
    FOnEnterPKCS12Passphrase: TOnEnterPassphrase;
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function ImportCertificate(FileInput: Longbool): Longint;
    property ParentHandle: THandle
      read FParentHandle
      write FParentHandle;
    property CertData: RawByteString
      read FCertData
      write FCertData;
  published
    property CertKeyPrompt: AnsiString
      read FCertKeyPrompt
      write FCertKeyPrompt;
    property CertPassPrompt: AnsiString
      read FCertPassPrompt
      write FCertPassPrompt;
    property OnFailure: TOnFailure
      read FOnFailure
      write FOnFailure;
    property OnCertImported: TOnCertImported
      read FOnCertImported
      write FOnCertImported;
    property OnEnterPKCS12Passphrase: TOnEnterPassphrase
      read FOnEnterPKCS12Passphrase
      write FOnEnterPKCS12Passphrase;
end;


// TX509CertExport -------------------------------------------------------------

type
  TOnCertExported = procedure(const CertPropsList: TX509CertPropsList; CertChain, ToFile: Longbool) of Object;
  TX509CertExport = class(TComponent)
  private
    FParentHandle: THandle;
    FArmorCert: Longbool;
    FCertChain: Longbool;
    FCertData: RawByteString;
    FCertKeyOrKeyAndPassPrompt: AnsiString;
    FExportPrivate: Longbool;
    FKeyHexIDorIASN: AnsiString;
    FOnFailure: TOnFailure;
    FOnCertExported: TOnCertExported;
    FOnEnterSecKeyPassphrase: TOnEnterPassphrase;
    procedure SetArmorCert(Value: Longbool);
    procedure SetCertChain(Value: Longbool);
    procedure SetExportPrivate(Value: Longbool);
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function ExportCertificate(FileOutput: Longbool): Longint;
    property ParentHandle: THandle
      read FParentHandle
      write FParentHandle;
    property CertData: RawByteString
      read FCertData
      write FCertData;
  published
    property ArmorCert: Longbool
      read FArmorCert
      write SetArmorCert;
    property CertChain: Longbool
      read FCertChain
      write SetCertChain;
    property CertKeyOrKeyAndPassPrompt: AnsiString
      read FCertKeyOrKeyAndPassPrompt
      write FCertKeyOrKeyAndPassPrompt;
    property ExportPrivate: Longbool
      read FExportPrivate
      write SetExportPrivate;
    property KeyHexIDorIASN: AnsiString
      read FKeyHexIDorIASN
      write FKeyHexIDorIASN;
    property OnFailure: TOnFailure
      read FOnFailure
      write FOnFailure;
    property OnCertExported: TOnCertExported
      read FOnCertExported
      write FOnCertExported;
    property OnEnterSecKeyPassphrase: TOnEnterPassphrase
      read FOnEnterSecKeyPassphrase
      write FOnEnterSecKeyPassphrase;
  end;


implementation

// Passphrase event handler ----------------------------------------------------

function GetPassphrase(var Passphrase: PAnsiChar; const PassphraseProc: TOnEnterPassphrase): Longint;
var
  Cancel: Longbool;
begin
  Passphrase := PGPNewSecureData(PGPGetDefaultMemoryMgr, succ(MaxPassphraseLength) * UTF8Factor, kPGPMemoryMgrFlags_Clear);
  if Passphrase <> NIL then begin
    Cancel := FALSE;
    Result := kPGPError_UserAbort;
    PassphraseProc(Passphrase, Cancel);
    if not Cancel then begin
      if SDKVersion >= PGP8XX then SecureAnsiToUtf8PChar(Passphrase, Passphrase, succ(MaxPassphraseLength) * UTF8Factor);
      Result := kPGPError_NoErr;
    end;
  end
  else Result := kPGPError_OutOfMemory;
end;


// TCertAttributes -------------------------------------------------------------

function TCertAttributes.GetSerialNumber: Longint;
begin
  Result := StrToIntDef(Values.SerialNumber, 0);
end;

procedure TCertAttributes.SetSerialNumber(Value: Longint);
begin
  Values.SerialNumber := IntToStr(Value);
end;


// TX509RequestHandler ---------------------------------------------------------

constructor TX509RequestHandler.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCertAttributes := TCertAttributes.Create;
  if AOwner is TForm then FParentHandle := TForm(AOwner).Handle;
end;

destructor TX509RequestHandler.Destroy;
begin
  FCertAttributes.Free;
  inherited Destroy;
end;

function TX509RequestHandler.CreateCertRequest: Longint;
var
  CertKeyPassphrase: PAnsiChar;
begin
  Result := kPGPError_BadParams;
  try
    if (FKeyAndPassPrompt <> E) or Assigned(FOnEnterPassphrase) then begin
      repeat
	CertKeyPassphrase := NIL;
	if (FKeyAndPassPrompt = E) and Assigned(FOnEnterPassphrase) then begin
	  Result := GetPassphrase(CertKeyPassphrase, FOnEnterPassphrase);
	end;
	try
	  if (FKeyAndPassPrompt <> E) or (Result = 0) then begin
	    Result := CreateX509CertificateRequest(FKeyAndPassPrompt, FKeyHexIDorIASN, CertKeyPassphrase,
						   FCertAttributes.Values, FArmor, FRequestData,
						   FKeyPropsRec, GetKeyPropsFlags(FKeyProps), FParentHandle);
	  end;
	finally
	  PGPFreeData(CertKeyPassphrase);
	end;
      until (Result <> kPGPError_BadPassphrase) or not Assigned(FOnEnterPassphrase);
    end;
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TX509RequestHandler.CreateCertFromRequest: Longint;
var
  SignKeyPassphrase: PAnsiChar;
  CertPropsList: TX509CertPropsList;
begin
  Result := kPGPError_BadParams;
  try
    if (FKeyAndPassPrompt <> E) or Assigned(FOnEnterPassphrase) then begin
      repeat
	SignKeyPassphrase := NIL;
	if (FKeyAndPassPrompt = E) and Assigned(FOnEnterPassphrase) then begin
	  Result := GetPassphrase(SignKeyPassphrase, FOnEnterPassphrase);
	end;
	try
	  if (FKeyAndPassPrompt <> E) or (Result = 0) then begin
	    CertPropsList := NIL;
	    try
	      Result := CreateX509CertificateFromRequest(FKeyAndPassPrompt, FKeyHexIDorIASN, SignKeyPassphrase,
							 FCertAttributes.Values, FValidDays, FArmor, FRequestData,
							 FCertificate, CertPropsList, FParentHandle);
	      if (Result = 0) and (CertPropsList.Count > 0) then begin
		FCertPropsRec := PX509CertPropsRec(CertPropsList.Objects[0])^;
	      end;
	    finally
	      CertPropsList.Free;
	    end;
	  end;
	finally
	  PGPFreeData(SignKeyPassphrase);
	end;
      until (Result <> kPGPError_BadPassphrase) or not Assigned(FOnEnterPassphrase);
    end;
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;


// TX509GetCertProps -----------------------------------------------------------

constructor TRingCertProps.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCertIDs := TRawByteStrings.Create;
  FCertPropsList := TX509CertPropsList.Create;
  FRingCertFilter := TCertificateFilter(CertFilter_X509Cert);
end;

destructor TRingCertProps.Destroy;
begin
  FCertPropsList.Free;
  FCertIDs.Free;
  inherited Destroy;
end;

function TRingCertProps.GetCertIDs: Longint;
var
  CertIndex	: Longint;
begin
  Result := 0;
  if (FCertIDs <> NIL) and (FCertPropsList <> NIL) and (FCertPropsList.Count > 0) then begin
    FCertIDs.Clear;
    FCertIDs.Sorted := TRUE;
    with FCertPropsList do begin
      for CertIndex := 0 to pred(Count) do with pX509CertPropsRec(Objects[CertIndex])^ do begin
	Move(CertIndex, FCertIDs.Add(x509OwnerLongName));
	inc(Result);
      end;
    end;
    if Assigned(FComboBox) then FComboBox.Items.Assign(FCertIDs);
    if Assigned(FListBox) then FListBox.Items.Assign(FCertIDs);
  end;
end;

function TRingCertProps.Update: Longint;
begin
  try
    FCertPropsList.Clear;
    Result := FindX509CertProps(E, E, E,
    				GetKeyFilter(FRingAlgorithmFilter, FRingBoolFilter, TCertKeyFilter(FRingCertFilter)),
				FALSE, FCertPropsList);
    if Result >= 0 then
      Result := GetCertIDs
    else begin
      if Assigned(FComboBox) then FComboBox.Items.Clear;
      if Assigned(FListBox) then FListBox.Items.Clear;
      ShowError(FOnFailure, Result, E);
    end;
  except
    on EFailException do Raise;
    on E: Exception do begin
      Result := ShowError(FOnFailure, -1, E.Message);
      if Assigned(FComboBox) then FComboBox.Items.Clear;
      if Assigned(FListBox) then FListBox.Items.Clear;
    end;
  end;
end;

constructor TX509GetCertProps.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCertFilter := TCertificateFilter(CertFilter_X509Cert);
  if AOwner is TForm then FParentHandle := TForm(AOwner).Handle;
end;

destructor TX509GetCertProps.Destroy;
begin
  inherited Destroy;
end;

function TX509GetCertProps.DisplayCertProps(FileInput, B64Decode: Longbool): Longint;
begin
  try
    if B64Decode and not FileInput then
      Result := ShowCertificateDialog(MimeDecodeString(FCertData), FCertDlgCaption, FileInput, FParentHandle)
    else Result := ShowCertificateDialog(FCertData, FCertDlgCaption, FileInput, FParentHandle);
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TX509GetCertProps.VerifyCertChain(FileInput: Longbool): Longint;
begin
  try
    Result := VerifyX509CertificateChain(FCertData, FileInput);
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TX509GetCertProps.GetCertProps(FileInput: Longbool): Longint;
var
  CertPropsList: TX509CertPropsList;
begin
  try
    CertPropsList := NIL;
    try
      Result := GetX509CertificateProperties(FCertData, FileInput, CertPropsList);
      if Result >= 0 then begin
	if Assigned(FOnGetCertProps) then FOnGetCertProps(CertPropsList);
      end
      else ShowError(FOnFailure, Result, E);
    finally
      CertPropsList.Free;
    end;
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TX509GetCertProps.GetRingCertProps: Longint;
var
  CertPropsList: TX509CertPropsList;
begin
  try
    CertPropsList := NIL;
    try
      Result := FindX509CertProps(FCertKeyUserID, FCertKeyHexID, FSignKeyHexID,
				  GetKeyFilter(FAlgorithmFilter, FBoolFilter, TCertKeyFilter(FCertFilter)),
				  FIncludeSigner, CertPropsList);
      if Result >= 0 then begin
	if Assigned(FOnGetCertProps) then FOnGetCertProps(CertPropsList);
      end
      else ShowError(FOnFailure, Result, E);
    finally
      CertPropsList.Free;
    end;
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TX509GetCertProps.GetCertProp(CertPropsList: TX509CertPropsList; CertIndex: Longint;
				       CertProp: TX509CertProp): Variant;
begin
  Result := kPGPError_NoErr;
  if (CertPropsList <> NIL) and (CertIndex >= 0) and (CertIndex < CertPropsList.Count) then begin
    with PX509CertPropsRec(CertPropsList.Objects[CertIndex])^ do begin
      case CertProp of
	CertProp_CertSerialNumber: Result := x509CertificateID;
	CertProp_IssuerKeyAlgorithm: Result := x509IssuerKeyAlgorithm;
	CertProp_IssuerHashAlgorithm: Result := x509IssuerHashAlgorithm;
	CertProp_IssuerLongName: Result := x509IssuerLongName;
	CertProp_IssuerSerialNumber: Result := x509IssuerSerialNumber;
	CertProp_OwnerHexID: Result := x509OwnerKeyHexID;
	CertProp_OwnerLongName: Result := x509OwnerLongName;
	CertProp_OwnerSerialNumber: Result := x509OwnerSerialNumber;
	CertProp_CreaTimeStr: Result := x509CreaTimeStr;
	CertProp_ExpTimeStr: Result := x509ExpTimeStr;
	CertProp_CreaTimeNum: Result := x509CreaTimeNum;
	CertProp_ExpTimeNum: Result := x509ExpTimeNum;
	CertProp_CanEncryptData: Result := x509CanEncryptData;
	CertProp_CanSignData: Result := x509CanSignData;
	CertProp_CanSignCerts: Result := x509CanSignCerts;
	CertProp_CanSignCRLs: Result := x509CanSignCRLs;
	CertProp_Verified: Result := x509IsVerified;
	CertProp_Corrupt: Result := x509IsCorrupt;
	CertProp_Revoked: Result := x509IsRevoked;
	CertProp_Expired: Result := x509IsExpired;
	CertProp_Root: Result := x509IsRoot;
      else
	ShowError(FOnFailure, kPGPError_BadParams, E);
      end;
    end;
  end
  else ShowError(FOnFailure, kPGPError_BadParams, E);
end;


// TX509SetCertProps -----------------------------------------------------------

constructor TX509SetCertProps.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if AOwner is TForm then FParentHandle := TForm(AOwner).Handle;
end;

destructor TX509SetCertProps.Destroy;
begin
  inherited Destroy;
end;

function TX509SetCertProps.RevokeCertificate(ValidDays: Longint): Longint;
var
  SignKeyPassphrase: PAnsiChar;
  CertPropsList: TX509CertPropsList;
begin
  Result := kPGPError_BadParams;
  try
    if (FSignKeyAndPassPrompt <> E) or Assigned(FOnEnterRevokerPassphrase) then begin
      repeat
	SignKeyPassphrase := NIL;
	if (FSignKeyAndPassPrompt = E) and Assigned(FOnEnterRevokerPassphrase) then begin
	  Result := GetPassphrase(SignKeyPassphrase, FOnEnterRevokerPassphrase);
	end;
	try
	  if (FSignKeyAndPassPrompt <> E) or (Result = 0) then begin
	    CertPropsList := NIL;
	    try
	      Result := CreateX509CertificateRevocationList(FCertKeyPrompt, FSignKeyAndPassPrompt,
							    FTargetCertIASN, FSignCertIASN,
							    SignKeyPassphrase, ValidDays,
							    CertPropsList, FParentHandle);
	      if (Result = 0) and (CertPropsList.Count > 0) and Assigned(FOnCertPropsChanged) then begin
		FOnCertPropsChanged(CertPropsList);
	      end;
	    finally
	      CertPropsList.Free;
	    end;
	  end;
	finally
	  PGPFreeData(SignKeyPassphrase);
	end;
      until (Result <> kPGPError_BadPassphrase) or not Assigned(FOnEnterRevokerPassphrase);
    end;
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;

function TX509SetCertProps.RemoveCertificate(IncludeUserID: Longbool): Longint;
var
  CertPropsList: TX509CertPropsList;
begin
  try
    CertPropsList := TX509CertPropsList.Create;
    try
      Result := RemoveX509Certificate(FCertKeyPrompt, FTargetCertIASN, IncludeUserID, CertPropsList, ParentHandle);
      if (Result = 0) and (CertPropsList.Count > 0) and Assigned(FOnCertPropsChanged) then begin
	FOnCertPropsChanged(CertPropsList);
      end;
    finally
      CertPropsList.Free;
    end;
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;


// TX509CertGenerate -----------------------------------------------------------

constructor TX509CertGenerate.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCertAttributes := TCertAttributes.Create;
  FCertPropsList := TX509CertPropsList.Create;
  if AOwner is TForm then FParentHandle := TForm(AOwner).Handle;
end;

destructor TX509CertGenerate.Destroy;
begin
  FCertAttributes.Free;
  FCertPropsList.Free;
  inherited Destroy;
end;

function TX509CertGenerate.CreateCertificate(SelfSigned: Longbool): Longint;
var
  SignKeyPassphrase: PAnsiChar;
begin
  Result := kPGPError_BadParams;
  try
    if (FSignKeyAndPassPrompt <> E) or Assigned(FOnEnterIssuerPassphrase) then begin
      repeat
	SignKeyPassphrase := NIL;
	if (FSignKeyAndPassPrompt = E) and Assigned(FOnEnterIssuerPassphrase) then begin
	  Result := GetPassphrase(SignKeyPassphrase, FOnEnterIssuerPassphrase);
	end;
	try
	  if (FSignKeyAndPassPrompt <> E) or (Result = 0) then begin
	    FCertPropsList.Clear;
	    Result := CreateX509Certificate(FCertKeyPrompt, FSignKeyAndPassPrompt, FCertKeyHexID, FSignCertIASN,
					    SignKeyPassphrase, FCertAttributes.Values, FValidDays,
					    SelfSigned, FCertPropsList, FParentHandle);
	  end;
	finally
	  PGPFreeData(SignKeyPassphrase);
	end;
      until (Result <> kPGPError_BadPassphrase) or not Assigned(FOnEnterIssuerPassphrase);
    end;
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;


// TX509CertImport -------------------------------------------------------------

constructor TX509CertImport.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if AOwner is TForm then FParentHandle := TForm(AOwner).Handle;
end;

destructor TX509CertImport.Destroy;
begin
  inherited Destroy;
end;

function TX509CertImport.ImportCertificate(FileInput: Longbool): Longint;
var
  CertPropsList: TX509CertPropsList;
  ImportCertPassphrase: PAnsiChar;
begin
  Result := kPGPError_NoErr;
  try
    repeat
      ImportCertPassphrase := NIL;
      if (FCertPassPrompt = E) and Assigned(FOnEnterPKCS12Passphrase) then begin
	Result := GetPassphrase(ImportCertPassphrase, FOnEnterPKCS12Passphrase);
      end;
      try
	if Result = 0 then begin
	  CertPropsList := NIL;
	  try
	    Result := ImportX509Certificate(FCertKeyPrompt, FCertPassPrompt, FCertData, ImportCertPassphrase,
					    ImportCertPassphrase, FileInput, CertPropsList, FParentHandle);
	    if (Result = 0) or ((Result = kPGPError_FilePermissions) and (CertPropsList.Count > 0)) then begin
	      if Assigned(FOnCertImported) then FOnCertImported(CertPropsList, CertPropsList.Count);
	    end;
	  finally
	    CertPropsList.Free;
	  end;
	end;
      finally
	PGPFreeData(ImportCertPassphrase);
      end;
    until (Result <> kPGPError_BadPassphrase) or not Assigned(FOnEnterPKCS12Passphrase);
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;


// TX509CertExport -------------------------------------------------------------

constructor TX509CertExport.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if AOwner is TForm then FParentHandle := TForm(AOwner).Handle;
end;

destructor TX509CertExport.Destroy;
begin
  inherited Destroy;
end;

procedure TX509CertExport.SetArmorCert(Value: Longbool);
begin
  FArmorCert := Value and not FExportPrivate;
end;

procedure TX509CertExport.SetCertChain(Value: Longbool);
begin
  FCertChain := Value;
  if FCertChain then FExportPrivate := FALSE;
end;

procedure TX509CertExport.SetExportPrivate(Value: Longbool);
begin
  FExportPrivate := Value;
  if FExportPrivate then begin
    FArmorCert := FALSE;
    FCertChain := FALSE;
  end;
end;

function TX509CertExport.ExportCertificate(FileOutput: Longbool): Longint;
var
  ExportCertPassphrase: PAnsiChar;
  CertPropsList: TX509CertPropsList;
begin
  Result := kPGPError_BadParams;
  try
    if (FCertKeyOrKeyAndPassPrompt <> E)
    or ((FKeyHexIDorIASN <> E) and (FCertChain or (not FExportPrivate) or Assigned(FOnEnterSecKeyPassphrase))) then begin
      if not FileOutput or (FCertData <> E) then begin
	repeat
	  ExportCertPassphrase := NIL;
	  if not FExportPrivate then
	    Result := kPGPError_NoErr
	  else if (FCertKeyOrKeyAndPassPrompt = E) and Assigned(FOnEnterSecKeyPassphrase) then begin
	    Result := GetPassphrase(ExportCertPassphrase, FOnEnterSecKeyPassphrase);
	  end;
	  try
	    if (FCertKeyOrKeyAndPassPrompt <> E) or (Result = 0) then begin
	      CertPropsList := NIL;
	      try
		Result := ExportX509Certificate(FCertKeyOrKeyAndPassPrompt, FKeyHexIDorIASN, ExportCertPassphrase,
						FArmorCert, FCertChain, FExportPrivate, FileOutput,
						FCertData, CertPropsList, FParentHandle);
		if (Result = 0) and Assigned(FOnCertExported) then FOnCertExported(CertPropsList, FCertChain, FileOutput);
	      finally
		CertPropsList.Free;
	      end;
	    end;
	  finally
	    PGPFreeData(ExportCertPassphrase);
	  end;
	until (Result <> kPGPError_BadPassphrase) or not Assigned(FOnEnterSecKeyPassphrase);
      end;
    end;
    if Result <> 0 then ShowError(FOnFailure, Result, E);
  except
    on EFailException do Raise;
    on E: Exception do Result := ShowError(FOnFailure, -1, E.Message);
  end;
end;


// Register --------------------------------------------------------------------

procedure Register;
begin
  RegisterComponents('X509', [TX509RequestHandler, TX509GetCertProps, TX509SetCertProps,
			      TX509CertGenerate, TX509CertImport, TX509CertExport]);
end;


end.

