{$J+,Z4}
unit pgpBase;

{ $DEFINE SILENT_FAILURE}

// by removing the space preceding $DEFINE you may enable SILENT_FAILURE,
// which disables all initialization error messages in case of a
// missing or incorrect PGP installation; this requires checking
// PGPInitErrorCode before using any PGP functionality, though

{ $DEFINE FORCE_LOCAL_EXEC}

// for some reason PGP 7.x doesn't show the key generation progress properly any more
// if kPGPFlags_ForceLocalExecution is omitted as library initialization flag;
// but if enabled PGP 7.x returns a "file locked" error when accessing keyring
// files while either PGPnet or PGPkeys are running simultaneously;
// it's now up to you to decide which one is more important ...

{**********************************************************************************}
{                                                                                  }
{ The contents of this file are subject to the Mozilla Public License Version 1.1  }
{ (the "License"); you may not use this file except in compliance with the         }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.    }
{                                                                                  }
{ Software distributed under the License is distributed on an "AS IS" basis,       }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the }
{ specific language governing rights and limitations under the License.            }
{                                                                                  }
{ The Original Code is the "Borland Delphi Runtime Library PGPsdk" released 10 Apr }
{ 2000, available at http://www.oz.net/~srheller/dpgp/sdk/.                        }
{                                                                                  }
{ The Initial Developer of the Original Code is Steven R. Heller.                  }
{                                                                                  }
{ Portions created by Steven R. Heller are Copyright (C) 2000 Steven R. Heller.    }
{ All Rights Reserved.                                                             }
{                                                                                  }
{ Contributor(s): Michael in der Wiesche <idw.doc@t-online.de> ("idw").            }
{                                                                                  }
{ The original file is pgpBase.pas based on pgpBase.h                              }
{ from the PGP sources which are Copyright (C) PGP Corporation.                    }
{                                                                                  }
{ Modifications by "idw" (other than stated in the code below):                    }
{                                                                                  }
{**********************************************************************************}

interface

uses
  Windows, Messages,
  {$IFDEF UNICODE}
  AnsiStrings,
  {$ENDIF}
  SysUtils;

{$IFNDEF UNICODE}
type
  AnsiString = type String;
  RawByteString = String;
{$ENDIF}

const	// added by idw
  PathPGP	= '\PGP Corporation\PGP';
  PathNAI	= '\Network Associates\PGP';
  ErrorTitle	= 'PGP Components fatal error:';
  SDKError	= 'PGP Software Development Kit Core Library not found';
  SDKNLError	= 'PGP Software Development Kit Networking Library not found';
  SDKUIError	= 'PGP Software Development Kit User Interface Library not found';
  CLError	= 'PGP Client Library not found';
  SCError	= 'PGP Cryptographic Support Library not found';
  SDKInitError	= 'Cannot initialize PGP SDK (e.g. PGP(sdk)Serv.exe)';
  VersionError	= 'Wrong PGP version: 6.5.x, 7.x, 8.x, 9.x or 10.x required';

type	// added by idw
  TSDKVersion = (PGPXXX, PGP65X, PGP7XX, PGP8XX, PGP81X, PGP9XX, PGP95X, PGP96X, PGP97X, PGP98X, PGP99X, PGP91XX, PGP10XX);
  TPGPInitErrorCode = (ieNone, ieSDK, ieSDKNL, ieSDKUI, ieCL, ieSC, ieUTF8, ieSDKInit, ieVersion);
  TVersionString = Array[0..255] of AnsiChar;

const	// added by idw
  PGPInitErrorCode: TPGPInitErrorCode = ieNone;
  SDKVersion: TSDKVersion = PGPXXX;
  PGPPath: RawByteString = '';
  PGPVersion: RawByteString = '';
  PGPsdkLib: PAnsiChar = NIL;
  PGPsdkNLLib: PAnsiChar = NIL;
  PGPsdkUILib: PAnsiChar = NIL;
  PGPclLib: PAnsiChar = NIL;
  PGPscLib: PAnsiChar = NIL;
  hPGPsdkLib: Longint = 0;
  hPGPsdkNLLib: Longint = 0;
  hPGPsdkUILib: Longint = 0;
  hPGPclLib: Longint = 0;
  hPGPscLib: Longint = 0;
  InitFlags: Longint = 0;

type
  PGPBoolean	  = Byte;			// can be TRUE (1) or FALSE (0)
  PGPInt8	  = Byte;
  PGPUInt8	  = Byte;
  PGPInt16	  = Smallint;
  PGPUInt16	  = Word;
  PGPInt32	  = Longint;
  PGPUInt32	  = DWord;			// changed from Cardinal to DWord by idw
  PGPByte	  = PGPUInt8;
  PGPError	  = PGPInt32;
  PGPEnumType	  = Longint;			// added by idw
  PGPUserValue	  = Pointer;
  PGPSize	  = PGPUInt32;
  PGPFlags	  = PGPUInt32;
  PGPFileOffset	  = PGPInt32;
  PGPTime	  = PGPUInt32;
  PGPTimeInterval = PGPUInt32;			// in milliseconds

  TPGPVersion = Record
    MajorVersion: PGPUInt16;
    MinorVersion: PGPUInt16;
  end;

const
  PGPFalse  = 0;
  PGPTrue   = 1;

const	// 7.x & later from pgpUtilities.h
  kPGPFlags_ForceLocalExecution	= $02;
  kPGPFlags_SuppressCacheThread	= $04;

const	// added by idw
  MaxPassphraseLength = 255;

var	// added by idw
  MyVersion: TVersionString;

var	// added by idw from pgpUtilities.h
  PGPsdkInit: function(Options: PGPFlags): PGPError; cdecl;		// Options is ignored prior to 7.x
  PGPsdkCleanup: function: PGPError; cdecl;
  PGPsdkReconnect: function: PGPError; cdecl;

var
  // 6.5.x: PGPGetSDKString from pgpFeatures.h
  // 7.x & later: PGPGetPGPsdkVersionString from pgpUtilities.h
  PGPGetSDKString: function(const TheString: TVersionString): PGPError; cdecl;
  // from pgpErrors.h
  PGPGetErrorString: function(TheErrorCode: PGPError; AvailLength: PGPSize; TheErrorText: PAnsiChar): PGPError; cdecl;

procedure ProcessMessages;								// added by idw
function GetProductVersion(const BinaryName: RawByteString): RawByteString;		// added by idw
function GetRegEntry(Key: HKEY; const Path, Name: RawByteString): RawByteString;	// added by idw


implementation	// code added by idw

uses
  pgpErrors;

procedure ProcessMessages;
var
  Msg: TMsg;
begin
  while PeekMessageA(Msg, 0, 0, 0, PM_REMOVE) and (Msg.Message <> WM_QUIT) do begin
    TranslateMessage(Msg);
    DispatchMessageA(Msg);
  end;
end;

function GetProductVersion(const BinaryName: RawByteString): RawByteString;

  function VersionInfo(const FileName: RawByteString): RawByteString;
  var
    dwSize: DWord;
    hInfo: THandle;
  begin
    dwSize := GetFileVersionInfoSizeA(PAnsiChar(FileName), hInfo);
    if dwSize > 0 then begin
      SetString(Result, NIL, dwSize);
      if GetFileVersionInfoA(PAnsiChar(FileName), hInfo, dwSize, Pointer(Result)) then Exit;
    end;
    Result := '';
  end;

  function VerValue(const Info, Name: RawByteString; var Data: Pointer): DWord;
  begin
    Result := 0;
    if not VerQueryValueA(Pointer(Info), PAnsiChar(Name), Data, Result) then Exit;
  end;

  function VerString(const Info, Language, Value: RawByteString): RawByteString;
  const
    FILE_INFO = '\StringFileInfo\';
  var
    p: Pointer;
    c: Cardinal;
  begin
    c := VerValue(Info, FILE_INFO + Language + Value, p);
    if c > 0 then begin
      SetString(Result, PAnsiChar(p), pred(c));
      Exit;
    end;
    Result := '';
  end;

  function VerTranslation(const Info: RawByteString): RawByteString;
  const
    VFI_TRANSLATION = '\VarFileInfo\Translation';
  type
    PTranslationInfo = ^TTranslationInfo;
    TTranslationInfo = Record
      wLanguage: Word;
      wCodePage: Word;
    end;
  var
    pti: PTranslationInfo;
  begin
    if VerValue(Info, VFI_TRANSLATION, Pointer(pti)) > 0 then begin
      with pti^ do Result := RawByteString(IntToHex(wLanguage shl 16 + wCodePage, 8));
      Exit;
    end;
    Result := '';
  end;

const
  PRODUCT_VERSION = '\ProductVersion';
var
  sInfo: RawByteString;
begin
  sInfo := VersionInfo(BinaryName);
  if sInfo <> '' then begin
    Result := VerString(sInfo, VerTranslation(sInfo), PRODUCT_VERSION);
    Exit;
  end;
  Result := '';
end;

function GetRegEntry(Key: HKEY; const Path, Name: RawByteString): RawByteString;
var
  hRegKey	 : HKEY;
  dwBufSize	 : DWORD;
begin
  if RegOpenKeyExA(Key, PAnsiChar(Path), 0, KEY_QUERY_VALUE, hRegKey) = ERROR_SUCCESS then begin
    try
      dwBufSize := 0;
      if RegQueryValueExA(hRegKey, PAnsiChar(Name), NIL, NIL, NIL, @dwBufSize) = ERROR_SUCCESS then begin
	SetString(Result, NIL, pred(dwBufSize));
	if RegQueryValueExA(hRegKey, PAnsiChar(Name), NIL, NIL, Pointer(Result), @dwBufSize) = ERROR_SUCCESS then Exit;
      end;
    finally
      RegCloseKey(hRegKey);
    end;
  end;
  Result := '';
end;

function PGPInitError(ErrorCode: TPGPInitErrorCode): Longbool;
{$IFNDEF SILENT_FAILURE}
var
  ErrorMsg: RawByteString;
{$ENDIF}
begin
  PGPInitErrorCode := ErrorCode;
  {$IFNDEF SILENT_FAILURE}
  ErrorMsg := '';
  case ErrorCode of
    ieSDK:	ErrorMsg := SDKError;
    ieSDKNL:	ErrorMsg := SDKNLError;
    ieSDKUI:	ErrorMsg := SDKUIError;
    ieCL:	ErrorMsg := CLError;
    ieSC:	ErrorMsg := SCError;
    ieSDKInit:	ErrorMsg := SDKInitError;
    ieVersion:	ErrorMsg := VersionError;
  end;
  MessageBoxA(0, PAnsiChar(ErrorMsg), ErrorTitle, MB_ICONERROR);
  //Halt(ERROR_DLL_INIT_FAILED);
  {$ENDIF}
  Result := TRUE;
end;

function GetMajorVersion(const Version: RawByteString): Cardinal;
const
  DIGITS = ['0'..'9'];
var
  sVersion: RawByteString;
  pVersion: PAnsiChar;
  iIndex, iCount: Integer;
begin
  Result := 0;
  if Version <> '' then begin
    iIndex := 1;
    while (Version[iIndex] <> #0) and not (Version[iIndex] in DIGITS) do inc(iIndex);
    if Version[iIndex] = #0 then Exit;
    SetString(sVersion, NIL, succ(Length(Version) - iIndex));
    for iCount := 0 to 1 do begin
      pVersion := Pointer(sVersion);
      while Version[iIndex] in DIGITS do begin
	pVersion^ := Version[iIndex];
	inc(pVersion);
	inc(iIndex);
      end;
      inc(iIndex);
      pVersion^ := #0;
      Result := Result shl 8 + Cardinal(StrToIntDef(String(sVersion), 0));
    end;
  end;
end;

// MajorVersion evaluation see unit pgpUtilities.h;
// PGP's version info returns inconsistent values
function CheckSDKVersion: TSDKVersion;
begin
  case GetMajorVersion(GetProductVersion(PGPsdkLib)) of
    $0400: Result := PGP10XX;
    $030C: Result := PGP91XX;
    $030B: Result := PGP99X;
    $030A: Result := PGP98X;
    $0309: Result := PGP97X;
    $0308: Result := PGP96X;
    $0307: Result := PGP95X;
    $0305: Result := PGP9XX;
    $0302: Result := PGP81X;
    $0300: Result := PGP8XX;
    $0200: Result := PGP7XX;
    $0107: Result := PGP65X;
  else
    Result := PGPXXX;
  end;
  SDKVersion := Result;
end;

function GetPGPPath(const RegPath: RawByteString): RawByteString;
begin
  Result := GetRegEntry(HKEY_LOCAL_MACHINE, PAnsiChar(RegPath), 'InstallPath');
end;

function GetPGPVersion: RawByteString;
var
  hInfo: THandle;
  dwSize: DWord;
  sInfo: RawByteString;
  dwFFSize: UINT;
  pFFInfo: Pointer;
begin
  dwSize := succ(GetFileVersionInfoSizeA(PGPclLib, hInfo));
  if dwSize > 1 then begin
    SetString(sInfo, NIL, dwSize);
    if GetFileVersionInfoA(PGPclLib, hInfo, dwSize, PAnsiChar(sInfo)) then begin
      if VerQueryValueA(PAnsiChar(sInfo), '\', pFFInfo, dwFFSize) and (dwFFSize <> 0) then begin
	with PVSFixedFileInfo(pFFInfo)^ do begin
	  Result := {$IFDEF UNICODE}AnsiStrings.{$ENDIF}Format('%d.%d.%d', [dwProductVersionMS shr 16,
									    dwProductVersionMS and $FFFF,
									    dwProductVersionLS shr 16]
							      );
	  Exit;
	end;
      end;
    end;
  end;
  Result := '';
end;

procedure Initialize;
begin
  // Load Software Development Kit Core Library
  PGPsdkLib := 'PGPSDK.dll';			// 5.x, 7.x & later
  hPGPsdkLib := LoadLibraryA(PGPsdkLib);
  if hPGPsdkLib = 0 then begin
    PGPsdkLib := 'PGP_SDK.dll';			// 6.x
    hPGPsdkLib := LoadLibraryA(PGPsdkLib);
  end;
  if (hPGPsdkLib = 0) and PGPInitError(ieSDK) then Exit;

  {$IFDEF FORCE_LOCAL_EXEC}
  InitFlags := kPGPFlags_ForceLocalExecution;
  {$ENDIF}

  // Check SDK Version
  PGPsdkInit := GetProcAddress(hPGPsdkLib, 'PGPsdkInit');
  PGPsdkCleanup := GetProcAddress(hPGPsdkLib, 'PGPsdkCleanup');
  if (PGPsdkInit(InitFlags) <> 0) and PGPInitError(ieSDKInit) then Exit;
  if (CheckSDKVersion = PGPXXX) and PGPInitError(ieVersion) then Exit;

  // For Checking PGPsdkService Status
  if SDKVersion > PGP65X then PGPsdkReconnect := GetProcAddress(hPGPsdkLib, 'PGPsdkReconnect');

  // Get SDK Version String
  if SDKVersion > PGP65X then
    PGPGetSDKString := GetProcAddress(hPGPsdkLib, 'PGPGetPGPsdkVersionString')
  else PGPGetSDKString := GetProcAddress(hPGPsdkLib, 'PGPGetSDKString');
  if PGPGetSDKString(MyVersion) <> 0 then MyVersion := 'N/A';

  PGPGetErrorString := GetProcAddress(hPGPsdkLib, 'PGPGetErrorString');

  // Load Software Development Kit Networking Library
  PGPsdkNLLib := 'PGPsdkNL.dll';
  hPGPsdkNLLib := LoadLibraryA(PGPsdkNLLib);
  if (hPGPsdkNLLib = 0) and PGPInitError(ieSDKNL) then Exit;

  // Load Software Development Kit User Interface Library
  PGPsdkUILib := 'PGPsdkUI.dll';
  hPGPsdkUILib := LoadLibraryA(PGPsdkUILib);
  if (hPGPsdkUILib = 0) and PGPInitError(ieSDKUI) then Exit;

  // Load Client Library
  if SDKVersion in [PGP7XX..PGP81X] then
    PGPclLib := 'PGPclientLib.dll'		// 7.x, 8.x
  else PGPclLib := 'PGPcl.dll';			// 6.x, 9.x
  hPGPclLib := LoadLibraryA(PGPclLib);
  if (hPGPclLib = 0) and PGPInitError(ieCL) then Exit;

  // Load Cryptographic Support Library
  PGPscLib := 'PGPsc.dll';
  hPGPscLib := LoadLibraryA(PGPscLib);
  if (hPGPscLib = 0) and PGPInitError(ieSC) then Exit;

  // Get PGP's Home Directory
  if SDKVersion > PGP7XX then
    PGPPath := GetPGPPath('Software' + PathPGP)
  else PGPPath := GetPGPPath('Software' + PathNAI);

  // Get PGP Version String
  PGPVersion := GetPGPVersion;
end;

procedure Finalize;
begin
  if Assigned(PGPsdkCleanup) then PGPsdkCleanup;

  if hPGPscLib <> 0 then FreeLibrary(hPGPscLib);
  if hPGPclLib <> 0 then FreeLibrary(hPGPclLib);
  if hPGPsdkUILib <> 0 then FreeLibrary(hPGPsdkUILib);
  if hPGPsdkNLLib <> 0 then FreeLibrary(hPGPsdkNLLib);
  if hPGPsdkLib <> 0 then FreeLibrary(hPGPsdkLib);
end;


initialization
  Initialize;

finalization
  Finalize;


end.

